
// This is file Annotation.C.  It implements a class for managing the
// document annotations and their data, described in Annotation.h.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/



#include "Annotation.h"
#include "TopShell.h"
#include "ADialogs.h"
#include <nihcl/Integer.h>
#include <string.h>
#include <stdlib.h>

#define String X11String
#define Object X11Object

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/Text.h>
#include <Xm/Form.h>
#include <Xm/ScrolledW.h>
#include <Xm/DialogS.h>
#include <Xm/Separator.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/ArrowB.h>
#include <Xm/MenuShell.h>
#include <Xm/RowColumn.h>

#undef String
#undef Object

class DocWindow;


#define	THIS	Annotation
#define	BASE	UIObject
#define	BASE_CLASSES BASE::desc()
#define	MEMBER_CLASSES
#define VIRTUAL_BASE_CLASSES

DEFINE_REDUCED_CLASS(Annotation,1,"$Header: /home/cellar/ginsburg/browser/RCS/Annotation.C,v 2.1 1996/07/05 15:42:58 ginsburg Exp $",NULL)

bool Annotation::isEqual(const Object& p) const
// Test two annotations for equality
{
  return (this == &p);
}


// Create a new annotation.  The form widget for the annotation has
// already been created, and the page, position, and annotation text
// are specified.  parent_obj gives the address of the DocWindow
// responsible for this Annotation, new_parent_deleteCB is a callback
// that deletes a specified annotation, and parent_drag_addr is the
// address to use for the "icon_drag" action.  This constructor form
// is used for creating new annotations in response to drop events.

Annotation::Annotation(Widget new_icon_form, const int new_page,
		       const float new_xpos, const float new_ypos,
		       const char *new_text, DocWindow *parent_obj,
		       CBType new_parent_deleteCB, char *parent_drag_addr)
:page(new_page),
 xpos(new_xpos),
 ypos(new_ypos),
 xpos_virtual(new_xpos),
 ypos_virtual(new_ypos),
 icon_form(new_icon_form),
 parent(parent_obj),
 small_dialog(NULL),
 parent_deleteCB(new_parent_deleteCB),
 components(0),
 link(NULL),
 link_int(-1)
{
  register_eh(icon_form, ButtonPressMask,
	      (EHType)&Annotation::raise_appropriate_menuEH, NULL, False);

  // copy the annotation text; strdup is not used so that XtMalloc can
  // perform the allocation:
  text = XtMalloc(strlen(new_text) + 1);
  strcpy(text, new_text);

  // Register ActionProcs for annotation property editing and dialog
  // link point dragging:
  edit_annotation_addr =
    register_ap("edit_annotation", (APType)&Annotation::edit_annotationAP);
  link_drag_addr = register_ap("link_drag", (APType)&Annotation::link_dragAP);

  // Create the annotation label:
  icon_label = XtVaCreateManagedWidget("icon_label", xmLabelWidgetClass, icon_form,
				       XmNlabelString, XmStringCreateSimple(""),
				       XmNtopAttachment, XmATTACH_FORM,
				       XmNbottomAttachment, XmATTACH_FORM,
				       XmNleftAttachment, XmATTACH_FORM,
				       XmNrightAttachment, XmATTACH_FORM,
				       NULL);

  register_eh(icon_label, ButtonPressMask,
	      (EHType)&Annotation::raise_appropriate_menuEH, NULL, False);

  // The translations must be recompiled for each instance as the
  // edit_annotation and parent_drag_addr addresses change:

  static const char button_translation_format[] = "<Btn2Down>: icon_drag(%s)\n";

  char *button_translations = new char[strlen(button_translation_format) + 1*16];
  sprintf(button_translations, button_translation_format, parent_drag_addr);

  button_drag_translations = XtParseTranslationTable(button_translations);

  delete [] button_translations;

  static const char label_translation_format[] = "<Btn2Down>: icon_drag(%s)\n\
                                                  <Btn1Up>(2): edit_annotation(%s)\n";

  char *label_translations = new char[strlen(label_translation_format) + 2*16];
  sprintf(label_translations, label_translation_format,
	  parent_drag_addr, edit_annotation_addr);

  label_drag_translations = XtParseTranslationTable(label_translations);

  delete [] label_translations;

  // Configure the annotation label with the appropriate translations
  XtOverrideTranslations(icon_label, label_drag_translations);

  // Acquire an annotation dialog, set its members for defining a new
  // annotation, and make it visible:
  full_dialog = get_full_dialog();

  full_dialog->set_fields(new_text, "", "OK", "Cancel", this,
			  (CBType)&Annotation::create_annotationCB,
			  (CBType)&Annotation::cancel_annotationCB,
			  (CBType)&Annotation::drop_and_linkCB);
  full_dialog->set_link(link_drag_addr);
  
  full_dialog->popup_and_manage();
}



// Create a new annotation.  The form widget for the annotation has
// already been created, and the page, position, and annotation text
// are specified.  parent_obj gives the address of the DocWindow
// responsible for this Annotation, new_parent_deleteCB is a callback
// that deletes a specified annotation, and parent_drag_addr is the
// address to use for the "icon_drag" action.  new_name gives the name
// of the annotation, and new_link/new_backlinks give the integer
// indices of the links and backlinks.  This constructor form is used
// for loading annotations that have been saved with the document.

Annotation::Annotation(Widget new_icon_form, const int new_page,
		       const float new_xpos, const float new_ypos,
		       const char *new_text,
		       DocWindow *parent_obj, CBType new_parent_deleteCB,
		       char *parent_drag_addr, char *new_name, int new_link,
		       OrderedCltn *new_backlinks)
:page(new_page),
 xpos(new_xpos),
 ypos(new_ypos),
 xpos_virtual(new_xpos),
 ypos_virtual(new_ypos),
 icon_form(new_icon_form),
 parent(parent_obj),
 small_dialog(NULL),
 full_dialog(NULL),
 parent_deleteCB(new_parent_deleteCB),
 components(0),
 link(NULL),
 link_int(new_link),
 backlinks_int(new_backlinks)
{
  register_eh(icon_form, ButtonPressMask,
	      (EHType)&Annotation::raise_appropriate_menuEH, NULL, False);

  // Copy the annotation text and name strings:
  text = XtMalloc(strlen(new_text) + 1);
  strcpy(text, new_text);

  name = XtMalloc(strlen(new_name) + 1);
  strcpy(name, new_name);

  // Register the annotation actions for property editing and dialog
  // link point dragging:
  edit_annotation_addr =
    register_ap("edit_annotation", (APType)&Annotation::edit_annotationAP);
  link_drag_addr = register_ap("link_drag", (APType)&Annotation::link_dragAP);

  // Create the annotation label:
  icon_label = XtVaCreateManagedWidget("icon_label", xmLabelWidgetClass, icon_form,
				       XmNlabelString, XmStringCreateSimple(""),
				       XmNtopAttachment, XmATTACH_FORM,
				       XmNbottomAttachment, XmATTACH_FORM,
				       XmNleftAttachment, XmATTACH_FORM,
				       XmNrightAttachment, XmATTACH_FORM,
				       NULL);

  register_eh(icon_label, ButtonPressMask,
	      (EHType)&Annotation::raise_appropriate_menuEH, NULL, False);

  // The translations must be recompiled for each instance as the
  // edit_annotation and parent_drag_addr addresses change:

  static const char button_translation_format[] = "<Btn2Down>: icon_drag(%s)\n";

  char *button_translations = new char[strlen(button_translation_format) + 1*16];
  sprintf(button_translations, button_translation_format, parent_drag_addr);

  button_drag_translations = XtParseTranslationTable(button_translations);

  delete [] button_translations;

  static const char label_translation_format[] = "<Btn2Down>: icon_drag(%s)\n\
                                                  <Btn1Up>(2): edit_annotation(%s)\n";

  char *label_translations = new char[strlen(label_translation_format) + 2*16];
  sprintf(label_translations, label_translation_format,
	  parent_drag_addr, edit_annotation_addr);

  label_drag_translations = XtParseTranslationTable(label_translations);

  delete [] label_translations;

  // Set the (drag and edit) translations for the annotation label
  XtOverrideTranslations(icon_label, label_drag_translations);

  // Create annotation buttons according to the value of components
  create_appropriate_buttons();
}


// Deallocate an annotation and release its dialogs

Annotation::~Annotation()
{
  delete [] edit_annotation_addr;

  if (link)
    link->backlinks.remove(*this);

  if (!backlinks.isEmpty())
    clear_backlinks();

  // Release the dialogs to the cache:
  if (full_dialog)
    {
      XtPopdown(full_dialog->get_shell());
      release_full_dialog(full_dialog);
    }
  if (small_dialog)
    {
      XtPopdown(small_dialog->get_shell());
      release_small_dialog(small_dialog);
    }

  XtFree(text);
}


// Remove the link component from the backlinked annotations,
// i.e. when this annotation is deleted.

void Annotation::clear_backlinks()
{
  for(int i = 0; i < backlinks.size(); i++)
    {
      // Unlink the annotations that reference this one:
      Annotation *backlink = (Annotation*)(backlinks[i]);

      // clear the link member and remove the hyperlink button:
      backlink->link = NULL;
      if (backlink->components & OBJ_HYPERLINK)
	{
	  backlink->components &= ~OBJ_HYPERLINK;
	  backlink->remove_hyperlink_button();
	}

      // reset the dialog link point to "unlinked":
      if (backlink->full_dialog)
	backlink->full_dialog->set_link(backlink->link_drag_addr);
    }
}


// Exchange xpos/ypos

void Annotation::swap_dimensions()
{
  const float temp = xpos;
  xpos = ypos;
  ypos = temp;
}


// xpos := 1 - ypos; ypos := xpos

void Annotation::rotate_cl()
{
  const float temp = xpos;
  xpos = 1 - ypos;
  ypos = temp;
}


// ypos := 1 - xpos; xpos := ypos

void Annotation::rotate_ccl()
{
  const float temp = xpos;
  xpos = ypos;
  ypos = 1 - temp;
}


// Display the text annotation in a SmallAnnotationDialog

void Annotation::popup_for_annotationCB(Widget, XtPointer, XtPointer)
{
  if (!small_dialog)
    small_dialog = get_small_dialog();

  small_dialog->set_fields(text);
  small_dialog->popup_and_manage();
}


// Functions for manipulating the dialog cache.  get_full/small_dialog
// retrieves a free dialog, if one exists, and otherwise creates a new
// one:

FullAnnotationDialog *Annotation::get_full_dialog()
{
  if (!(parent->full_dialog_stack.isEmpty()))
    // Pop an existing dialog and return it
    return (FullAnnotationDialog*)(parent->full_dialog_stack.pop());
  else
    // create a new dialog
    return new FullAnnotationDialog(parent->gvarea);
}


SmallAnnotationDialog *Annotation::get_small_dialog()
{
  if (!(parent->annotation_dialog_stack.isEmpty()))
    // Pop an existing dialog and return it
    return (SmallAnnotationDialog*)(parent->annotation_dialog_stack.pop());
  else
    // create a new dialog
    return new SmallAnnotationDialog(parent->gvarea);
}


// Process the results of the new annotation properties dialog

void Annotation::create_annotationCB(Widget, XtPointer, XtPointer)
{
  parent->annotations_modified = True;

  // Set the annotation name and text:
  XtFree(text);
  text = full_dialog->get_annotation();
  name = full_dialog->get_name();

  create_appropriate_buttons();

  // Update the backlinks with the new name:
  for (int i = 0; i < backlinks.size(); i++)
    {
      Annotation *this_annotation = (Annotation*)(backlinks[i]);
      if (this_annotation->full_dialog)
	this_annotation->full_dialog->set_link((char*)name,
					       parent->doc->pages[page].label);
    }

  // Popdown and release the dialog:
  XtPopdown(full_dialog->get_shell());
  release_full_dialog(full_dialog);
  full_dialog = NULL;
}


// Delete this annotation, i.e. "Cancel" from the new annotation
// dialog

void Annotation::cancel_annotationCB(Widget widget, XtPointer, XtPointer)
{
  // Popdown and release the dialog
  XtPopdown(full_dialog->get_shell());
  release_full_dialog(full_dialog);
  full_dialog = NULL;

  // Deallocate this annotation:
  clear_backlinks();
  (parent->*parent_deleteCB)(widget, this, NULL);
}


// Cancel editing changes and revert to saved values, i.e. "Cancel"
// from the annotation properties dialog

void Annotation::cancel_changesCB(Widget, XtPointer, XtPointer)
{
  // Popdown and release the dialog
  XtPopdown(full_dialog->get_shell());
  release_full_dialog(full_dialog);
  full_dialog = NULL;

  // Undo the link if necessary:
  if (link != oldlink)
    {
      // Remove this annotation from the backlinks of *link:
      if (link)
	link->backlinks.remove(*this);

      // Revert link to the saved value:
      link = oldlink;
      if (link)
	link->backlinks.add(*this);
    }
}


// Start a drag from the annotation dialog's link point with the goal
// of creating a new linked annotation

void Annotation::link_dragAP(Widget widget, XEvent *event, char *[], int num_args)
{
  // No arguments expected:
  if (num_args)
    {
      cerr << "link_drag: invalid argument list" << endl;
      return;
    }
  
  // Store this pointer in widget's userData, then pass widget to convertProc
  XtVaSetValues(widget,
		XmNuserData, this,
		NULL);

  // Start the drag:
  Atom exportList[1];

  exportList[0] = create_link_atom;

  static Arg args[10];
  int n = 0;

  // Create a drag_icon widget:
  Widget drag_icon = TopShell::create_drag_icon(widget, destination_icon_pixmaps[PIXMAPS_IMAGE],
						destination_width, destination_height);

  // Use a dragDropFinish callback to deallocate the icon:
  static XtCallbackRec dragDropFinishCB[] =
    {
      {(XtCallbackProc)destroyWidgetCB, NULL},
      {NULL, NULL}
    };

  dragDropFinishCB[0].closure = (XtPointer)drag_icon;

  // Configure a drag context and start the drag:
  XtSetArg(args[n], XmNexportTargets, exportList); n++;
  XtSetArg(args[n], XmNnumExportTargets, XtNumber(exportList)); n++;
  XtSetArg(args[n], XmNdragOperations, XmDROP_MOVE); n++;
  XtSetArg(args[n], XmNdragDropFinishCallback, dragDropFinishCB); n++;
  XtSetArg(args[n], XmNconvertProc, LinkConvertProc); n++;
  XtSetArg(args[n], XmNclientData, (XtPointer)this); n++;
  XtSetArg(args[n], XmNsourcePixmapIcon, drag_icon); n++;

  Widget drag_context = XmDragStart(widget, event, args, n);

  register_cb(drag_context, XmNdropFinishCallback,
	      (CBType)&Annotation::drop_finishCB, NULL);

  // Set the docwin_context of XtWindow(drag_context) to the DocWindow's address for
  // retrieval by drop routines
  XSaveContext(dpy, XtWindow(drag_context), docwin_context, (XPointer)parent);
}


// Setup for creating a new annotation; the actual data is the address
// of a pointer to the new object.  When the drop is received by the
// DocWindow's TransferProc, a new annotation is created and its
// address is stored in the pointer.

Boolean Annotation::LinkConvertProc(Widget dc, Atom *, Atom *target,
				    Atom *typeRtn, XtPointer *valueRtn,
				    unsigned long *lengthRtn, int *formatRtn,
				    unsigned long *, int, XtRequestId *)
{
  if (*target != create_link_atom)
    return(False);

  // Retrieve "this" pointer from the DragContext:
  Annotation *this_obj;
  XtVaGetValues(dc,
		XmNclientData, &this_obj,
		NULL);

  // Create a new Annotation** and store the address of the "link"
  // member:
  Annotation ***ptrs = new Annotation**[1];
  ptrs[0] = &(this_obj->link);

  // Describe the value format:
  *typeRtn = create_link_atom;
  *valueRtn = (XtPointer)ptrs;
  *lengthRtn = 1;
  *formatRtn = sizeof(Annotation**);

  return(True);
}


// Add this object as a backlink for *link

void Annotation::drop_finishCB(Widget, XtPointer, XmDropFinishCallbackStruct *)
{
  if (!link)
    return;

  link->backlinks.add(*this);
}


// Receive a drop in the annotation dialog's link point

void Annotation::drop_and_linkCB(Widget, int annotation_index, XtPointer)
{
  // Remove this object as a backlink from the existing link:
  if (link)
    link->backlinks.remove(*this);

  // Locate the new link:
  link = (Annotation*)(parent->annotations[annotation_index]);
  link->backlinks.add(*this);

  // Configure the dialog:
  if (full_dialog)
    full_dialog->set_link(link->name, parent->doc->pages[link->page].label);
}


// Create a dialog to edit the values of this annotation

void Annotation::edit_annotationCB(Widget, XtPointer, XtPointer)
{
  edit_annotationAP(NULL, NULL, NULL, 0);
}

void Annotation::edit_annotationAP(Widget, XEvent*, char*[], int)
{
  // Obtain a dialog:
  if (!full_dialog)
    full_dialog = get_full_dialog();

  oldlink = link;		// for a potential "Cancel"

  // Configure the dialog:
  full_dialog->set_fields(text, name, "Update", "Cancel", "Remove", this,
			  (CBType)&Annotation::create_annotationCB,
			  (CBType)&Annotation::cancel_changesCB,
			  (CBType)&Annotation::cancel_annotationCB,
			  (CBType)&Annotation::drop_and_linkCB);

  if(link)
    full_dialog->set_link(link->name, parent->doc->pages[link->page].label);
  else
    full_dialog->set_link(link_drag_addr);

  // Make the dialog visible
  full_dialog->popup_and_manage();
}


// Create a button for the text annotation and add it to icon_form

void Annotation::create_annotation_button()
{
  components |= OBJ_ANNOTATION;	// update components

  // Create the button:
  icon_annotation_button =
    XtVaCreateWidget("A", xmPushButtonWidgetClass, icon_form,
		     XmNleftAttachment, XmATTACH_FORM,
		     XmNtopAttachment, XmATTACH_FORM,
		     XmNbottomAttachment, XmATTACH_FORM,
		     XmNmarginWidth, 0,
		     XmNmarginHeight, 0,
		     NULL);

  register_eh(icon_annotation_button, ButtonPressMask,
	      (EHType)&Annotation::raise_appropriate_menuEH, NULL, False);

  // Set the labelPixmap
  assign_label_pixmap(icon_annotation_button, note_pixmaps);

  // Enable dragging
  XtOverrideTranslations(icon_annotation_button, button_drag_translations);

  // Activating the button displays the annotation:
  register_cb(icon_annotation_button, XmNactivateCallback,
	      (CBType)&Annotation::popup_for_annotationCB, NULL);

  // Change the left attachment for the label to accomodate the button:
  XtVaSetValues(icon_label,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, icon_annotation_button,
		NULL);
}


// Remove an existing annotation text button

void Annotation::remove_annotation_button()
{
  components &= ~(OBJ_ANNOTATION); // update components

  // Reconfigure the label constrants and destroy the button widget:
  XtUnmanageChild(icon_form);
  XtVaSetValues(icon_label,
		XmNleftAttachment, XmATTACH_FORM,
		NULL);

  XtDestroyWidget(icon_annotation_button);
  XtManageChild(icon_form);
}


// Create a button for the hyperlink and add it to icon_form

void Annotation::create_hyperlink_button()
{
  components |= OBJ_HYPERLINK;	// update components

  // Set the labelPixmap
  icon_link_button =
    XtVaCreateWidget("L", xmPushButtonWidgetClass, icon_form,
		     XmNrightAttachment, XmATTACH_FORM,
		     XmNtopAttachment, XmATTACH_FORM,
		     XmNleftAttachment, XmATTACH_WIDGET,
		     XmNleftWidget, icon_label,
		     XmNmarginWidth, 0,
		     XmNmarginHeight, 0,
		     NULL);

  register_eh(icon_link_button, ButtonPressMask,
	      (EHType)&Annotation::raise_appropriate_menuEH, NULL, False); 

 // Set the labelPixmap
  assign_label_pixmap(icon_link_button, link_pixmaps);

  // Enable dragging
  XtOverrideTranslations(icon_link_button, button_drag_translations);

  // Activating the button moves the display to the linked annotation:
  register_cb(icon_link_button, XmNactivateCallback,
	      (CBType)&Annotation::follow_linkCB, NULL);

  // Change the right attachment for the label to accomodate the button:
  XtVaSetValues(icon_label,
		XmNrightAttachment, XmATTACH_NONE,
		NULL);
}


// Remove an existing annotation text button

void Annotation::remove_hyperlink_button()
{
  components &= ~(OBJ_HYPERLINK); // update components

  // Reconfigure the label constrants and destroy the button widget:
  XtUnmanageChild(icon_form);

  XtDestroyWidget(icon_link_button);
  XtManageChild(icon_form);
}


// Create and initialize the elements of the document annotation
// according to the value of components

void Annotation::create_appropriate_buttons()
{
  // Create/remove the annotation text button as necessary:
  if ((strlen(text) != 0) && !(components & OBJ_ANNOTATION))
    create_annotation_button();
  else if ((strlen(text) == 0) && (components & OBJ_ANNOTATION))
    remove_annotation_button();

  // Set the annotation name in the label:
  set_labelString(icon_label, name);

  if(strlen(name) != 0)
    components |= OBJ_LABEL;
  else
    components &= ~OBJ_LABEL;

  // Create/remove the hyperlink button as necessary:
  if ((link || (link_int != -1)) && !(components & OBJ_HYPERLINK))
    create_hyperlink_button();
  else if ((!link && (link_int == -1)) && (components & OBJ_HYPERLINK))
    remove_hyperlink_button();

  // Set the top/bottom attachment for the label:
  XtVaSetValues(icon_label,
		XmNtopAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_FORM,
		NULL);

  // If no components are in use, use a blank note icon for the label:
  if (!components)
    {
      XtVaSetValues(icon_label,
		    XmNlabelType, XmPIXMAP,
		    NULL);

      assign_label_pixmap(icon_label, marker_pixmaps);
    }
  else
    XtVaSetValues(icon_label,
		  XmNlabelType, XmSTRING,
		  NULL);

  // Manage the widgets:
  XtManageChild(icon_label);
  if (components & OBJ_ANNOTATION)
    XtManageChild(icon_annotation_button);
  if (components & OBJ_HYPERLINK)
    XtManageChild(icon_link_button);
}


// Convert link_int, backlinks_int from indices to annotation pointers
// and store the results in link, backlinks.  The argument specifies
// the collection of annotations to which the indices apply.

void Annotation::indices_to_objects(OrderedCltn &annotations)
{
  // Convert link_int to link:
  if ((link_int >= 0) && (link_int < annotations.size()))
    link = (Annotation*)(annotations[link_int]);
  else
    link = NULL;

  link_int = -1;

  // Convert the backlink indices to annotation pointers and add them
  // to backlinks:
  for(int i = 0; i < backlinks_int->size(); i++)
    {
      Integer *this_int = (Integer*)(backlinks_int->at(i));

      if ((this_int->value() >= 0) && (this_int->value() < annotations.size()))
	backlinks.add(*(Annotation*)(annotations[(int)(this_int->value())]));
      else
	cerr << "Warning: hyperlink index out of range (ignored)" << endl;

      delete this_int;		// delete the backlink_int contents
    }

  delete backlinks_int;
}


// Move the page display to the location of *link

void Annotation::follow_linkCB(Widget, XtPointer, XtPointer)
{
  // Change the page if necessary:

  if (parent->get_current_page() != link->page)
    {
      parent->draw_page(link->page);
      parent->update_scan(link->page);
    }

  // Pan to center the annotation in the window
  parent->jump_to_pos(link->xpos, link->ypos);
}


// Reorient this annotation, i.e. when the page orientation changes

void Annotation::set_orientation(const int new_orientation)
{
  // Rotate the coordinates to the new orientation:
  int orientation_difference = new_orientation - OS_PORTRAIT;

  xpos = xpos_virtual;
  ypos = ypos_virtual;

  while(orientation_difference > 0)
    {
      rotate_cl();
      orientation_difference--;
    }
}


void Annotation::raise_appropriate_menuEH(Widget widget, XtPointer,
					  XButtonPressedEvent *event, Boolean*)
{
  parent->popup_callbacks[CB_PROPERTIES] =
    (CBType)&Annotation::edit_annotationCB;
  parent->popup_callbacks[CB_TEXT] =
    (CBType)&Annotation::popup_for_annotationCB;
  parent->popup_callbacks[CB_HYPERLINK] =
    (CBType)&Annotation::follow_linkCB;

  parent->callback_root = this;

  Widget popup;

  if ((components & OBJ_HYPERLINK) && (components & OBJ_ANNOTATION))
    popup = parent->annotation_popup_both_buttons;
  else if (components & OBJ_HYPERLINK)
    popup = parent->annotation_popup_link_button;
  else if (components & OBJ_ANNOTATION)
    popup = parent->annotation_popup_text_button;
  else
    popup = parent->annotation_popup_no_buttons;

  if (event->button != 3) return;

  XmMenuPosition(popup, event);
  XtManageChild(popup);
}
