//  This may look like C, but it is really -*- C++ -*-

// This is file Annotation.h.  It describes a class for managing the
// document annotations and their data, implemented in Annotation.C.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/



#ifndef Annotation_H
#define Annotation_H


#include "UIObject.h"
#include "TopShell.h"
#include "ADialogs.h"
#include <nihcl/OrderedCltn.h>


class DocWindow;


// The three components of a document annotation:
#define OBJ_ANNOTATION  (1 << 0)
#define OBJ_LABEL  (1 << 1)
#define OBJ_HYPERLINK  (1 << 2)


class Annotation: public UIObject {
  DECLARE_REDUCED_MEMBERS(Annotation);

private:
  // Dialogs, if they are in use by this annotation.  Dialogs are
  // cached by the parent DocWindow object and are allocated to
  // individual Annotations only when necessary.  These pointers are
  // set to NULL when there is no dialog assigned to this annotation:
  FullAnnotationDialog *full_dialog;
  SmallAnnotationDialog *small_dialog;


  // xpos/ypos give the coordinates on the physical Ghostview widget;
  // xpos/ypos_virtual give the coordinates were the page rendered
  // with orientation "Portrait".  That is, xpos/ypos_virtual are
  // saved with the document, and do not change unless the annotation
  // is moved.  xpos/ypos change when the page orientation is changed.
  float xpos, ypos, xpos_virtual, ypos_virtual;

  char *text;			// the annotation text
  Annotation *oldlink;		// a pointer to the previous link
  DocWindow *parent;		// the DocWindow annotated by "this"
  CBType parent_deleteCB;	// delete the specified Annotation

  // Action addresses for drag, double click translations
  char *link_drag_addr, *edit_annotation_addr;

  // When annotations are loaded from a document file, the link and
  // backlinks are specified as integer indices into the list of
  // annotations.  When all annotations have been loaded, they are
  // converted to pointers (stored in link and backlinks).  These
  // members store the indices, which are needed only until the
  // conversion is performed:
  int link_int;			// index of link annotation
  OrderedCltn *backlinks_int;	// collection of backlink indices

  // Display the text annotation in a SmallAnnotationDialog
  void popup_for_annotationCB(Widget, XtPointer, XtPointer);
  // Process the results of the annotation dialog
  void create_annotationCB(Widget, XtPointer, XtPointer);
  // Delete this annotation
  void cancel_annotationCB(Widget, XtPointer, XtPointer);
  // Cancel editing changes and revert to saved values
  void cancel_changesCB(Widget, XtPointer, XtPointer);
  // Add this object as a backlink for *link
  void drop_finishCB(Widget, XtPointer, XmDropFinishCallbackStruct*);
  // Start a drag from the annotation dialog's link point
  void link_dragAP(Widget, XEvent *, char *[], int);
  // Move the page display to the location of *link
  void follow_linkCB(Widget, XtPointer, XtPointer);
  // Receive a drop in the annotation dialog's link point
  void drop_and_linkCB(Widget, int, XtPointer);
  // Create a dialog to edit the values of this annotation
  void edit_annotationAP(Widget, XEvent*, char*[], int);
  void edit_annotationCB(Widget, XtPointer, XtPointer);

  // Setup for creating a new annotation by dropping
  static Boolean Annotation::LinkConvertProc(Widget, Atom*, Atom*,
					     Atom*, XtPointer*, unsigned long*, int*,
					     unsigned long*, int, XtRequestId*);

  // Functions for manipulating the dialog cache.
  // get_full/small_dialog retrieves a free dialog, if one exists, and
  // otherwise creates a new one.  release_full/small_dialog returns
  // an unused dialog to the cache:
  FullAnnotationDialog *get_full_dialog();
  SmallAnnotationDialog *get_small_dialog();
  void release_full_dialog(FullAnnotationDialog *dialog)
    { parent->full_dialog_stack.push(*dialog); };
  void release_small_dialog(SmallAnnotationDialog *dialog)
    { parent->annotation_dialog_stack.push(*dialog); };

  // Create/remove buttons from the document annotation:
  void create_annotation_button();
  void create_hyperlink_button();
  void remove_annotation_button();
  void remove_hyperlink_button();

  // Create and initialize the elements of the document annotation
  // according to the value of components
  void create_appropriate_buttons();

  // The compiled translations for the annotation label (drag/edit)
  // and the annotation buttons (drag):
  XtTranslations label_drag_translations;
  XtTranslations button_drag_translations;

  // Raise the appropriate popup menu from the parent DocWindow
  void raise_appropriate_menuEH(Widget, XtPointer, XButtonPressedEvent*, Boolean*);

public:
  const int page;		// the page this annotation is on
  // Widgets for the annotation form, buttons, and label:
  Widget icon_form;
  Widget icon_annotation_button, icon_link_button, icon_label;
  char *name;			// the annotation name
  // A collection of annotations that link to this one
  OrderedCltn backlinks;
  Annotation *link;		// a link to another annotation
  int components;		// bitwise OR of OBJ_HYPERLINK, etc.


  // Create a new annotation either directly (first form) or from a
  // saved specification:
  Annotation(Widget, const int, const float, const float, const char *,
	     DocWindow *, CBType, char *);
  Annotation(Widget, const int, const float, const float, const char *,
	     DocWindow *, CBType, char *, char *, int, OrderedCltn *);
  ~Annotation();		// delete the annotation

  // Convert link/backlink indices to pointers
  void indices_to_objects(OrderedCltn&);
  // Move this annotation:
  void set_coordinates(const float new_x, const float new_y)
    { xpos = new_x; ypos = new_y; }
  void set_virtual_coordinates(const float new_x, const float new_y)
    { xpos_virtual = new_x; ypos_virtual = new_y; }
  // Reorient this annotation, i.e. when the page orientation changes
  void set_orientation(const int);

  // Access to private members:
  const char *get_text() const { return text; }
  const float get_xpos() const { return xpos; }
  const float get_ypos() const { return ypos; }
  const float get_xpos_virtual() const { return xpos_virtual; }
  const float get_ypos_virtual() const { return ypos_virtual; }

  void swap_dimensions();	// Exchange xpos/ypos
  void rotate_cl();		// xpos := 1 - ypos; ypos := xpos
  void rotate_ccl();		// ypos := 1 - xpos; xpos := ypos

  // Remove the link component from the backlinked annotations, i.e.
  // when this annotation is deleted
  void clear_backlinks();

public:
  bool operator==(const Annotation& a) const
    { return (this == &a); }
  bool operator!=(const Annotation& a) const
    { return !(*this==a); }
  virtual bool isEqual(const Object&) const;
#ifdef FULL_NIHCL
  virtual const Class* species() const
    { return &classDesc; }
#endif
};

#endif
