
// This is file DocWindow.C.  It implements a class for managing
// document-level windows as well as the PostScript file itself.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/



// 	$Id: DocWindow.C,v 2.3 1996/07/05 15:42:51 ginsburg Exp ginsburg $	


#include "DocWindow.h"
#include "TopShell.h"
#include "ScanControl.h"
#include "MessageBox.h"
#include "Thumbnail.h"
#include "Annotation.h"
#include "ExportDialog.h"

#include <nihcl/String.h>
#include <nihcl/OrderedCltn.h>
#include <nihcl/Integer.h>
#include <nihcl/Iterator.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#ifdef sgi
#include <libgen.h>
#endif

#define String X11String
#define Object X11Object

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/Shell.h>
#include <X11/Composite.h>
#include <X11/cursorfont.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/SeparatoG.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/DrawingA.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrolledW.h>
#include <Xm/ToggleB.h>
#include <Xm/DialogS.h>
#include <Xm/Frame.h>
#include <Xm/CascadeB.h>
#include <Xm/DragDrop.h>
#include <Xm/DragIcon.h>
#include <Xm/DropSMgr.h>

extern "C" {
#include "Ghostview.h"
#include "Generic.h"
}

#undef String
#undef Object

// Compatibility defines:

#ifdef sgi
// Select an appropriate temp directory
#define TEMP_DIR(FILENAME)  dirname(FILENAME)

// Signal-handling:
#define SIGNALS_INIT()
#define SIGNAL_SET_IGNORE(SIG)  sigset(SIG, SIG_IGN)
#define SIGNAL_SET_DEFAULT(SIG)  sigset(SIG, SIG_DFL)

#else // not sgi
// Select an appropriate temp directory
#define TEMP_DIR(FILENAME)  "/usr/tmp"

// basename may not exist
static inline char *basename(char *s) {return s;}

// Signal-handling:
#define SIGNALS_INIT()  struct sigaction action; action.sa_flags = SA_RESETHAND
#define SIGNAL_SET_IGNORE(SIG)  action.sa_handler = SIG_IGN; sigaction(SIG, &action, NULL)
#define SIGNAL_SET_DEFAULT(SIG)  action.sa_handler = SIG_DFL; sigaction(SIG, &action, NULL)

#endif // sgi


static char BROWSER_ANNOTATION_NAME[] = "%BrowserAnnotationName";
static char BROWSER_ANNOTATION_TEXT[] = "%BrowserAnnotationText";
static char BROWSER_ANNOTATION_END[] = "%BrowserAnnotationEnd";


#define	THIS	DocWindow
#define	BASE	UIObject
#define	BASE_CLASSES BASE::desc()
#define	MEMBER_CLASSES
#define VIRTUAL_BASE_CLASSES

DEFINE_REDUCED_CLASS(DocWindow,1,"$Header: /home/cellar/ginsburg/browser/RCS/DocWindow.C,v 2.3 1996/07/05 15:42:51 ginsburg Exp ginsburg $",NULL)


// A macro to set float arguments, as defined in the ghostview sources:

#define XtSetFloatArg(arg, n, d) \
    if (sizeof(float) > sizeof(XtArgVal)) { \
	XtSetArg(arg, n, &(d)); \
    } else { \
	XtArgVal *ld = (XtArgVal *)&(d); \
	XtSetArg(arg, n, *ld); \
    }


bool DocWindow::isEqual(const Object& p) const
// Test two objects for equality
{
  return &p == this;
}


//  Declare the EditRes event handler
extern "C" void _XEditResCheckMessages();


DocWindow::DocWindow(Widget parent_widget, FILE *file, const String &filename,
		     CBType parent_deleteCB, TopShell *parent_obj,
		     int &successful, pid_t &pid)
:psfile(file),
 psfilename(filename),
 scan(NULL),
 parent(parent_obj),
 size_established(False),
 icon_drag_addr(register_ap("icon_drag", (APType)&DocWindow::icon_dragAP)),
 annotations_modified(False),
 bookmarks_modified(False),
 allow_annotations(True),
 x_slider_size(0),
 auto_orientation(PORTRAIT),
 current_media(papersizes)
{
  // Initialize the bounding box to {0, 0, 0, 0}
  memset(current_bounding_box, 0, sizeof(int) * 4);

  // Get the initial modification time of the file:
  stat((const char *)filename, &file_stats);
  file_modification_time = file_stats.st_mtime;

  shell = XtCreatePopupShell("docwindow", topLevelShellWidgetClass,
			     parent_widget, NULL, 0);

  XtVaSetValues(shell,
		XmNdeleteResponse, XmDO_NOTHING,
		XmNallowShellResize, True,
		XmNiconName, "DeckView Document",
		XmNkeyboardFocusPolicy, XmPOINTER,
		NULL);

  register_wm_cb(shell, delete_window_atom, parent_deleteCB,
		 (XtPointer)this, (UIObject*)parent_obj);

  XtAddEventHandler(shell, (EventMask)0, True,
		    (XtEventHandler)_XEditResCheckMessages, NULL);

  main_window = XtVaCreateWidget("main_window", xmFormWidgetClass, shell,
				 NULL);

  // Load the document:
  doc = psscan(psfile);

  current_page = 0;

  // Alert the user if the document is non-conforming
  if ((!doc) || (doc->numpages == 0))
    {
      static char *labels[] = {"Use Ghostview", "Abort", "Help"};

      char *base = basename((char*)(const char *)psfilename);

      char error_msg[350];
      sprintf(error_msg, "File \"%s\" does not conform to the\nDocument Structuring Conventions, and cannot be used with DeckView.", base);

      MessageBox msg_box(shell, XmDIALOG_ERROR, error_msg, labels,
			 MBOkButton | MBCancelButton,
			 XmDIALOG_PRIMARY_APPLICATION_MODAL);

      int response = msg_box.get_response();

      if (response == MBOkButton)
	{
	  pid = fork();
	  if (pid == -1)
	    perror("Error: could not fork");

	  else if (pid == 0)
	    {
	      // child
	      execl(TopShell::appdata.gv_path,
		    "ghostview", (char*)(const char *)psfilename, NULL);

	      perror("DeckView: exec failed");
	      exit(1);
	    }
	  // Parent returns pid as a reference
	}
      else			// MBCancel ==> no Ghostview instance
	pid = -1;

      successful = False;
      return;
    }

  // Load the saved bookmarks:
  rewind(psfile);

  fseek(psfile, doc->endheader, SEEK_SET);
 
  char line[PSLINELENGTH], comment_start[PSLINELENGTH],
       page_name[PSLINELENGTH];
  int page_number;

  // Skip over the "%%BeginProlog" line

  OrderedCltn *old_bookmarks = new OrderedCltn();
  int num_read;

  while(True)
    {
      fgets(line, PSLINELENGTH, psfile);

      num_read =
	sscanf(line, "%%BrowserBookmark %d %s", &page_number, page_name);

      if ((num_read == EOF) || (num_read != 2))
	break;

      old_bookmarks->add(*(new Integer(page_number)));
    }

  // Create the page area widgets

  // Add actions and translations

  gvform = XtVaCreateManagedWidget("gvform", xmFormWidgetClass,
				   main_window,
				   XmNfractionBase, 100,
				   XmNleftAttachment, XmATTACH_FORM,
				   XmNrightAttachment, XmATTACH_FORM,
				   NULL);

  scan =
    new ScanControl(this, gvform, doc, old_bookmarks, (const char*)psfilename);

  int scan_offset;
  XtVaGetValues(scan->get_top_widget(),
		XmNrightOffset, &scan_offset,
		NULL);

  save_dialog =
    new ExportDialog(shell, this, TYPE_SAVE_DIALOG, (const char*)psfilename);
  print_dialog =
    new ExportDialog(shell, this, TYPE_PRINT_DIALOG,
		     TopShell::appdata.print_command);

  gvscrolled =
    XtVaCreateManagedWidget("gvscrolled", xmScrolledWindowWidgetClass, gvform,
			    XmNtopAttachment, XmATTACH_FORM,
			    XmNrightAttachment, XmATTACH_FORM,
			    XmNbottomAttachment, XmATTACH_FORM,
			    XmNleftAttachment, XmATTACH_FORM,
			    XmNleftOffset, -scan_offset,
			    NULL);

  gvhscroll = XtVaCreateManagedWidget("gvhscroll", xmScrollBarWidgetClass, gvscrolled,
				      XmNorientation, XmHORIZONTAL,
				      XmNheight, SCROLLBAR_THICKNESS,
				      NULL);

  gvvscroll = XtVaCreateManagedWidget("gvvscroll", xmScrollBarWidgetClass, gvscrolled,
				      XmNorientation, XmVERTICAL,
				      XmNwidth, SCROLLBAR_THICKNESS,
				      NULL);

  gvarea =
    XtVaCreateManagedWidget("gvarea", genericWidgetClass, gvscrolled, NULL);

  register_eh(gvarea, ButtonPressMask, (EHType)&DocWindow::pointer_clickEH,
	      NULL);
  register_eh(gvarea, ButtonReleaseMask, (EHType)&DocWindow::pointer_clickEH,
	      NULL);
  register_eh(gvarea, Button1MotionMask | PointerMotionHintMask,
	      (EHType)&DocWindow::pointer_motionEH, NULL);

  XmScrolledWindowSetAreas(gvscrolled, gvhscroll, gvvscroll, gvarea);

  clicked_in_gvarea = False;

  gc = XCreateGC(dpy, RootWindowOfScreen(scrn), 0, NULL);

  char *expose_hack_addr =
    register_ap("expose_hack", (APType)&DocWindow::expose_hackAP);
  char *gvkey_addr = register_ap("gvkey", (APType)&DocWindow::process_keyAP);

  static const char ghostview_translation_format[] =
    "<Message>:	message()\n\
     <Expose>: expose_hack(%s)\n\
     <Key>Left: gvkey(left, %s)\n\
     <Key>Right: gvkey(right, %s)\n\
     <Key>Up: gvkey(up, %s)\n\
     <Key>Down: gvkey(down, %s)\n\
     <Key>Prior: gvkey(pup, %s)\n\
     <Key>Next: gvkey(pdown, %s)\n\
     <Key>KP_Left: gvkey(left, %s)\n\
     <Key>KP_Right: gvkey(right, %s)\n\
     <Key>KP_Up: gvkey(up, %s)\n\
     <Key>KP_Down: gvkey(down, %s)\n\
     <Key>KP_Prior: gvkey(pup, %s)\n\
     <Key>KP_Next: gvkey(pdown, %s)\n";


  char *ghostview_translations =
    new char[strlen(ghostview_translation_format) + 11*16];

  sprintf(ghostview_translations, ghostview_translation_format,
	  expose_hack_addr,
	  gvkey_addr, gvkey_addr, gvkey_addr, gvkey_addr,
	  gvkey_addr, gvkey_addr, gvkey_addr, gvkey_addr,
	  gvkey_addr, gvkey_addr, gvkey_addr, gvkey_addr);

  delete [] expose_hack_addr;

  XtTranslations compiled_ghostview_translations =
    XtParseTranslationTable(ghostview_translations);

  ghostview =
    XtVaCreateManagedWidget("ghostview",ghostviewWidgetClass, gvarea,
			    XtNtranslations, compiled_ghostview_translations,
			    XmNbackground, WhitePixelOfScreen(scrn),
			    XmNforeground, BlackPixelOfScreen(scrn),
			    XtNinterpreter, TopShell::appdata.gs_path,
			    //  For Drag/Drop callbacks w/o client_data:
			    XmNuserData, this,
			    XtVaTypedArg, XtNcursor, XtRString,
			    "fleur", sizeof("fleur"),
			    NULL);

  register_eh(ghostview, ButtonPressMask,
	      (EHType)&DocWindow::pointer_clickEH, NULL);
  register_eh(ghostview, ButtonReleaseMask,
	      (EHType)&DocWindow::pointer_clickEH, NULL);
  register_eh(ghostview, Button1MotionMask | PointerMotionHintMask,
	      (EHType)&DocWindow::pointer_motionEH, NULL);
  register_eh(ghostview, KeyPressMask,
	      (EHType)&DocWindow::process_keyEH, NULL);

  XmProcessTraversal(ghostview, XmTRAVERSE_CURRENT);


  Arg args[10];
  int n = 0;

  Atom importList[3];

  importList[0] = XA_STRING;
  importList[1] = existing_annotation_atom;
  importList[2] = create_link_atom;

  // Set the docwin_context of XtWindow(ghostview) to this
  XSaveContext(dpy, XtWindow(ghostview), docwin_context, (XPointer)this);

  XtSetArg(args[n], XmNimportTargets, importList); n++;
  XtSetArg(args[n], XmNnumImportTargets, XtNumber(importList)); n++;
  XtSetArg(args[n], XmNanimationStyle, XmDRAG_UNDER_SHADOW_IN); n++;
  XtSetArg(args[n], XmNdropSiteOperations, XmDROP_COPY | XmDROP_MOVE); n++;
  XtSetArg(args[n], XmNdropProc, handle_dropCB); n++;

  XmDropSiteRegister(ghostview, args, n);

  register_cb(ghostview, XtNdestroyCallback, parent_deleteCB,
	      (XtPointer)this, (UIObject*)parent_obj);
  register_cb(ghostview, XtNoutputCallback,
	      (CBType)&DocWindow::write_messageCB, NULL);
  register_cb(ghostview, XtNmessageCallback,
	      (CBType)&DocWindow::handle_gs_messageCB, NULL);


  XtVaGetValues(ghostview,
		XtNxdpi, &full_page_xdpi,
		XtNydpi, &full_page_ydpi,
		NULL);

  // Load the annotations and create their buttons:
  // line contains the next line after the end of the bookmarks section
  float x_coord, y_coord;
  int new_page;

  char text_line[PSLINELENGTH], name[PSLINELENGTH];
  int link_addr;

  Position gv_x, gv_y;
  XtVaGetValues(ghostview,
		XmNx, &gv_x,
		XmNy, &gv_y,
		NULL);

  while(True)
    {
      if (sscanf(line, "%%BrowserAnnotation %d %f %f",
		 &new_page, &x_coord, &y_coord) != 3)
	break;

      OrderedCltn *backlinks = new OrderedCltn;

      fgets(line, PSLINELENGTH, psfile);

      if (strncmp(BROWSER_ANNOTATION_NAME, line,
		  (sizeof BROWSER_ANNOTATION_NAME) - 1) != 0)
	{
	  cerr << "PS format error (annotation name); load aborted" << endl;
	  successful = False;
	  return;
	}
      // Otherwise, the line begins with BROWSER_ANNOTATION_NAME
      if (line[(sizeof BROWSER_ANNOTATION_NAME) - 1] == ' ')
	{
	  strcpy(name, line + sizeof(BROWSER_ANNOTATION_NAME));
	  name[strlen(name) - 1] = '\0';  // Delete the trailing newline
	}
      else
	// Otherwise, the name is empty; make it the empty string
	name[0] = '\0';

      fgets(line, PSLINELENGTH, psfile);

      if (sscanf(line, "%%BrowserAnnotationLink %d", &link_addr) == 1)
	fgets(line, PSLINELENGTH, psfile);
      else
	link_addr = -1;

      int new_backlink;
      while(sscanf(line, "%%BrowserAnnotationBacklink %d", &new_backlink) == 1)
	{
	  Integer *new_int = new Integer(new_backlink);
	  backlinks->add(*new_int);
	  fgets(line, PSLINELENGTH, psfile);
	}

      String text_store;

      while (strncmp(BROWSER_ANNOTATION_TEXT, line,
		     (sizeof BROWSER_ANNOTATION_TEXT) - 1) == 0)
	{
	  if (line[(sizeof BROWSER_ANNOTATION_TEXT) - 1] == ' ')
	    {
	      strcpy(text_line, line + sizeof(BROWSER_ANNOTATION_TEXT));
	      text_store &= text_line;
	    }
	  else
	    // Otherwise, the name is empty; make it the empty string
	    text_store &= "\n";

	  fgets(line, PSLINELENGTH, psfile);
	}

      if (strncmp(line, BROWSER_ANNOTATION_END,
		  strlen(BROWSER_ANNOTATION_END)) != 0)
	{
	  cerr << "PS format error; load aborted" << endl;
	  cerr << "Line is " << line << endl;
	  successful = False;
	  return;
	}

      int *bbox = bounding_box_of_page(new_page);

      Widget icon_form =
	XtVaCreateWidget("annotation_form", xmFormWidgetClass, gvarea,
			 XmNx,
			 (Position)(x_coord * (float)(bbox[URX] - bbox[LLX])
				    / 72.0 * full_page_xdpi),
			 XmNy,
			 (Position)(y_coord * (float)(bbox[URY] - bbox[LLY])
				    / 72.0 * full_page_ydpi),
			 XmNorientation, XmHORIZONTAL,
			 XmNborderWidth, 1,
			 XmNborderColor, 1,
			 NULL);

      delete [] bbox;

      annotations.add(*(new Annotation(icon_form, new_page, x_coord, y_coord,
				       (const char *)text_store, this,
				       (CBType)&DocWindow::delete_annotationCB,
				       icon_drag_addr, name,
				       link_addr, backlinks)));

      if(new_page == current_page)
	XtManageChild(icon_form);

      fgets(line, PSLINELENGTH, psfile);
    }

  for(int i = 0; i < annotations.size(); i++)
    ((Annotation*)(annotations[i]))->indices_to_objects(annotations);

  // Configure the scrollbars:
  XtVaGetValues(gvarea,
		XmNwidth, &areawidth,
		XmNheight, &areaheight,
		NULL);

  XtVaSetValues(gvvscroll,
		XmNminimum, 0,
		XmNpageIncrement, areaheight,
		XmNincrement, 10,
		NULL);

  XtVaSetValues(gvhscroll,
		XmNminimum, 0,
		XmNpageIncrement, areawidth,
		XmNincrement, 10,
		NULL);

  x_value = y_value = 0;

  // Register the ScrollBar callbacks:
  register_cb(gvhscroll, XmNvalueChangedCallback,
	      (CBType)&DocWindow::scrollCB, (XtPointer)XmHORIZONTAL);
  register_cb(gvhscroll, XmNdragCallback,
	      (CBType)&DocWindow::scrollCB, (XtPointer)XmHORIZONTAL);
  register_cb(gvvscroll, XmNvalueChangedCallback,
	      (CBType)&DocWindow::scrollCB, (XtPointer)XmVERTICAL);
  register_cb(gvvscroll, XmNdragCallback,
	      (CBType)&DocWindow::scrollCB, (XtPointer)XmVERTICAL);


  // Register the DrawingArea callbacks:
  static const char gvarea_translation_format[] =
    "<Configure>: gvresize(%s)\n\
     <Key>Left: gvkey(left, %s)\n\
     <Key>Right: gvkey(right, %s)\n\
     <Key>Up: gvkey(up, %s)\n\
     <Key>Down: gvkey(down, %s)\n\
     <Key>Prior: gvkey(pup, %s)\n\
     <Key>Next: gvkey(pdown, %s)\n\
     <Key>KP_Left: gvkey(left, %s)\n\
     <Key>KP_Right: gvkey(right, %s)\n\
     <Key>KP_Up: gvkey(up, %s)\n\
     <Key>KP_Down: gvkey(down, %s)\n\
     <Key>KP_Prior: gvkey(pup, %s)\n\
     <Key>KP_Next: gvkey(pdown, %s)\n";

  char *gvresize_addr =
    register_ap("gvresize", (APType)&DocWindow::layout_page_geometryAP);

  // Registered above for ghostview widget translations
  char *gvarea_translations =
    new char[strlen(gvarea_translation_format) + 14*16];
  sprintf(gvarea_translations, gvarea_translation_format, gvresize_addr,
	  gvkey_addr, gvkey_addr, gvkey_addr, gvkey_addr,
	  gvkey_addr, gvkey_addr, gvkey_addr, gvkey_addr,
	  gvkey_addr, gvkey_addr, gvkey_addr, gvkey_addr);

  delete [] gvresize_addr;
  delete [] gvkey_addr;

  XtOverrideTranslations(gvarea,
			 XtParseTranslationTable(gvarea_translations));

  delete [] gvarea_translations;
  
  // Create the popup menus for the Annotation class:
  
  static MenuItem popup_both_buttons_items[]=
    {{"Properties...", &xmPushButtonGadgetClass, 'P', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB,
      (XtPointer)CB_PROPERTIES, NULL},
     {"Show Text...", &xmPushButtonGadgetClass, 'T', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB, (XtPointer)CB_TEXT, NULL},
     {"Follow Link", &xmPushButtonGadgetClass, 'L', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB,
      (XtPointer)CB_HYPERLINK, NULL},
     {"Sep", &xmSeparatorGadgetClass, 'e', NULL, NULL, NULL, NULL, NULL},
     {"Remove", &xmPushButtonGadgetClass, 'R', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB, (XtPointer)CB_DELETE, NULL},
     NULL};

  static MenuItem popup_no_buttons_items[]=
    {{"Properties...", &xmPushButtonGadgetClass, 'P', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB,
      (XtPointer)CB_PROPERTIES, NULL},
     {"Sep", &xmSeparatorGadgetClass, 'e', NULL, NULL, NULL, NULL, NULL},
     {"Remove", &xmPushButtonGadgetClass, 'R', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB, (XtPointer)CB_DELETE, NULL},
     NULL};

  static MenuItem popup_text_button_items[]=
    {{"Properties...", &xmPushButtonGadgetClass, 'P', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB,
      (XtPointer)CB_PROPERTIES, NULL},
     {"Show Text...", &xmPushButtonGadgetClass, 'T', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB, (XtPointer)CB_TEXT, NULL},
     {"Sep", &xmSeparatorGadgetClass, 'e', NULL, NULL, NULL, NULL, NULL},
     {"Remove", &xmPushButtonGadgetClass, 'R', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB, (XtPointer)CB_DELETE, NULL},
     NULL};

  static MenuItem popup_link_button_items[]=
    {{"Properties...", &xmPushButtonGadgetClass, 'P', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB,
      (XtPointer)CB_PROPERTIES, NULL},
     {"Follow Link", &xmPushButtonGadgetClass, 'L', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB,
      (XtPointer)CB_HYPERLINK, NULL},
     {"Sep", &xmSeparatorGadgetClass, 'e', NULL, NULL, NULL, NULL, NULL},
     {"Remove", &xmPushButtonGadgetClass, 'R', NULL, NULL,
      (CBType)&DocWindow::call_popup_callbackCB, (XtPointer)CB_DELETE, NULL},
     NULL};


  annotation_popup_no_buttons =
    BuildPulldownMenu(ghostview, "Annotation", 'A', popup_no_buttons_items,
		      XmMENU_POPUP);
  annotation_popup_both_buttons =
    BuildPulldownMenu(ghostview, "Annotation", 'A', popup_both_buttons_items,
		      XmMENU_POPUP);
  annotation_popup_text_button =
    BuildPulldownMenu(ghostview, "Annotation", 'A', popup_text_button_items,
		      XmMENU_POPUP);
  annotation_popup_link_button =
    BuildPulldownMenu(ghostview, "Annotation", 'A', popup_link_button_items,
		      XmMENU_POPUP);


  page_dialog = XmCreatePromptDialog(shell, "page_dialog", NULL, 0);

  XtVaSetValues(XtParent(page_dialog),
		XmNtitle, "Go to Page",
		NULL);

  XmString page_prompt = XmStringCreateSimple("Go to Page #:");
  XmString this_page = XmStringCreateSimple(doc->pages[0].label);

  XtVaSetValues(page_dialog,
		XmNselectionLabelString, page_prompt,
		XmNtextString, this_page,
		XmNautoUnmanage, False,
		NULL);

  XmStringFree(page_prompt);
  XmStringFree(this_page);

  XtUnmanageChild(XmSelectionBoxGetChild(page_dialog,
					 XmDIALOG_HELP_BUTTON));
  XtManageChild(XmSelectionBoxGetChild(page_dialog, XmDIALOG_APPLY_BUTTON));

  register_cb(page_dialog, XmNokCallback,
	      (CBType)&DocWindow::goto_pageCB, (XtPointer)True);
  register_cb(page_dialog, XmNapplyCallback,
	      (CBType)&DocWindow::goto_pageCB, (XtPointer)False);
  register_cb(page_dialog, XmNcancelCallback,
	      (CBType)&UIObject::unmanageCB, page_dialog);
  register_cb(XtParent(page_dialog), XmNpopdownCallback,
	      (CBType)&UIObject::unmanageCB, page_dialog);


  mag_dialog = XmCreatePromptDialog(shell, "mag_dialog", NULL, 0);

  XtVaSetValues(XtParent(mag_dialog),
		XmNtitle, "Document Zoom",
		NULL);

  XmString mag_prompt = XmStringCreateSimple("Zoom Percentage:");
  XmString one_hundred = XmStringCreateSimple("100%");
          
  XtVaSetValues(mag_dialog,
		XmNselectionLabelString, mag_prompt,
		XmNtextString, one_hundred,
		XmNautoUnmanage, False,
		NULL);

  XmStringFree(mag_prompt);
  XmStringFree(one_hundred);

  XtUnmanageChild(XmSelectionBoxGetChild(mag_dialog,
					 XmDIALOG_HELP_BUTTON));
  XtManageChild(XmSelectionBoxGetChild(mag_dialog, XmDIALOG_APPLY_BUTTON));

  register_cb(mag_dialog, XmNokCallback,
	      (CBType)&DocWindow::set_magnificationCB, (XtPointer)True);
  register_cb(mag_dialog, XmNapplyCallback,
	      (CBType)&DocWindow::set_magnificationCB, (XtPointer)False);
  register_cb(mag_dialog, XmNcancelCallback,
	      (CBType)&UIObject::unmanageCB, mag_dialog);
  register_cb(XtParent(mag_dialog), XmNpopdownCallback,
	      (CBType)&UIObject::unmanageCB, mag_dialog);

  // Create the menus:

  menubar = XmCreateMenuBar(main_window, "menubar", NULL, 0);

  XtVaSetValues(menubar,
		XmNleftAttachment, XmATTACH_FORM,
		XmNtopAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		NULL);

  static MenuItem file_items[]=
    {{"Get Info", &xmPushButtonWidgetClass, 'I', NULL, NULL,
      (CBType)&DocWindow::get_infoCB, NULL, NULL},
     {"Export Postscript...", &xmPushButtonWidgetClass, 'E', NULL, NULL,
      (CBType)&DocWindow::exportCB, NULL, NULL},
     {"Print...", &xmPushButtonWidgetClass, 'P', NULL, NULL,
      (CBType)&DocWindow::exportCB, NULL, NULL},
     {"Sep", &xmSeparatorWidgetClass, 'e', NULL, NULL, NULL, NULL, NULL},
     {"Close", &xmPushButtonWidgetClass, 'C', "Ctrl<Key>c", "Ctrl+C",
      NULL, NULL, NULL},
     NULL};

  // Avoid "sorry, not implemented" errors ("general initializer in
  // initializer list")
  file_items[1].cb_data = (XtPointer)save_dialog;
  file_items[2].cb_data = (XtPointer)print_dialog;

  filemenu = BuildPulldownMenu(menubar, "File", 'F', file_items);

  // A hack to force the menubar to the same background color as the
  // menu panes:
  unsigned int menu_background;
  XtVaGetValues(filemenu, XmNbackground, &menu_background, NULL);
  XtVaSetValues(menubar, XmNbackground, menu_background, NULL);

  register_cb(XtNameToWidget(shell,"*Close"),XmNactivateCallback,
	      parent_deleteCB, (XtPointer)this, (UIObject*)parent_obj);

  static MenuItem zoomitems[] = 
    {{"50%", &xmPushButtonWidgetClass, '5', NULL, NULL,
      (CBType)&DocWindow::set_magnification_directlyCB, (XtPointer)50, NULL},
     {"75%", &xmPushButtonWidgetClass, '7', NULL, NULL,
      (CBType)&DocWindow::set_magnification_directlyCB, (XtPointer)75, NULL},
     {"100%", &xmPushButtonWidgetClass, '1', NULL, NULL,
      (CBType)&DocWindow::set_magnification_directlyCB, (XtPointer)100, NULL},
     {"150%", &xmPushButtonWidgetClass, '0', NULL, NULL,
      (CBType)&DocWindow::set_magnification_directlyCB, (XtPointer)150, NULL},
     {"200%", &xmPushButtonWidgetClass, '2', NULL, NULL,
      (CBType)&DocWindow::set_magnification_directlyCB, (XtPointer)200, NULL},
     {"Fit width", &xmPushButtonWidgetClass, 'w', NULL, NULL,
      (CBType)&DocWindow::fit_widthCB, NULL, NULL},
     {"Fit length", &xmPushButtonWidgetClass, 'l', NULL, NULL,
      (CBType)&DocWindow::fit_heightCB, NULL, NULL},
     {"Fit to window", &xmPushButtonWidgetClass, 'F', NULL, NULL,
      (CBType)&DocWindow::fit_to_windowCB, NULL, NULL},
     {"Custom...", &xmPushButtonWidgetClass, 'C', NULL, NULL,
      NULL, NULL, NULL},
     NULL};

  static MenuItem orientationitems[] = 
    {{"Auto (Currently Portrait)", &xmToggleButtonWidgetClass, 'A',
      NULL, NULL, NULL, NULL, NULL},
     {"Portrait", &xmToggleButtonWidgetClass, 'P', NULL, NULL,
      NULL, NULL, NULL},
     {"Landscape", &xmToggleButtonWidgetClass, 'L', NULL, NULL,
      NULL, NULL, NULL},
     {"Inverted", &xmToggleButtonWidgetClass, 'I', NULL, NULL,
      NULL, NULL, NULL},
     {"Seascape", &xmToggleButtonWidgetClass, 'S', NULL, NULL,
      NULL, NULL, NULL},
     NULL};

  static MenuItem pageitems[] =
    {{"Next",&xmPushButtonWidgetClass,'N',"Ctrl<Key>n","Ctrl+N",
      (CBType)&DocWindow::change_pageCB,(XtPointer)+1, NULL},
     {"Previous",&xmPushButtonWidgetClass,'P',"Ctrl<Key>p","Ctrl+P",
      (CBType)&DocWindow::change_pageCB,(XtPointer)-1, NULL},
     {"Go to...",&xmPushButtonWidgetClass,'G',"Ctrl<Key>g","Ctrl+G",
      NULL,NULL, NULL},
     {"Redisplay",&xmPushButtonWidgetClass,'R',NULL,NULL,
      (CBType)&DocWindow::change_pageCB,(XtPointer)0, NULL},
     {"Sep",&xmSeparatorWidgetClass,'E',NULL,NULL,NULL,NULL, NULL},
     {"Zoom",&xmCascadeButtonWidgetClass,'Z',NULL,NULL,NULL,NULL, zoomitems},
     {"Orientation",&xmCascadeButtonWidgetClass,'O',NULL,NULL,NULL,NULL,
      orientationitems},
     NULL};

  static MenuItem viewitems[] =
    {{"Thumbnails",&xmToggleButtonWidgetClass,'T',NULL,NULL,NULL,NULL,NULL},
     {"Annotations",&xmToggleButtonWidgetClass,'A',NULL,NULL,NULL,NULL, NULL},
     NULL};

  // Avoid "sorry, not implemented...":
  zoomitems[7].cb = (CBType)&UIObject::manageCB;
  zoomitems[7].cb_data = mag_dialog;
  pageitems[2].cb = (CBType)&UIObject::manageCB;
  pageitems[2].cb_data = page_dialog;

  pagemenu = BuildPulldownMenu(menubar, "Page", 'P', pageitems);
  viewmenu = BuildPulldownMenu(menubar, "View", 'V', viewitems);

  Widget annotation_toggle = XtNameToWidget(shell, "*Annotations");
  Widget thumbnail_toggle = XtNameToWidget(shell, "*Thumbnails");

  XtVaSetValues(annotation_toggle,
		XmNindicatorType, XmN_OF_MANY,
		XmNindicatorSize, 15,
		XmNvisibleWhenOff, True,
		XmNset, True,
		XtVaTypedArg, XmNselectColor, XtRString,
		"yellow", sizeof "yellow",
		XmNfillOnSelect, True,
		NULL);

  XtVaSetValues(thumbnail_toggle,
		XmNindicatorType, XmN_OF_MANY,
		XmNindicatorSize, 15,
		XmNvisibleWhenOff, True,
		XmNset, True,
		XmNsensitive, False,
		XtVaTypedArg, XmNselectColor, XtRString,
		"yellow", sizeof "yellow",
		XmNfillOnSelect, True,
		NULL);

  register_cb(annotation_toggle, XmNvalueChangedCallback,
	      (CBType)&DocWindow::toggle_annotationsCB, NULL);

  register_cb(thumbnail_toggle, XmNvalueChangedCallback,
	      (CBType)&DocWindow::toggle_regionCB,
	      (XtPointer)REGION_THUMBNAILS);

  static char *orientation_labels[] = {"*Auto (Currently Portrait)",
				       "*Portrait", "*Landscape",
				       "*Inverted", "*Seascape"};

  for (int i = 0; i < XtNumber(orientation_labels); i++)
    {
      orientation_buttons[i] = XtNameToWidget(shell, orientation_labels[i]);

      XtVaSetValues(orientation_buttons[i],
		    XmNindicatorType, XmONE_OF_MANY,
		    XmNindicatorSize, 15,
		    XmNvisibleWhenOff, True,
		    XmNset, i ? False : True,
		    NULL);

      register_cb(orientation_buttons[i], XmNvalueChangedCallback,
		  (CBType)&DocWindow::set_orientationCB, (XtPointer)i);
    }

  active_orientation = OS_AUTO;
  physical_orientation = OS_PORTRAIT;

  XtManageChild(menubar);

  // Create the message area:

  msgform =
    XtVaCreateManagedWidget("msgform", xmFormWidgetClass, main_window,
			    XmNfractionBase, 100,
			    XmNleftAttachment, XmATTACH_FORM,
			    XmNbottomAttachment, XmATTACH_FORM,
			    XmNrightAttachment, XmATTACH_FORM,
			    XmNleftOffset, 5,
			    XmNrightOffset, 5,
			    XmNbottomOffset, 5,
			    NULL);

  XtVaSetValues(gvform,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, menubar,
		XmNtopOffset, 5,
		XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, msgform,
		XmNbottomOffset, 5,
		XmNleftOffset, 5,
		XmNrightOffset, 5,
		NULL);

  next_page_button =
    XtVaCreateManagedWidget("next_page_button", xmPushButtonWidgetClass,
			    msgform,
			    XmNtopAttachment, XmATTACH_FORM,
			    XmNrightAttachment, XmATTACH_FORM,
			    XtVaTypedArg, XmNlabelString, XtRString,
			    "Next Page",
			    sizeof("Next Page"),
			    NULL);

  prev_page_button =
    XtVaCreateManagedWidget("prev_page_button", xmPushButtonWidgetClass,
			    msgform,
			    XmNtopAttachment, XmATTACH_FORM,
			    XmNrightAttachment, XmATTACH_WIDGET,
			    XmNrightWidget, next_page_button,
			    XmNbottomAttachment, XmATTACH_OPPOSITE_WIDGET,
			    XmNbottomWidget, next_page_button,
			    XtVaTypedArg, XmNlabelString, XtRString,
			    "Previous Page",
			    sizeof("Previous Page"),
			    NULL);

  register_cb(next_page_button, XmNactivateCallback,
	      (CBType)&DocWindow::change_pageCB, (XtPointer)+1);

  register_cb(prev_page_button, XmNactivateCallback,
	      (CBType)&DocWindow::change_pageCB, (XtPointer)-1);

  Widget annotation_frame =
    XtVaCreateWidget("Annotation frame", xmFrameWidgetClass, msgform,
		     XmNtopAttachment, XmATTACH_FORM,
		     XmNrightAttachment, XmATTACH_WIDGET,
		     XmNrightWidget, prev_page_button,
		     XmNrightOffset, 15,
		     XmNbottomAttachment, XmATTACH_OPPOSITE_WIDGET,
		     XmNbottomWidget, next_page_button,
		     XmNtraversalOn, False,
		     XmNshadowType, XmSHADOW_IN,
		     XmNshadowThickness, 3,
		     NULL);

  Widget annotation_label =
    XtVaCreateManagedWidget("Annotation Label", xmLabelWidgetClass, msgform,
			    XmNtopAttachment, XmATTACH_FORM,
			    XmNrightAttachment, XmATTACH_WIDGET,
			    XmNrightWidget, annotation_frame,
			    XmNrightOffset, 10,
			    XmNbottomAttachment, XmATTACH_OPPOSITE_WIDGET,
			    XmNbottomWidget, next_page_button,
			    XtVaTypedArg, XmNlabelString, XtRString,
			    "Annotation:", sizeof("Annotation:"),
			    NULL);

  annotation_icon =
    XtVaCreateManagedWidget("New Annotation", xmPushButtonWidgetClass,
			    annotation_frame,
			    XmNtraversalOn, False,
			    XmNshadowThickness, 0,
			    XmNchildType, XmFRAME_WORKAREA_CHILD,
			    NULL);

  assign_label_pixmap(annotation_icon, annotation_icon_pixmaps);

  XtManageChild(annotation_frame);

  char *init_drag_addr =
    register_ap("init_drag", (APType)&DocWindow::init_dragAP);

  static const char drag_translation_format[] =
    "<Btn2Down>: init_drag(annotation,%s)\n\
     <Btn1Down>(1+):\n\
     <Btn1Up>(1+):\n";

  char *drag_translations = new char[strlen(drag_translation_format) + 1*16];

  sprintf(drag_translations, drag_translation_format, init_drag_addr);

  XtOverrideTranslations(annotation_icon,
			 XtParseTranslationTable(drag_translations));

  XtManageChild(main_window);
  XtRealizeWidget(shell);
  XtManageChild(shell);

  // Raise the annotation buttons by lowering ghostview:
  XLowerWindow(dpy, XtWindow(ghostview));

  scan->scan_realize();

  change_pageCB(shell, 0, NULL); // Draw the first page

  successful = True;
}


DocWindow::~DocWindow()
{
  // A good indicator of whether the object was successfully created:
  if (!scan)
    return;

  int bookmark_response = MBCancelButton;
  int annotation_response = MBCancelButton;

  static char *confirmation_names[] = {"Yes", "No", NULL};

  if (bookmarks_modified && annotations_modified)
    {
      MessageBox confirm_save(shell, XmDIALOG_QUESTION,
			      "Persistent bookmarks and annotations were modified; update document file?",
			      confirmation_names, MBOkButton | MBCancelButton,
			      XmDIALOG_PRIMARY_APPLICATION_MODAL);

      annotation_response = bookmark_response = confirm_save.get_response();
    }
  else if (bookmarks_modified)
    {
      MessageBox confirm_save(shell, XmDIALOG_QUESTION,
			      "Persistent bookmarks were modified; update document file?",
			      confirmation_names, MBOkButton | MBCancelButton,
			      XmDIALOG_PRIMARY_APPLICATION_MODAL);

      bookmark_response = confirm_save.get_response();
    }
  else if (annotations_modified)
    {
      MessageBox confirm_save(shell, XmDIALOG_QUESTION,
			      "Annotations were modified; update document file?",
			      confirmation_names, MBOkButton | MBCancelButton,
			      XmDIALOG_PRIMARY_APPLICATION_MODAL);

      annotation_response = confirm_save.get_response();
    }

  if ((bookmark_response == MBOkButton) ||
      (annotation_response == MBOkButton))
    {
      // Update the file with new bookmarks and annotations:

      char *name_copy = strdup((const char *)psfilename);
      char *temp_name = tempnam(TEMP_DIR(name_copy), "br");

      if (rename(psfilename, psfilename & ".bak") == -1)
	{
	  perror("Could not rename file");
	  fprintf(stderr, "Save aborted\n");
	  return;
	}

      FILE *thisfile = fopen(psfilename & ".bak", "r");
      FILE *newfile = fopen(psfilename, "w");

      if (!(newfile && thisfile))
	{
	  cerr << "Error: could not open output files." << endl;
	  return;
	}

      pscopy(thisfile, newfile, 0, doc->endheader);

      char saved_line[PSLINELENGTH];

      if (bookmark_response == MBOkButton)
	{
	  SortedCltn &bookmarks = scan->get_column(LEFT_COLUMN);

	  for(int i = 0; i < bookmarks.size(); i++)
	    {
	      fprintf(newfile, "%%BrowserBookmark ");
	      ((Thumbnail*)(bookmarks[i]))->print_page(newfile);
	    }

	  // Discard the old bookmarks:
	  char discard_line[PSLINELENGTH];
	  int num_read = 2;
	  int page_number;
	  char page_name[PSLINELENGTH];

	  while(num_read == 2)
	    {
	      fgets(discard_line, PSLINELENGTH, thisfile);
	      num_read = sscanf(discard_line, "%%BrowserBookmark %d %s",
				&page_number, page_name);
	    }
	  strcpy(saved_line, discard_line);
	}
      else
	{
	  // Skip over the old bookmarks:
	  char copy_line[PSLINELENGTH], page_name[PSLINELENGTH];
	  int page_number, num_read;

	  fgets(copy_line, PSLINELENGTH, thisfile);
	  num_read = sscanf(copy_line, "%%BrowserBookmark %d %s",
			    &page_number, page_name);
	  while(num_read == 2)
	    {
	      fputs(copy_line, newfile);
	      fgets(copy_line, PSLINELENGTH, thisfile);
	      num_read = sscanf(copy_line, "%%BrowserBookmark %d %s",
				&page_number, page_name);
	    }
	  strcpy(saved_line, copy_line);
	}

      if (annotation_response == MBOkButton)
	{
	  // Save the annotations:
	  for(int i = 0; i < annotations.size(); i++)
	    {
	      Annotation * const this_annotation =
		((Annotation*)(annotations[i]));

	      fprintf(newfile, "%%BrowserAnnotation %d %f %f\n",
		      this_annotation->page,
		      this_annotation->get_xpos_virtual(),
		      this_annotation->get_ypos_virtual());

	      fprintf(newfile, "%%BrowserAnnotationName %s\n",
		      this_annotation->name);

	      if (this_annotation->components & OBJ_HYPERLINK)
		fprintf(newfile, "%%BrowserAnnotationLink %d\n",
			annotations.indexOf(*(this_annotation->link)));

	      // print the backlinks:
	      for(int j = 0; j < this_annotation->backlinks.size(); j++)
		fprintf(newfile, "%%BrowserAnnotationBacklink %d\n",
			annotations.indexOf(*(Annotation*)
					    (this_annotation->backlinks[j])));

	      if (this_annotation->components & OBJ_ANNOTATION)
		{
		  char *text = (char*)this_annotation->get_text();

		  const int text_len = strlen(text);
		  if (text[text_len - 1] != '\n')
		    {
		      const char * const text_temp = text;
		      text = new char[text_len + 2];
		      strcpy(text, text_temp);
		      text[text_len] = '\n';
		      text[text_len + 1] = '\0';
		    }

		  char *this_nl = strchr(text, '\n'),
		       *this_token = (char*)text;

		  while(this_nl)
		    {
		      // Write a NULL in place of the newline:
		      *this_nl = '\0';

		      fprintf(newfile, "%%BrowserAnnotationText %s\n",
			      this_token);
	  
		      this_token = this_nl + 1;
		      this_nl = strchr(this_token, '\n');
		    }
		}

	      fprintf(newfile, "%%BrowserAnnotationEnd\n");
	    }

	  // Discard the old annotations:
	  char discard_text[PSLINELENGTH];
	  float discard_x, discard_y;
	  int discard_page;

	  while(sscanf(saved_line, "%%BrowserAnnotation %d %f %f",
		       &discard_page, &discard_x, &discard_y) == 3)
	    {
	      fgets(saved_line, PSLINELENGTH, thisfile);
	      while(strncmp(BROWSER_ANNOTATION_END, saved_line,
			    strlen(BROWSER_ANNOTATION_END)) != 0)
		fgets(saved_line, PSLINELENGTH, thisfile);
	      fgets(saved_line, PSLINELENGTH, thisfile);
	    }
	  fputs(saved_line, newfile);
	}
      else
	fputs(saved_line, newfile);

      // Copy the rest of the file:
      const long start_position = ftell(thisfile);
      fseek(thisfile, -1, SEEK_END);
      const long stop_position = ftell(thisfile);

      pscopy(thisfile, newfile, start_position, stop_position);

      if ((ferror(newfile) != 0) || (ferror(thisfile) != 0))
	cerr << "I/O error detected; file save aborted." << endl
	  << "Original file has been saved with \".bak\" extension."
	  << endl;

      fclose(newfile);
      fclose(thisfile);

      free(name_copy);
      free(temp_name);
    }

  // Delete the annotations
  // (must be done *after* saving annotations and links):
  for (int i = 0; i < annotations.size(); i++)
    delete (Annotation*)(annotations[i]);

  // Destroy window and deallocate structures:
  GhostviewDisableInterpreter(ghostview);
  delete scan;
  delete save_dialog;
  delete print_dialog;

  // shell can be destroyed only when child UI objects have been deleted
  XtDestroyWidget(shell);
  fclose(psfile);

  delete [] icon_drag_addr;

  if (doc)
    psfree(doc);
}


inline int DocWindow::virtual_to_real_page(int virtual_page) const
{
  if (doc->pageorder == DESCEND)
    virtual_page = (doc->numpages - 1) - virtual_page;
  return virtual_page;
}


void DocWindow::write_messageCB(Widget, XtPointer, XtPointer call_data)
{
  cerr << (char*)call_data << endl;
}


void DocWindow::write_message(const char *msg)
{
  cerr << msg << endl;
}


void DocWindow::handle_gs_messageCB(Widget, XtPointer, XtPointer call_data)
{
  String msg = (char*)call_data;

  if (msg == "Failed")
    write_message("PostScript intepreter failed.\n");
  else if (msg == "BadAlloc")
    write_message("Failed to allocate backing pixmap\n");
  else if (msg == "Refresh")
    change_pageCB(NULL,0,NULL);  // Resend the page
  else if (msg == "Page")
    {}				// Do nothing
  else if (msg == "Done")
    write_message("Interpreter exited\n");
  else
    {
      const String buf = "Unknown interpreter error: " & msg;
      write_message((const char *)buf);
    }
}


void DocWindow::update_scan(const int page)
{
  scan->change_page(page);
}


static inline Boolean bounding_box_nonzero(const int box[])
{
  return ((box[URX] - box[LLX] > 0) && (box[URY] - box[LLY]) > 0);
}


static inline Boolean bounding_boxes_equal(const int box1[], const int box2[])
{
  return ((box1[0] == box2[0]) && (box1[1] == box2[1]) &&
	  (box1[2] == box2[2]) && (box1[3] == box2[3]));
}


static inline void media_to_bounding_box(const documentmedia *media, int box[])
{
  box[LLX] = box[LLY] = 0;
  box[URX] = media->width;
  box[URY] = media->height;
}


int *DocWindow::bounding_box_of_page(const int page)
{
  const int len = sizeof(int) * 4;
  int *bounding_box = new int[4];

  if (bounding_box_nonzero(doc->pages[page].boundingbox))
    memcpy(bounding_box, doc->pages[page].boundingbox, len);
  else if (bounding_box_nonzero(doc->default_page_boundingbox))
    memcpy(bounding_box, doc->default_page_boundingbox, len);
  else if (bounding_box_nonzero(doc->boundingbox))
    memcpy(bounding_box, doc->boundingbox, len);
  else if (doc->pages[page].media)
    media_to_bounding_box(doc->pages[page].media, bounding_box);
  else if (doc->default_page_media)
    media_to_bounding_box(doc->default_page_media, bounding_box);
  else if (doc->media)
    media_to_bounding_box(doc->media, bounding_box);
  else
    media_to_bounding_box(papersizes, bounding_box);	// Letter-size

  return bounding_box;
}


documentmedia *DocWindow::get_page_media(const int page)
{
  if (doc->pages[page].media)
    return doc->pages[page].media;
  else if (doc->default_page_media)
    return doc->default_page_media;
  else if (doc->media)
    return doc->media;
  else
    return papersizes;	// Letter-size
}


int *DocWindow::get_page_boundingbox(const int page)
{
  if (bounding_box_nonzero(doc->pages[page].boundingbox))
    return doc->pages[page].boundingbox;
  else if (bounding_box_nonzero(doc->default_page_boundingbox))
    return doc->default_page_boundingbox;
  else if (bounding_box_nonzero(doc->boundingbox))
    return doc->boundingbox;
  else
    return NULL;
}


int DocWindow::get_page_orientation(const int page)
{
  // Set the automatic orientation
  if (doc->pages[page].orientation != NONE)
    return doc->pages[page].orientation;
  else if (doc->default_page_orientation != NONE)
    return doc->default_page_orientation;
  else if (doc->orientation != NONE)
    return doc->orientation;
  else
    return PORTRAIT;
}


void DocWindow::draw_page(const int page)
{
  current_page = page;

  Boolean need_geometry_update = False;

  documentmedia *new_media = get_page_media(page);

  int *bounding_box = get_page_boundingbox(page); // Could be NULL

  if ((bounding_box &&
       !bounding_boxes_equal(bounding_box, current_bounding_box))
      || (!bounding_box && (new_media != current_media)))
    {
      if (bounding_box)
	memcpy(current_bounding_box, bounding_box, sizeof(int) * 4);
      else
	media_to_bounding_box(new_media, current_bounding_box);

      XtVaSetValues(ghostview,
		    XtNllx, current_bounding_box[LLX],
		    XtNlly, current_bounding_box[LLY],
		    XtNurx, current_bounding_box[URX],
		    XtNury, current_bounding_box[URY],
		    NULL);

      full_page_width =
	(Dimension)((current_bounding_box[URX] - current_bounding_box[LLX])
		    / 72.0 * full_page_xdpi + 0.5);

      full_page_height =
	(Dimension)((current_bounding_box[URY] - current_bounding_box[LLY])
		    / 72.0 * full_page_ydpi + 0.5);

      XtVaGetValues(ghostview,
		    XmNwidth, &gvwidth,
		    XmNheight, &gvheight,
		    NULL);

      need_geometry_update = True;
    }

  current_media = new_media;

  // Sensitize/desensitize the next and previous page buttons as necessary:
  if (current_page == 0)
    {
      XtSetSensitive(XtNameToWidget(shell,"*Previous"),False);
      XtSetSensitive(prev_page_button, False);
    }
  else
    {
      XtSetSensitive(XtNameToWidget(shell,"*Previous"),True);
      XtSetSensitive(prev_page_button, True);
    }

  if (current_page == doc->numpages - 1)
    {
      XtSetSensitive(XtNameToWidget(shell,"*Next"),False);
      XtSetSensitive(next_page_button, False);
    }
  else
    {
      XtSetSensitive(XtNameToWidget(shell,"*Next"),True);
      XtSetSensitive(next_page_button, True);
    }

  // Set the window title:
  static char window_title[350];

  sprintf(window_title, "Document: %s (Page %s)", (const char*)psfilename,
	  doc->pages[current_page].label);

  XtVaSetValues(shell,
		XmNtitle, window_title,
		NULL);

  // Set the automatic orientation
  int new_auto_orientation = get_page_orientation(current_page);

  if (auto_orientation != new_auto_orientation)
    {
      auto_orientation = new_auto_orientation;

      // Set the menu label and callback accordingly:
      if (new_auto_orientation == PORTRAIT)
	set_labelString(orientation_buttons[OS_AUTO],
			"Auto (Currently Portrait)");
      else
	set_labelString(orientation_buttons[OS_AUTO],
			"Auto (Currently Landscape)");

      if (active_orientation == OS_AUTO)
	{
	  change_gv_orientation((new_auto_orientation == PORTRAIT)
				? OS_PORTRAIT : OS_LANDSCAPE);
	  need_geometry_update = True;
	}
    }

  // Force the page to the upper left corner:
  Position gv_x = 0, gv_y = 0;
  XtVaGetValues(ghostview,
		XmNx, gv_x,
		XmNy, gv_y,
		NULL);

/*  XtVaSetValues(ghostview,
		XmNx, 0,
		XmNy, 0,
		NULL);*/

  // Manage the annotation buttons for the current page:
  for(int i = 0; i < annotations.size(); i++)
    {
      Annotation * const this_annotation = ((Annotation*)(annotations[i]));

      if ((this_annotation->page == current_page) && allow_annotations)
	{
	  this_annotation->
	    set_orientation((active_orientation == OS_AUTO) ?
			    ((new_auto_orientation == PORTRAIT) ?
			     OS_PORTRAIT : OS_LANDSCAPE) :
			    active_orientation);

	  const int x_coord =
	    gv_x + (int)(this_annotation->get_xpos() * (float)gvwidth);
	  const int y_coord =
	    gv_y + (int)(this_annotation->get_ypos() * (float)gvheight);

	  XtVaSetValues(this_annotation->icon_form,
			XmNx, x_coord,
			XmNy, y_coord,
			NULL);

	  XtManageChild(this_annotation->icon_form);
	}
      else
	XtUnmanageChild(this_annotation->icon_form);
    }

  if (need_geometry_update)
    layout_page_geometryAP(NULL, NULL, NULL, NULL);

  // Reset the view to the upper-left:
/*  if (x_slider_size) // (but only after these values have been initialized)
    {
      XmScrollBarSetValues(gvhscroll, 0, x_slider_size,
			   x_inc, x_page_inc, True);
      XmScrollBarSetValues(gvvscroll, 0, y_slider_size,
			   y_inc, y_page_inc, True);
    }
*/
  // Send PS to Ghostview
  send_current_page_ps();
}


void DocWindow::send_current_page_ps()
{
  // Check that the modification time is unchanged:
  if ((stat((const char *)psfilename, &file_stats) == -1)
      || (file_modification_time != file_stats.st_mtime))
    {
      MessageBox
	alert_user(shell, XmDIALOG_QUESTION,
		   "File has changed; reload the document?",
		   NULL, MBOkButton | MBCancelButton,
		   XmDIALOG_PRIMARY_APPLICATION_MODAL);

      // This is legal since send_current_page_ps() is always called
      // at the tail of a function stack that begins with a callback;
      // thus, the data from "this" is never needed after this point:
      if (alert_user.get_response() == MBOkButton)
	{
	  // Don't allow this to be deleted twice:
	  XtRemoveAllCallbacks(ghostview, XtNdestroyCallback);

	  parent->delete_and_reload(this, psfilename);
	}

      return;
    }

  // Draw the page in the main window:
  GhostviewEnableInterpreter(ghostview);
  GhostviewSendPS(ghostview, psfile, doc->beginprolog,
		  doc->lenprolog, False);
  GhostviewSendPS(ghostview, psfile, doc->beginsetup,
		  doc->lensetup, False);

  const int real_page = virtual_to_real_page(current_page);

  GhostviewSendPS(ghostview, psfile, doc->pages[real_page].begin,
		  doc->pages[real_page].len, False);
}


void DocWindow::change_pageCB(Widget, XtPointer inc, XtPointer)
{
  // Redraw the scan control
  scan->change_page(current_page + (int)inc);

  draw_page(current_page + (int)inc);
}


void DocWindow::set_magnificationCB(Widget, XtPointer close_dialog,
				    XtPointer call_data)
{
  const XmSelectionBoxCallbackStruct *cbs =
    (XmSelectionBoxCallbackStruct*)call_data;

  char *magstring;
  if (!XmStringGetLtoR(cbs->value,XmSTRING_DEFAULT_CHARSET,&magstring))
    return;

  if (!*magstring)  // No magnification specified
    {
      XtFree(magstring);
      return;
    }

  int magnification = atoi(magstring);
  XtFree(magstring);

  if ((magnification < 25) || (magnification > 400))
    {
      MessageBox mag_error(shell, XmDIALOG_ERROR,
			   "Magnification out of range\n\
                            (Please specify a value in the range 25-400%)",
			   NULL, MBOkButton,
			   XmDIALOG_PRIMARY_APPLICATION_MODAL);

      int discard = mag_error.get_response();
      return;
    }

  if ((int)close_dialog)
    XtUnmanageChild(mag_dialog);

  set_magnification_directlyCB(NULL, magnification, NULL);
}


void DocWindow::set_magnification_directlyCB(Widget, int magnification,
					     XtPointer)
{
  const float mag_ratio = (float)magnification/(float)100;
  float magnified_xdpi = mag_ratio * full_page_xdpi;
  float magnified_ydpi = mag_ratio * full_page_ydpi;

  GhostviewDisableInterpreter(ghostview);

  Arg args[5];
  int num_args = 0;
  XtSetFloatArg(args[num_args], XtNxdpi, magnified_xdpi); num_args++;
  XtSetFloatArg(args[num_args], XtNydpi, magnified_ydpi); num_args++;
  XtSetValues(ghostview, args, num_args);

  gvwidth = (int)((float)full_page_width * mag_ratio);
  gvheight = (int)((float)full_page_height * mag_ratio);

  XtVaSetValues(ghostview,
		XmNwidth, gvwidth,
		XmNheight, gvheight,
		NULL);

  Position gv_x, gv_y;
  XtVaGetValues(ghostview,
		XmNx, &gv_x,
		XmNy, &gv_y,
		NULL);

   for(int i = 0; i < annotations.size(); i++)
     {
       Annotation * const p = (Annotation*)(annotations[i]);

       const int x_coord = gv_x + (int)(p->get_xpos() * (float)gvwidth);
       const int y_coord = gv_y + (int)(p->get_ypos() * (float)gvheight);

       XtVaSetValues(p->icon_form,
		     XmNx, x_coord,
		     XmNy, y_coord,
		     NULL);
     }

  GhostviewEnableInterpreter(ghostview);

  layout_page_geometryAP(NULL,NULL,NULL,NULL);

  // Redraw the page
  send_current_page_ps();
}


void DocWindow::layout_page_geometryAP(Widget, XConfigureEvent*,
				       char*[], int*)
{
  XtVaGetValues(gvarea,
		XmNwidth, &areawidth,
		XmNheight, &areaheight,
		NULL);

  XtVaGetValues(ghostview,
		XmNwidth, &gvwidth,
		XmNheight, &gvheight,
		NULL);

  // Horizontal ScrollBar:
  XtVaGetValues(gvhscroll,
		XmNvalue, &x_value,
		XmNsliderSize, &x_slider_size,
		XmNpageIncrement, &x_page_inc,
		XmNincrement, &x_inc,
		XmNminimum, &x_min,
		XmNmaximum, &x_max,
		NULL);

  x_value = max(0, min((int)(x_max - areawidth),
		       (int)(gvwidth * ((float)x_value / (float)x_max))));

  XtVaSetValues(gvhscroll,
		XmNmaximum, gvwidth,
		XmNsliderSize, min(areawidth, gvwidth),
		XmNpageIncrement, areawidth,
		XmNvalue, x_value,
		NULL);

  x_slider_size = min(areawidth, gvwidth);
  x_max = gvwidth;
  x_page_inc = areawidth;

  // Vertical ScrollBar:
  XtVaGetValues(gvvscroll,
		XmNvalue, &y_value,
		XmNsliderSize, &y_slider_size,
		XmNincrement, &y_inc,
		XmNpageIncrement, &y_page_inc,
		XmNminimum, &y_min,
		XmNmaximum, &y_max,
		NULL);

  y_value = max(0, min((int)(y_max - areaheight),
		       (int)(gvheight * ((float)y_value / (float)y_max))));

  XtVaSetValues(gvvscroll,
		XmNmaximum, gvheight,
		XmNsliderSize, min(areaheight, gvheight),
		XmNpageIncrement, areaheight,
		XmNvalue, y_value,
		NULL);

  y_slider_size = min(areaheight, gvheight);
  y_max = gvheight;
  y_page_inc = areaheight;

  // Ghostview widget:
  move_gv_and_annotations((areawidth <= gvwidth) ?
			  - x_value : (areawidth - gvwidth) / 2,
			  (areaheight <= gvheight) ?
			  - y_value : (areaheight - gvheight) / 2);

  if (XtIsRealized(gvarea))
    XClearWindow(dpy, XtWindow(gvarea));
}


void DocWindow::pointer_clickEH(Widget, XtPointer, XButtonEvent *event,
				Boolean*)
{
  if ((!XtIsRealized(gvarea)) || (event->button != Button1))
    return;

  if (event->type == ButtonPress)
    {
      orig_x_value = x_value;
      orig_y_value = y_value;

      button_down_x = event->x_root;
      button_down_y = event->y_root;

      clicked_in_gvarea = True;
    }
  else
    {
      if (!clicked_in_gvarea)
	return;

      clicked_in_gvarea = False;

      int x_offset = button_down_x - event->x_root;
      int y_offset = button_down_y - event->y_root;

      const float x_ratio = (float)x_offset/(float)gvwidth;
      const float y_ratio = (float)y_offset/(float)gvheight;

      const int delta_x = (int)(x_ratio * (float)(x_max - x_min));
      const int delta_y = (int)(y_ratio * (float)(y_max - y_min));

      const int final_x =
	max(min(orig_x_value + delta_x, x_max - x_slider_size), x_min);
      const int final_y =
	max(min(orig_y_value + delta_y, y_max - y_slider_size), y_min);

      XmScrollBarSetValues(gvhscroll, final_x, x_slider_size,
			   x_inc, x_page_inc, True);
      XmScrollBarSetValues(gvvscroll, final_y, y_slider_size,
			   y_inc, y_page_inc, True);
    }
}



void DocWindow::pointer_motionEH(Widget widget, XtPointer, XButtonEvent *,
				 Boolean*)
{
  if (!XtIsRealized(gvarea) || !clicked_in_gvarea)
    return;

  int root_x, root_y, pos_x, pos_y;
  Window root, child;
  unsigned int state;

  if (!XQueryPointer(dpy, XtWindow(widget), &root, &child, &root_x, &root_y,
		     &pos_x, &pos_y, &state))
    return;			// Pointer not in this screen

  int x_offset = button_down_x - root_x;
  int y_offset = button_down_y - root_y;

  const float x_ratio = (float)x_offset/(float)gvwidth;
  const float y_ratio = (float)y_offset/(float)gvheight;

  const int delta_x = (int)(x_ratio * (float)(x_max - x_min));
  const int delta_y = (int)(y_ratio * (float)(y_max - y_min));

  const int final_x =
    max(min(orig_x_value + delta_x, x_max - x_slider_size), x_min);
  const int final_y =
    max(min(orig_y_value + delta_y, y_max - y_slider_size), y_min);

  XmScrollBarSetValues(gvhscroll, final_x, x_slider_size,
		       x_inc, x_page_inc, True);
  XmScrollBarSetValues(gvvscroll, final_y, y_slider_size,
		       y_inc, y_page_inc, True);
}


int DocWindow::page_label_to_virtual_page(const String &label) const
{
  if (!doc) return -1;

  for(int count=0; count < doc->numpages; count++)
    if (names_approx_equal(label, doc->pages[count].label)) return count;

  // Page not found
  return -1;
}



void DocWindow::goto_pageCB(Widget, XtPointer close_dialog,
			    XtPointer call_data)
{
  const XmSelectionBoxCallbackStruct *cbs =
    (XmSelectionBoxCallbackStruct*)call_data;

  char *pagestring;
  if (!XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &pagestring))
    return;

  if (!*pagestring)  // No page specified
    {
      XtFree(pagestring);
      return;
    }

  const int virtual_page = page_label_to_virtual_page(pagestring);

  if (virtual_page == -1)
    {
      char error_msg[100];

      sprintf(error_msg,
	      "Page %s does not exist.\n(Pages are in the range %s - %s.)",
	      pagestring, doc->pages[0].label,
	      doc->pages[doc->numpages - 1].label);

      MessageBox page_error(shell, XmDIALOG_ERROR,
			    error_msg, NULL, MBOkButton,
			    XmDIALOG_PRIMARY_APPLICATION_MODAL);

      int discard = page_error.get_response();
      return;
    }

  XtFree(pagestring);
  if ((int)close_dialog)
    XtUnmanageChild(mag_dialog);

  // Redraw the page
  const int page_increment =
    virtual_to_real_page(virtual_page) - current_page;
  change_pageCB(NULL, (XtPointer)page_increment, NULL);
}


int DocWindow::names_approx_equal(const char *new_str1, const char *new_str2)
{
  char *str1 = strdup(new_str1);
  char *str2 = strdup(new_str2);

  char *tok = strtok(str1, IGNORE_THESE_CHARS);

  char *cmp_str1 = new char[strlen(str1) + 1];
  *cmp_str1 = '\0';

  while(tok)
    {
      strcat(cmp_str1, tok);
      tok = strtok(NULL, IGNORE_THESE_CHARS);
    }


  tok = strtok(str2, IGNORE_THESE_CHARS);

  char *cmp_str2 = new char[strlen(str2) + 1];
  *cmp_str2 = '\0';

  while(tok)
    {
      strcat(cmp_str2, tok);
      tok = strtok(NULL, IGNORE_THESE_CHARS);
    }

  int return_val = !(strcmp(cmp_str1, cmp_str2));

  delete [] cmp_str1;
  delete [] cmp_str2;

  free(str1);
  free(str2);

  return return_val;
}


void DocWindow::scrollCB(Widget, int client_data,
			 XmScrollBarCallbackStruct *cbs)
{
  Position x, y;

  XtVaGetValues(ghostview,
		XmNx, &x,
		XmNy, &y,
		NULL);

  if (client_data == XmHORIZONTAL)
    {
      x_value = cbs->value;

      move_gv_and_annotations((areawidth <= gvwidth) ? - cbs->value : 0, y);
    }
  else
    {
      y_value = cbs->value;

      move_gv_and_annotations(x, (areaheight <= gvheight) ? - cbs->value : 0);
    }
}


void DocWindow::expose_hackAP(Widget, XEvent*, char*[], int*)
{
  if (!size_established)
    {
      size_established = True;
      layout_page_geometryAP(NULL,NULL,NULL,NULL);
    }
}


void DocWindow::fit_to_windowCB(Widget widget, XtPointer, XtPointer)
{
  const int multiplier =
    min((int)(100.0 * (float)areawidth / (float)full_page_width),
	(int)(100.0 * (float)areaheight / (float)full_page_height));

  set_magnification_directlyCB(widget, multiplier, NULL);
}


void DocWindow::fit_widthCB(Widget widget, XtPointer, XtPointer)
{
  const int multiplier =
    (int)(100.0 * (float)areawidth / (float)full_page_width);

  set_magnification_directlyCB(widget, multiplier, NULL);
}


void DocWindow::fit_heightCB(Widget widget, XtPointer, XtPointer)
{
  const int multiplier =
    (int)(100.0 * (float)areaheight / (float)full_page_height);

  set_magnification_directlyCB(widget, multiplier, NULL);
}


void DocWindow::set_orientationCB(Widget, int new_orientation, XtPointer)
{
  // Reorient the thumbnails
  if (scan->reorient(new_orientation) == REORIENT_ERROR)
    {
      // Thumbnails could not be reoriented, presumably because an
      // existing rendering was still in progress; restore the toggle
      // state and return:
      XmToggleButtonSetState(orientation_buttons[active_orientation],
			     True, False);
      XmToggleButtonSetState(orientation_buttons[new_orientation],
			     False, False);

      return;
    }

  // Enforce mutually exclusive orientation choice
  XmToggleButtonSetState(orientation_buttons[active_orientation],
			 False, False);
  XmToggleButtonSetState(orientation_buttons[new_orientation], True, False);

  // Convert auto_orientation from PORTRAIT/LANDSCAPE to
  // OS_PORTRAIT/OS_LANDSCAPE
  const int os_auto_orientation =
    (auto_orientation == PORTRAIT) ? OS_PORTRAIT : OS_LANDSCAPE;

  // new_orientation with OS_AUTO processed out:
  const int new_physical_orientation =
    (new_orientation == OS_AUTO) ? os_auto_orientation : new_orientation;

  change_gv_orientation(new_physical_orientation);

  active_orientation = new_orientation;

  Position gv_x, gv_y;

  XtVaGetValues(ghostview,
		XmNwidth, &gvwidth,
		XmNheight, &gvheight,
		XmNx, &gv_x,
		XmNy, &gv_y,
		NULL);

  for(int i = 0; i < annotations.size(); i++)
    {
      Annotation * const this_annotation = ((Annotation*)(annotations[i]));

      this_annotation->set_orientation(new_physical_orientation);

      const int x_coord =
	gv_x + (int)(this_annotation->get_xpos() * (float)gvwidth);
      const int y_coord =
	gv_y + (int)(this_annotation->get_ypos() * (float)gvheight);

      XtVaSetValues(this_annotation->icon_form,
		    XmNx, x_coord,
		    XmNy, y_coord,
		    NULL);
    }

  layout_page_geometryAP(NULL, NULL, NULL, NULL);

  XtVaSetValues(gvhscroll,
		XmNvalue, 0,
		NULL);
  XtVaSetValues(gvvscroll,
		XmNvalue, 0,
		NULL);

  send_current_page_ps();
}



void DocWindow::change_gv_orientation(const int new_orientation)
{
  const int orientation_difference = new_orientation - physical_orientation;

  physical_orientation = new_orientation;

  if (abs(orientation_difference % 2) == 1)
    {
      // Dimensions rotated; swap full_page_width and full_page_height:
      swap(full_page_width, full_page_height);
      swap(full_page_xdpi, full_page_ydpi);
    }

  // Convert the orientation from OS_... to degrees for Ghostview
  int gv_orientation = os_to_gv_orientation(new_orientation);

  // Set the scroll bar values to 0 (other values will be reset later):
  XtVaSetValues(gvhscroll,
		XmNvalue, 0,
		NULL);
  XtVaSetValues(gvvscroll,
		XmNvalue, 0,
		NULL);

  // Change ghostview's orientation and record the new geometry
  GhostviewDisableInterpreter(ghostview);
  XtVaSetValues(ghostview,
		XtNorientation, gv_orientation,
		NULL);

  // Get the new values for gvwidth/height
  XtVaGetValues(ghostview,
		XmNwidth, &gvwidth,
		XmNheight, &gvheight,
		NULL);
}


void DocWindow::init_dragAP(Widget widget, XEvent *event,
			    char *[], int num_args)
{
  if (num_args != 1)
    {
      cerr << "init_drag: invalid argument list" << endl;
      return;
    }

  if (!allow_annotations)
    return;

  static Atom exportList[] = {XA_STRING};

  Widget drag_icon =
    TopShell::create_drag_icon(widget, annotation_icon_pixmaps[PIXMAPS_IMAGE],
			       annotation_width, annotation_height);

  static Arg args[10];
  int n = 0;

  static XtCallbackRec dragDropFinishCB[] =
    {
      {(XtCallbackProc)destroyWidgetCB, NULL},
      {NULL, NULL}
    };

  dragDropFinishCB[0].closure = (XtPointer)drag_icon;

  XtSetArg(args[n], XmNexportTargets, exportList); n++;
  XtSetArg(args[n], XmNnumExportTargets, 1); n++;
  XtSetArg(args[n], XmNdragOperations, XmDROP_COPY); n++;
  XtSetArg(args[n], XmNdragDropFinishCallback, dragDropFinishCB); n++;
  XtSetArg(args[n], XmNconvertProc, DragConvertProc); n++;
  XtSetArg(args[n], XmNclientData, (XtPointer)this); n++;
  XtSetArg(args[n], XmNsourcePixmapIcon, drag_icon); n++;

  Widget drag_context = XmDragStart(widget, event, args, n);

  // Set the docwin_context of XtWindow(ghostview) to this
  XSaveContext(dpy, XtWindow(drag_context), docwin_context, (XPointer)this);
}


Boolean DocWindow::DragConvertProc(Widget, Atom *, Atom *target,
				   Atom *typeRtn, XtPointer *valueRtn,
				   unsigned long *lengthRtn, int *formatRtn,
				   unsigned long *, char *, XtRequestId *)
{
  if (*target != XA_STRING)
    return(False);

  *typeRtn = XA_STRING;
  *valueRtn = (XtPointer) strdup("annotation");
  *lengthRtn = strlen((char *)*valueRtn) + 1;
  *formatRtn = 8;

  return(True);
}


void DocWindow::handle_dropCB(Widget widget, XtPointer,
			      XmDropProcCallbackStruct *cbs)
{
  XmDropTransferEntryRec transferEntries[3];
  Arg args[10];
  int n = 0;

  if ((cbs->dropAction != XmDROP) || (cbs->operation == XmDROP_LINK))
    {
      XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE); n++;
    }
  else
    {
      // Retrieve the object pointer from the DragContext's userData:
      DocWindow *this_obj = NULL;
      Widget dc = XmGetDragContext(widget, cbs->timeStamp);

      if (dc &&
	  (XFindContext(dpy, XtWindow(dc), docwin_context,
			(XPointer*)&this_obj) == 0) &&
	  this_obj)
	{
	  transferEntries[0].target = XA_STRING;
	  transferEntries[0].client_data = (XtPointer)this_obj;
	  transferEntries[1].target = existing_annotation_atom;
	  transferEntries[1].client_data = (XtPointer)this_obj;
	  transferEntries[2].target = create_link_atom;
	  transferEntries[2].client_data = (XtPointer)this_obj;
	  XtSetArg(args[n], XmNdropTransfers, transferEntries); n++;
	  XtSetArg(args[n], XmNnumDropTransfers, 3); n++;
	  XtSetArg(args[n], XmNtransferProc, TransferProc); n++;
	}
      else
	{
	  // No context; probably a drag from somewhere else
	  XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE); n++;
	}
    }
  XmDropTransferStart(cbs->dragContext, args, n);
}


// Warning: format might in fact be int; documentation is sketchy
void DocWindow::TransferProc(Widget, XtPointer closure,
			     Atom *, Atom *type, XtPointer value,
			     unsigned long *, int*)
{
  DocWindow * const this_obj = (DocWindow*)closure;

  this_obj->annotations_modified = True;

  if (((*type == XA_STRING) && (0 == strcmp("annotation", (char*)value))) ||
      (*type == create_link_atom))
    {
      Window root_win, child_win;
      int root_x, root_y, win_x, win_y;
      unsigned int mask_return;

      XQueryPointer(dpy, XtWindow(this_obj->ghostview),&root_win, &child_win,
		    &root_x, &root_y, &win_x, &win_y, &mask_return);

      int parent_x, parent_y;
      Window child_return;

      XTranslateCoordinates(dpy, XtWindow(this_obj->ghostview),
			    XtWindow(XtParent(this_obj->ghostview)),
			    win_x, win_y, &parent_x,
			    &parent_y, &child_return);

      Widget annotation_form =
 	XtVaCreateWidget("annotation_form", xmFormWidgetClass,
			 this_obj->gvarea,
 			 XmNx, (Position)parent_x,
 			 XmNy, (Position)parent_y,
 			 NULL);

      // Create and store a new annotation object:
      float x_coord = (float)win_x / (float)(this_obj->full_page_width);
      float y_coord = (float)win_y / (float)(this_obj->full_page_height);

      convert_to_virtual_coords(x_coord, y_coord,
				this_obj->physical_orientation);

      Annotation *new_annotation =
	new Annotation(annotation_form, this_obj->current_page,
		       x_coord, y_coord,
		       "", this_obj, (CBType)&DocWindow::delete_annotationCB,
		       this_obj->icon_drag_addr);

      new_annotation->set_orientation(this_obj->physical_orientation);

      this_obj->annotations.add(*new_annotation);

      // If the type is create_link_atom, the new annotation should be
      // stored in the supplied location:
      if (*type == create_link_atom)
	*(((Annotation***)value)[0]) = new_annotation;

      XtManageChild(annotation_form);
    }
  else if (*type == existing_annotation_atom)
    {
      int selected_icon = ((int*)value)[0];

      Window root_win, child_win;
      int root_x, root_y, win_x, win_y;
      unsigned int mask_return;

      XQueryPointer(dpy, XtWindow(this_obj->ghostview), &root_win, &child_win,
		    &root_x, &root_y, &win_x, &win_y, &mask_return);

      int parent_x, parent_y;
      Window child_return;

      XTranslateCoordinates(dpy, XtWindow(this_obj->ghostview),
			    XtWindow(XtParent(this_obj->ghostview)),
			    win_x, win_y, &parent_x, &parent_y,
			    &child_return);

      // Find the corresponding annotation:
      Annotation *this_annotation =
	(Annotation*)(this_obj->annotations[selected_icon]);

      // Move the button
      XtVaSetValues(this_annotation->icon_form,
		    XmNx, parent_x,
		    XmNy, parent_y,
		    NULL);

      float x_coord = (float)win_x / (float)(this_obj->gvwidth);
      float y_coord = (float)win_y / (float)(this_obj->gvheight);

      this_annotation->set_coordinates(x_coord, y_coord);
      convert_to_virtual_coords(x_coord, y_coord,
				this_obj->physical_orientation);
      this_annotation->set_virtual_coordinates(x_coord, y_coord);
    }
}


void DocWindow::move_gv_and_annotations(Position x, Position y)
{
  // Get the current position for offsets
  Position gv_x, gv_y;
  XtVaGetValues(ghostview,
		XmNx, &gv_x,
		XmNy, &gv_y,
		NULL);

  Position x_diff = x - gv_x;
  Position y_diff = y - gv_y;

  XtVaSetValues(ghostview,
		XmNx, x,
		XmNy, y,
		NULL);

  for(int i = 0; i < annotations.size(); i++)
    {
      const Widget this_icon = ((Annotation*)(annotations[i]))->icon_form;

      XtVaGetValues(this_icon,
		    XmNx, &gv_x,
		    XmNy, &gv_y,
		    NULL);

      XtVaSetValues(this_icon,
		    XmNx, gv_x + x_diff,
		    XmNy, gv_y + y_diff,
		    NULL);
    }
}



void DocWindow::icon_dragAP(Widget widget, XEvent *event, char *[],
			    int num_args)
{
  if (num_args)
    {
      cerr << "icon_drag: invalid argument list" << endl;
      return;
    }

  // Store this pointer in widget's userData, then pass widget to convertProc
  XtVaSetValues(widget,
		XmNuserData, this,
		NULL);

  // Start the drag:
  Atom exportList[1];

  exportList[0] = existing_annotation_atom;

  // Find the corresponding annotation:
  int annotation_index = -1;
  for (int i = 0; i < annotations.size(); i++)
    {
      Annotation * const this_annotation = ((Annotation*)(annotations[i]));

      if ((this_annotation->icon_label == widget) ||
	  (this_annotation->icon_annotation_button == widget) ||
	  (this_annotation->icon_link_button == widget))
      {
	annotation_index = i;
	break;
      }
    }

  if (annotation_index == -1)
    {
      cerr << "Internal error: annotation not found." << endl;
      return;
    }

  // Create the drag icon:

  Widget annotation_form =
    ((Annotation*)(annotations[annotation_index]))->icon_form;

  Dimension icon_width, icon_height;
  unsigned int bg_color;

  XtVaGetValues(annotation_form,
		XmNwidth, &icon_width,
		XmNheight, &icon_height,
		XmNbackground, &bg_color,
		NULL);

  Pixmap annotation_image = XCreatePixmap(dpy, DefaultRootWindow(dpy),
					  icon_width, icon_height,
					  DefaultDepthOfScreen(scrn));

  GC gc = XCreateGC(dpy, RootWindowOfScreen(scrn), 0, NULL);

  XSetForeground(dpy, gc, bg_color);
  XFillRectangle(dpy, annotation_image, gc,
		 0, 0, icon_width, icon_height);

  XSetForeground(dpy, gc, BlackPixelOfScreen(scrn));
  XDrawRectangle(dpy, annotation_image, gc,
		 0, 0, icon_width - 1, icon_height - 1);

  XFreeGC(dpy, gc);

  Widget drag_icon =
    TopShell::create_drag_icon(widget, annotation_image,
			       icon_width, icon_height);

  int *indices = new int[1];
  indices[0] = annotation_index;

  static Arg args[10];
  int n = 0;

  static XtCallbackRec dragDropFinishCB[] =
    {
      {(XtCallbackProc)destroyWidgetCB, NULL},
      // Free pixmap as well
      {NULL, NULL}
    };

  dragDropFinishCB[0].closure = (XtPointer)drag_icon;

  XtSetArg(args[n], XmNexportTargets, exportList); n++;
  XtSetArg(args[n], XmNnumExportTargets, XtNumber(exportList)); n++;
  XtSetArg(args[n], XmNdragOperations, XmDROP_MOVE | XmDROP_LINK); n++;
  XtSetArg(args[n], XmNdragDropFinishCallback, dragDropFinishCB); n++;
  XtSetArg(args[n], XmNconvertProc, ButtonConvertProc); n++;
  XtSetArg(args[n], XmNclientData, (XtPointer)indices); n++;
  XtSetArg(args[n], XmNsourcePixmapIcon, drag_icon); n++;

  Widget drag_context = XmDragStart(widget, event, args, n);

  // Set the docwin_context of XtWindow(drag_context) to this for
  // retrieval by drop routines
  XSaveContext(dpy, XtWindow(drag_context), docwin_context, (XPointer)this);
}


Boolean DocWindow::ButtonConvertProc(Widget dc, Atom *, Atom *target,
				     Atom *typeRtn, XtPointer *valueRtn,
				     unsigned long *lengthRtn, int *formatRtn,
				     unsigned long *, int, XtRequestId *)
{
  if (*target != existing_annotation_atom)
    return(False);

  int *indices;
  XtVaGetValues(dc,
		XmNclientData, &indices,
		NULL);

  *typeRtn = existing_annotation_atom;
  *valueRtn = (XtPointer)indices;
  *lengthRtn = 1;
  *formatRtn = 8 * sizeof(int);

  return(True);
}


void DocWindow::add_drag_translations(Widget button)
{
  static const char button_translation_format[] =
    "<Btn2Down>: icon_drag(%s)\n";

  char *button_translations =
    new char[strlen(button_translation_format) + 1*16];
  sprintf(button_translations, button_translation_format, icon_drag_addr);

  XtOverrideTranslations(button,
			 XtParseTranslationTable(button_translations));

  delete [] button_translations;
}


void DocWindow::delete_annotationCB(Widget, Annotation *annotation, XtPointer)
{
  annotations_modified = True;

  annotations.removeId(*annotation);

  XtDestroyWidget(annotation->icon_form);
  delete annotation;
}


void DocWindow::exportCB(Widget, ExportDialog *dialog, XtPointer)
{
  int available = ENABLE_CENTER_COLUMN;

  if (scan->get_column_size(LEFT_COLUMN) > 0)
    available |= ENABLE_LEFT_COLUMN;
  if (scan->get_column_size(RIGHT_COLUMN) > 0)
    available |= ENABLE_RIGHT_COLUMN;

  dialog->set_available(available);

  dialog->popup_and_manage();
}



void DocWindow::process_keyAP(Widget, XKeyEvent*, char *args[], int num_args)
{
  if (num_args != 1)
    {
      cerr << "process_keyAP: invalid arguments" << endl;
      return;
    }

  if (strcmp("left", args[0]) == 0)
    XmScrollBarSetValues(gvhscroll, max(x_min, x_value - x_inc),
			 x_slider_size, x_inc, x_page_inc, True);
  else if (strcmp("right", args[0]) == 0)
    XmScrollBarSetValues(gvhscroll,
			 min(x_max - x_slider_size, x_value + x_inc),
			 x_slider_size, x_inc, x_page_inc, True);
  else if (strcmp("up", args[0]) == 0)
    XmScrollBarSetValues(gvvscroll, max(y_min, y_value - y_inc),
			 y_slider_size, y_inc, y_page_inc, True);
  else if (strcmp("down", args[0]) == 0)
    XmScrollBarSetValues(gvvscroll,
			 min(y_max - y_slider_size, y_value + y_inc),
			 y_slider_size, y_inc, y_page_inc, True);
  else if (strcmp("pup", args[0]) == 0)
    if (current_page > 0)
      change_pageCB(NULL, (XtPointer)-1, NULL);
  else if (strcmp("pdown", args[0]) == 0)
    if (current_page < doc->numpages - 1)
      change_pageCB(NULL, (XtPointer)+1, NULL);
}


void DocWindow::process_keyEH(Widget, XtPointer, XKeyEvent *event, Boolean*)
{
  char buffer[20];

  KeySym key;
  XComposeStatus compose;
  XLookupString(event, buffer, 20, &key, &compose);

  switch(key)
    {
    case XK_Left:
    case XK_KP_Left:
      XmScrollBarSetValues(gvhscroll, max(x_min, x_value - x_inc),
			   x_slider_size, x_inc, x_page_inc, True);
      break;
    case XK_Right:
    case XK_KP_Right:
      XmScrollBarSetValues(gvhscroll,
			   min(x_max - x_slider_size, x_value + x_inc),
			   x_slider_size, x_inc, x_page_inc, True);
      break;
    case XK_Up:
    case XK_KP_Up:
      XmScrollBarSetValues(gvvscroll, max(y_min, y_value - y_inc),
			   y_slider_size, y_inc, y_page_inc, True);
      break;
    case XK_Down:
    case XK_KP_Down:
      XmScrollBarSetValues(gvvscroll,
			   min(y_max - y_slider_size, y_value + y_inc),
			   y_slider_size, y_inc, y_page_inc, True);
      break;
    case XK_Prior:
    case XK_KP_Prior:
      if (current_page > 0)
	change_pageCB(NULL, (XtPointer)-1, NULL);
      break;
    case XK_Next:
    case XK_KP_Next:
      if (current_page < doc->numpages - 1)
	change_pageCB(NULL, (XtPointer)+1, NULL);
      break;
    }
}


void DocWindow::toggle_annotationsCB(Widget, XtPointer,
				     XmToggleButtonCallbackStruct *cbs)
{
  if (cbs->set)
    allow_annotations = True;
  else
    allow_annotations = False;

  XtSetSensitive(annotation_icon, allow_annotations);

  // Manage the annotation buttons for the current page:
  for(int i = 0; i < annotations.size(); i++)
    {
      Annotation * const this_annotation = ((Annotation*)(annotations[i]));

      if (this_annotation->page == current_page)
	if (allow_annotations)
	  XtManageChild(this_annotation->icon_form);
	else
	  XtUnmanageChild(this_annotation->icon_form);
    }
}


// Macros for string constants:
#define length(a) (sizeof(a)-1)
#define iscomment(a, b)	(strncmp(a, b, length(b)) == 0)
#define DSCcomment(a) (a[0] == '%' && a[1] == '%')


void DocWindow::save_file(const int column, const char *filename)
{
  FILE *output_file = fopen(filename, "w");

  if (!output_file)
    {
      write_message("Save failed: could not open output file\n");
      return;
    }

  export_ps(column, output_file);

  fclose(output_file);
}


void broken_pipe_handler()
{
  cerr << "Warning: broken pipe reported" << endl;
}


void DocWindow::print_file(const int column, const char *command)
{
  SIGNALS_INIT();
  SIGNAL_SET_IGNORE(SIGPIPE);

  FILE *printer = popen(command, "w");

  if (!printer)
    {
      write_message("Print failed: could not run specified command\n");
      return;
    }

  export_ps(column, printer);

  pclose(printer);
  SIGNAL_SET_DEFAULT(SIGPIPE);
}


void DocWindow::export_ps(const int column, FILE *output_file)
{
  FILE *input_file = fopen((char*)(const char*)psfilename, "r");

  if (!input_file)
    {
      write_message("Export failed: could not open input file\n");
      return;
    }

  if (column == CENTER_COLUMN)
    {
      // If the entire document is selected, just copy the input file:
      const long start_position = ftell(input_file);
      fseek(input_file, -1, SEEK_END);
      const long stop_position = ftell(input_file);

      pscopy(input_file, output_file, start_position, stop_position);
    }
  else
    {
      // Otherwise, extract the selected pages:
      SortedCltn &pages = scan->get_column(column);

      char text[PSLINELENGTH];
      char *comment;
      Boolean pages_written = False;
      Boolean pages_atend = False;
      Boolean marked_pages = False;
      const int num_pages = pages.size();
      int page = 1;

      long here = doc->beginheader;
      while (comment = pscopyuntil(input_file, output_file, here,
				   doc->endheader, "%%Pages:"))
	{
	  here = ftell(input_file);
	  if (pages_written || pages_atend)
	    {
	      free(comment);
	      continue;
	    }

	  sscanf(comment+length("%%Pages:"), "%s", text);
	  if (strcmp(text, "(atend)") == 0)
	    {
	      fputs(comment, output_file);
	      pages_atend = True;
	    }
	  else
	    {
	      int i;
	      switch (sscanf(comment+length("%%Pages:"), "%*d %d", &i))
		{
		case 1:
		  fprintf(output_file, "%%%%Pages: %d %d\n", num_pages, i);
		  break;
		default:
		  fprintf(output_file, "%%%%Pages: %d\n", num_pages);
		  break;
		}
	      pages_written = True;
	    }
	  free(comment);
	}

      fseek(input_file, doc->endheader, SEEK_SET);

      // Discard the saved bookmarks:
      char copy_line[PSLINELENGTH], page_name[PSLINELENGTH];
      int page_number, num_read;

      fgets(copy_line, PSLINELENGTH, input_file);
      num_read =
	sscanf(copy_line, "%%BrowserBookmark %d %s", &page_number, page_name);
      while(num_read == 2)
	{
	  fgets(copy_line, PSLINELENGTH, input_file);
	  num_read = sscanf(copy_line, "%%BrowserBookmark %d %s",
			    &page_number, page_name);
	}

      // Discard the annotations:
      char discard_text[PSLINELENGTH];
      float discard_x, discard_y;
      int discard_page;

      while(sscanf(copy_line, "%%BrowserAnnotation %d %f %f",
		   &discard_page, &discard_x, &discard_y) == 3)
	{
	  fgets(copy_line, PSLINELENGTH, input_file);

	  while(strncmp(BROWSER_ANNOTATION_END, copy_line,
			strlen(BROWSER_ANNOTATION_END)) != 0)
	    fgets(copy_line, PSLINELENGTH, input_file);
	  fgets(copy_line, PSLINELENGTH, input_file);
	}
      fputs(copy_line, output_file);

      pscopy(input_file, output_file, ftell(input_file), doc->endprolog);
      pscopy(input_file, output_file, doc->beginsetup, doc->endsetup);

      int i;
      for (i = 0; i < num_pages; i++)
	{
	  Thumbnail *tn = (Thumbnail*)(pages[i]);
	  const int j = tn->get_page_number(); //virtual_to_real_page(i);

	  comment = pscopyuntil(input_file, output_file,
				doc->pages[j].begin, doc->pages[j].end,
				"%%Page:");
	  fprintf(output_file, "%%%%Page: %s %d\n",
		  doc->pages[j].label, page++);
	  free(comment);
	  pscopy(input_file, output_file, -1, doc->pages[j].end);
	}

      here = doc->begintrailer;
      while (comment = pscopyuntil(input_file, output_file, here,
				   doc->endtrailer, "%%Pages:"))
	{
	  here = ftell(input_file);
	  if (pages_written)
	    {
	      free(comment);
	      continue;
	    }
	  switch (sscanf(comment+length("%%Pages:"), "%*d %d", &i))
	    {
	    case 1:
	      fprintf(output_file, "%%%%Pages: %d %d\n", num_pages, i);
	      break;
	    default:
	      fprintf(output_file, "%%%%Pages: %d\n", num_pages);
	      break;
	    }
	  pages_written = True;
	  free(comment);
	}
    }

  fclose(input_file);
}

#undef length
#undef iscomment
#undef DSCcomment


void DocWindow::jump_to_pos(float x, float y)
{
  XmScrollBarSetValues(gvhscroll,
		       min(x_max - x_slider_size,
			   max(0, (int)(x * gvwidth) - x_slider_size / 2)),
		       x_slider_size, x_inc, x_page_inc, True);
  XmScrollBarSetValues(gvvscroll,
		       min(y_max - y_slider_size,
			   max(0, (int)(y * gvheight) - y_slider_size / 2)),
		       y_slider_size, y_inc, y_page_inc, True);  
}


void DocWindow::get_infoCB(Widget, XtPointer, XtPointer)
{
  String info_msg = "File: ";
  char scratch[256];

  info_msg &= psfilename;
  info_msg &= "\n";

  if (doc->title)
    {
      info_msg &= doc->title;
      info_msg &= "\n";
    }

  if (doc->date)
    {
      info_msg &= "Created: ";
      info_msg &= doc->date;
      info_msg &= "\n";
    }

  sprintf(scratch, "%d total document mark-ups\n", annotations.size());

  info_msg &= scratch;

  sprintf(scratch, "Page %d (%s) of %d", current_page + 1,
	  doc->pages[current_page].label, doc->numpages);

  info_msg &= scratch;

  if (current_media)
    {
      sprintf(scratch, ", using media \"%s\"", current_media->name);
      info_msg &= scratch;
    }

  MessageBox page_error(shell, XmDIALOG_INFORMATION,
			(char*)(const char*)info_msg, NULL, MBOkButton,
			XmDIALOG_PRIMARY_APPLICATION_MODAL);

  int discard = page_error.get_response();
}


int DocWindow::os_to_gv_orientation(const int os_orientation)
{
  // Convert the orientation from OS_... to degrees for Ghostview
  switch (os_orientation)
    {
    case OS_PORTRAIT:
      return(XtPageOrientationPortrait);
    case OS_LANDSCAPE:
      return(XtPageOrientationLandscape);
    case OS_SEASCAPE:
      return(XtPageOrientationSeascape);
    case OS_INVERTED:
      return(XtPageOrientationUpsideDown);
    case OS_AUTO:
      cerr << "Warning: physical orientation set to auto (defaulted to portrait)." << endl;
      return(XtPageOrientationPortrait);
    }
}


void DocWindow::convert_to_virtual_coords(float &x, float &y,
					  const int orientation)
{
  float x_coord_virtual, y_coord_virtual;

  switch(orientation)
    {
    case OS_PORTRAIT:
      x_coord_virtual = x;
      y_coord_virtual = y;
      break;
    case OS_LANDSCAPE:
      x_coord_virtual = y;
      y_coord_virtual = 1 - x;
      break;
    case OS_INVERTED:
      x_coord_virtual = 1 - x;
      y_coord_virtual = 1 - y;
      break;
    case OS_SEASCAPE:
      x_coord_virtual = 1 - y;
      y_coord_virtual = x;
      break;
    }

  x = x_coord_virtual;
  y = y_coord_virtual;
}



void DocWindow::call_popup_callbackCB(Widget widget, int cb_index,
				      XtPointer call_data)
{
  if (cb_index == CB_DELETE)
    {
      delete_annotationCB(NULL, callback_root, NULL);
      callback_root = NULL;
    }

  else (callback_root->*(popup_callbacks[cb_index]))(widget, NULL, call_data);
}


void DocWindow::toggle_regionCB(Widget widget, int,
				XmToggleButtonCallbackStruct *cbs)
{
  Widget this_widget = scan->get_top_widget();

  if (cbs->set)
    {
      XtManageChild(this_widget);
      XtVaSetValues(gvscrolled,
		    XmNleftAttachment, XmATTACH_WIDGET,
		    XmNleftWidget, this_widget,
		    NULL);
    }
  else
    {
      XtUnmanageChild(this_widget);
      // Unattach the frame to let the window resize:
      XtVaSetValues(gvscrolled,
		    XmNleftAttachment, XmATTACH_FORM,
		    NULL);
    }
}
