
//  This is file ExportDialog.C.  It implements a class for the
//  "Export Postscript"/"Print" dialogs, as defined in ExportDialog.h.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/



#include "ExportDialog.h"
#include "TopShell.h"

#include <string.h>

#define String X11String
#define Object X11Object

#include <X11/Intrinsic.h>
#include <xpm.h>
#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/BulletinB.h>
#include <Xm/Form.h>
#include <Xm/Separator.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/FileSB.h>
#include <Xm/Text.h>

extern "C" {
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
}

#undef String
#undef Object


#define	THIS	ExportDialog
#define	BASE	UIObject
#define	BASE_CLASSES BASE::desc()
#define	MEMBER_CLASSES
#define VIRTUAL_BASE_CLASSES

DEFINE_REDUCED_CLASS(ExportDialog,1,"$Header: /home/cellar/ginsburg/browser/RCS/ExportDialog.C,v 2.2 1996/07/05 15:42:45 ginsburg Exp ginsburg $",NULL)


bool ExportDialog::isEqual(const Object& p) const
// Test two objects for equality
{
#ifdef FULL_NIHCL
  return p.isSpecies(classDesc) && *this==castdown(p);
#else
  return this == &p;
#endif
}


// Create a new ExportDialog.  The new dialog is created as a child of
// the specified widget, and its configuration is determined by the
// value of new_type, either TYPE_SAVE_DIALOG or TYPE_PRINT_DIALOG.
// The value of default_text is used to initialize the text line,
// which is used either as the print command or the filename.

ExportDialog::ExportDialog(Widget parent, DocWindow *new_parent_obj, const int new_type,
			   const char *default_text)
:parent_obj(new_parent_obj),
 type(new_type),
 column(0),
 default_filename(strdup(default_text))
{
  // Create a new dialog with the appropriate title:
  shell =
    XtCreatePopupShell("export", xmDialogShellWidgetClass, parent, NULL, 0);

  XtVaSetValues(shell,
		XmNtitle,
		(type == TYPE_SAVE_DIALOG) ?
		  "Export PostScript" : "Print selections",
		XmNallowShellResize, False,
		XmNautoUnmanage, False,
		NULL);

  // Use a form as the dialog's manager widget, store "this" pointer in
  // userData for later use:
  form = XtVaCreateWidget("form", xmFormWidgetClass, shell,
 			  XmNfractionBase, 5,
 			  XmNnoResize, True,
			  XmNautoUnmanage, False,
			  XmNuserData, this,
 			  NULL);

  // Create an "OK" button:
  ok_button = XtVaCreateWidget("OK", xmPushButtonWidgetClass, form,
			       XmNleftAttachment, XmATTACH_POSITION,
			       XmNleftPosition, 1,
			       XmNrightAttachment, XmATTACH_POSITION,
			       XmNrightPosition, 2,
			       XmNbottomAttachment, XmATTACH_FORM,
			       XmNbottomOffset, 10,
			       NULL);

  register_cb(ok_button, XmNactivateCallback, (CBType)&ExportDialog::okCB, NULL);

  // Create a "Cancel" button:
  cancel_button = XtVaCreateWidget("Cancel", xmPushButtonWidgetClass, form,
				   XmNleftAttachment, XmATTACH_POSITION,
				   XmNleftPosition, 3,
				   XmNrightAttachment, XmATTACH_POSITION,
				   XmNrightPosition, 4,
				   XmNbottomAttachment, XmATTACH_FORM,
				   XmNbottomOffset, 10,
				   NULL);

  register_cb(cancel_button, XmNactivateCallback,
	      (CBType)&UIObject::popdownCB, shell);

  // Deleting the window also cancels:
  register_wm_cb(shell, delete_window_atom,
		 (CBType)&UIObject::popdownCB, shell);


  // Create a separator between the buttons and the work area above:
  separator = XtVaCreateWidget("separator", xmSeparatorWidgetClass, form,
			       XmNleftAttachment, XmATTACH_FORM,
			       XmNrightAttachment, XmATTACH_FORM,
			       XmNbottomAttachment, XmATTACH_WIDGET,
			       XmNbottomWidget, ok_button,
			       XmNbottomOffset, 10,
			       NULL);

  // Create a text line to specify either the filename or the print command:
  text_line = XtVaCreateWidget("text_line", xmTextWidgetClass, form,
			       XmNleftAttachment, XmATTACH_FORM,
			       XmNleftOffset, 10,
			       XmNbottomAttachment, XmATTACH_WIDGET,
			       XmNbottomWidget, separator,
			       XmNbottomOffset, 10,
			       XmNeditMode, XmSINGLE_LINE_EDIT,
			       XmNmaxLength, 250,
			       XmNeditable, True,
			       XmNvalue, default_filename,
			       NULL);

  if (type == TYPE_SAVE_DIALOG)
    {
      // If this is a save dialog, create a "Browse" button next to
      // the text line and a file selection box:
      browse_button = XtVaCreateWidget("Browse...", xmPushButtonWidgetClass, form,
				       XmNbottomAttachment, XmATTACH_WIDGET,
				       XmNbottomWidget, separator,
				       XmNbottomOffset, 10,
				       XmNrightAttachment, XmATTACH_FORM,
				       XmNrightOffset, 10,
				       NULL);

      XtVaSetValues(text_line,
		    XmNrightAttachment, XmATTACH_WIDGET,
		    XmNrightWidget, browse_button,
		    XmNrightOffset, 10,
		    NULL);

      // Create the file selection dialog:
      filebox = XmCreateFileSelectionDialog(browse_button, "filebox", NULL, 0);

      XtVaSetValues(XtParent(filebox),
		    XtNtitle, "Select File for Export",
		    NULL);

      XtVaSetValues(filebox,
		    XtVaTypedArg, XmNpattern, XtRString,
		    TopShell::appdata.pattern,
		    strlen(TopShell::appdata.pattern) + 1,
		    NULL);

      register_cb(filebox, XmNokCallback,
		  (CBType)&ExportDialog::fileokCB, NULL);
      register_cb(filebox, XmNcancelCallback,
		  (CBType)&UIObject::unmanageCB, filebox);

      // Remove the help button
      XtUnmanageChild(XmFileSelectionBoxGetChild(filebox,
						 XmDIALOG_HELP_BUTTON));

      // Cause the "Browse" button to manage the file selection dialog:
      register_cb(browse_button, XmNactivateCallback,
		  (CBType)&UIObject::manageCB, (XtPointer)filebox);
    }
  else
    {
      // If this is a print dialog, do not create a button or file
      // selection dialog; instead, attach the text line to the right
      // edge of the dialog:
      XtVaSetValues(text_line,
		    XmNrightAttachment, XmATTACH_FORM,
		    XmNrightOffset, 10,
		    NULL);

      browse_button = filebox = NULL;
    }

  // Create an option menu to specify which column is to be exported:
  XmString menu_title =
    XmStringCreateSimple((type == TYPE_SAVE_DIALOG) ? "Save" : "Print");

  static XmString button_titles[3];

  button_titles[0] = XmStringCreateSimple("Persistent Bookmarks");
  button_titles[1] = XmStringCreateSimple("Entire Document");
  button_titles[2] = XmStringCreateSimple("Temporary Bookmarks");

  option_menu =
    XmVaCreateSimpleOptionMenu(form, "option_menu", menu_title,
			       (type == TYPE_SAVE_DIALOG) ? 'S' : 'P',
			       1, (XtCallbackProc)option_menuCB,
			       XmVaPUSHBUTTON, button_titles[0], 'P', NULL, NULL,
			       XmVaPUSHBUTTON, button_titles[1], 'D', NULL, NULL,
			       XmVaPUSHBUTTON, button_titles[2], 'T', NULL, NULL,
			       XmNleftAttachment, XmATTACH_FORM,
			       XmNleftOffset, 10,
			       XmNbottomAttachment, XmATTACH_WIDGET,
			       XmNbottomWidget, text_line,
			       XmNbottomOffset, 10,
			       XmNtopAttachment, XmATTACH_FORM,
			       XmNtopOffset, 10,
			       NULL);

  XmStringFree(menu_title);
  XmStringFree(button_titles[0]);
  XmStringFree(button_titles[1]);
  XmStringFree(button_titles[2]);

  // Store the menu button widgets in an array:
  menu_widgets[EXPORT_PERSISTENT_BOOKMARKS] = XtNameToWidget(shell, "*button_0");
  menu_widgets[EXPORT_ENTIRE_DOCUMENT] = XtNameToWidget(shell, "*button_1");
  menu_widgets[EXPORT_TEMPORARY_BOOKMARKS] = XtNameToWidget(shell, "*button_2");

  // Create a label to print the word "in" after the option menu.  The
  // final effect should be "Save [Entire Document] in [filename]" or
  // "Print [Entire Document] using [command]".
  in_label = XtVaCreateWidget((type == TYPE_SAVE_DIALOG) ? "in" : "using",
			      xmLabelWidgetClass, form,
			      XmNleftAttachment, XmATTACH_WIDGET,
			      XmNleftWidget, option_menu,
//			      XmNleftOffset, 10,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNtopOffset, 10,
			      //XmNrightAttachment, XmATTACH_FORM,
			      //XmNrightOffset, 10,
			      XmNbottomAttachment, XmATTACH_WIDGET,
			      XmNbottomWidget, text_line,
			      XmNbottomOffset, 10,
			      NULL);

  // Store the addresses of the form children for fast management:
  children = new Widget[6];

  children[0] = option_menu;
  children[1] = in_label;
  children[2] = text_line;
  children[3] = separator;
  children[4] = ok_button;
  children[5] = cancel_button;
}


// Set the dialog controls to their initial values, manage the
// widgets, and popup the dialog.

void ExportDialog::popup_and_manage()
{
  // Reset the text line
  XmTextSetString(text_line, default_filename);

  // Manage the widgets:
  XtManageChildren(children, 6);

  if (browse_button)
    XtManageChild(browse_button);

  XtManageChild(form);

  // Put the insertion point at the end of the text
  XmTextSetInsertionPosition(text_line, XmTextGetLastPosition(text_line));

  // Reset the option menu to the middle item ("Entire Document"):
  XtVaSetValues(option_menu,
		XmNmenuHistory, menu_widgets[EXPORT_ENTIRE_DOCUMENT],
		NULL);

  column = 2;			// Center column

  // Popup the dialog
  XtPopup(shell, XtGrabNonexclusive);
}


// Make the menu buttons sensitive or insensitive according to the
// value of column_mask, the bitwise OR of ENABLE_?_COLUMN values.

void ExportDialog::set_available(const int column_mask)
{
  // (The center button is never insensitive.)
  XtSetSensitive(menu_widgets[EXPORT_PERSISTENT_BOOKMARKS], column_mask & ENABLE_LEFT_COLUMN);
  XtSetSensitive(menu_widgets[EXPORT_TEMPORARY_BOOKMARKS], column_mask & ENABLE_RIGHT_COLUMN);
}


// Called when a menu option is selected; first obtains the value of
// "this" from the XmNuserData resource of the shell widget, i.e. the
// third parent of a menu button.

void ExportDialog::option_menuCB(Widget widget, int button, XtPointer)
{
  ExportDialog *this_obj;

  XtVaGetValues(XtParent(XtParent(XtParent(widget))),
		XmNuserData, &this_obj,
		NULL);

  // button ranges from 0..2; convert this to 1..3
  this_obj->column = button + 1;
}


// Obtain the value of the text line, then call the parent DocWindow's
// save_file or print_file methods according to the dialog type.

void ExportDialog::okCB(Widget, XtPointer, XtPointer)
{
  // Get the text, then call the appropriate DocWindow method with the
  // current column:
  char *text_string = XmTextGetString(text_line);

  if (type == TYPE_SAVE_DIALOG)
    parent_obj->save_file(column, text_string);
  else
    parent_obj->print_file(column, text_string);

  // Free the text, then popdown the dialog:
  XtFree(text_string);

  XtPopdown(shell);
}


// The "OK" button in the file selection dialog was pressed; copy the
// selection to the text line.

void ExportDialog::fileokCB(Widget, XtPointer, XmFileSelectionBoxCallbackStruct *cbs)
{
  char *text;

  // Get the selected filename, then set the contents of text line:
  if (XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &text))
    {
      XmTextSetString(text_line, text);
      XmTextSetInsertionPosition(text_line, XmTextGetLastPosition(text_line));
      XtFree(text);
    }

  XtUnmanageChild(filebox);
}


// Destroy associated widgets and deallocate storage.

ExportDialog::~ExportDialog()
{
  XtDestroyWidget(shell);
  delete [] children;
  free(default_filename);
}
