
// This is file MessageBox.C.  It implements a class for managing
// message dialogs, described in MessageBox.h.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/



#include "MessageBox.h"
#include "TopShell.h"

#define String X11String
#define Object X11Object

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/MessageB.h>

#undef String
#undef Object


#define	THIS	MessageBox
#define	BASE	UIObject
#define	BASE_CLASSES BASE::desc()
#define	MEMBER_CLASSES
#define VIRTUAL_BASE_CLASSES

DEFINE_REDUCED_CLASS(MessageBox,1,"$Header: /a/cellar62/home/ginsburg/browser/RCS/MessageBox.C,v 2.0 1996/03/13 20:51:45 ginsburg Exp $",NULL)


bool MessageBox::isEqual(const Object& p) const
// Test two objects for equality
{
//  return p.isSpecies(classDesc) && *this==castdown(p);
  return this == &p;
}


// Called when a button is pressed or the message box is popped down.

void MessageBox::responseCB(Widget, XtPointer client_data, XtPointer)
{
  // Unmanage the message box and record the answer:
  XtUnmanageChild(message_box);
  answer = (int)client_data;

  // Remove the popdownCallback to prevent it being invoked twice
  XtRemoveAllCallbacks(XtParent(message_box), XmNpopdownCallback);
}


// Create a new message dialog as a child of parent.  dialog_type
// specifies the type, e.g. XmDIALOG_WARNING, and msg gives the text
// to display.  button_labels is an array of three strings that
// specify the titles of the buttons in the dialog.  If button_labels
// is NULL, a set of default titles is used instead.  If any of the
// three elements of a non-null button_labels array is NULL, the
// default title for that element is used.  buttons is the bitwise-OR
// of the MB?Button values from MessageBox.h, and describes which
// buttons are to be managed in the dialog.  modality is set to one of
// the dialog style options from <Xm/Xm.h>.

MessageBox::MessageBox(Widget parent, int dialog_type, char *msg,
		       char *button_labels[] = NULL,
		       int buttons = MBOkButton | MBCancelButton,
		       int modality = XmDIALOG_MODELESS)
{
  XmString cs_labels[3], msg_text;
  static char *default_labels[] = {"OK", "Cancel", "Help"};

  // Convert the appropriate button labels to XmStrings:
  for (int i = 0; i < 3; i++)
    if (button_labels && button_labels[i])
      cs_labels[i] = XmStringCreateSimple(button_labels[i]);
    else
      cs_labels[i] = XmStringCreateSimple(default_labels[i]);

  // Create the message text; XmStringCreateLtoR converts newlines to
  // separators.
  msg_text = XmStringCreateLtoR(msg, XmFONTLIST_DEFAULT_TAG);

  // Create the dialog box:
  message_box = XmCreateMessageDialog(parent, "message_box", NULL, 0);
  XtVaSetValues(message_box,
		XmNdialogType, dialog_type,
		XmNmessageString, msg_text,
		XmNdialogStyle, modality,
		XmNokLabelString, cs_labels[0],
		XmNcancelLabelString, cs_labels[1],
		XmNhelpLabelString, cs_labels[2],
		XmNautoUnmanage, False,
		NULL);

  // Set the title resource and deleteResponse of the actual dialog:
  XtVaSetValues(XtParent(message_box),
		XmNtitle, "Message",
		XmNdeleteResponse, XmUNMAP,
		NULL);

  // Free the XmStrings that have been allocated:
  for (int i = 0; i < 3; i++)
    XmStringFree(cs_labels[i]);

  XmStringFree(msg_text);

  // Unmanage any unused buttons:
  if(!(buttons & MBOkButton))
    XtUnmanageChild(XmMessageBoxGetChild(message_box, XmDIALOG_OK_BUTTON));
  if(!(buttons & MBCancelButton))
    XtUnmanageChild(XmMessageBoxGetChild(message_box, XmDIALOG_CANCEL_BUTTON));
  if(!(buttons & MBHelpButton))
    XtUnmanageChild(XmMessageBoxGetChild(message_box, XmDIALOG_HELP_BUTTON));

  // Register responseCB as the callback for the different buttons and
  // for popdownCallback, i.e. when the window has been deleted:
  register_cb(message_box, XmNokCallback,
	      (CBType)&MessageBox::responseCB, (XtPointer)MBOkButton);
  register_cb(message_box, XmNcancelCallback,
	      (CBType)&MessageBox::responseCB, (XtPointer)MBCancelButton);
  register_cb(message_box, XmNhelpCallback,
	      (CBType)&MessageBox::responseCB, (XtPointer)MBHelpButton);

  // Guess which button should be activated when the dialog is
  // deleted.  First try CancelButton, the OkButton, then HelpButton:
  if (buttons & MBCancelButton)
    register_cb(XtParent(message_box), XmNpopdownCallback,
		(CBType)&MessageBox::responseCB, (XtPointer)MBCancelButton);
  else if (buttons & MBOkButton)
    register_cb(XtParent(message_box), XmNpopdownCallback,
		(CBType)&MessageBox::responseCB, (XtPointer)MBOkButton);
  else
    register_cb(XtParent(message_box), XmNpopdownCallback,
		(CBType)&MessageBox::responseCB, (XtPointer)MBHelpButton);
}


// Manage the message box and wait for a user response.  Return a
// value corresponding to the button pressed:

int MessageBox::get_response()
{
  answer = MBNoButton;
  XtManageChild(message_box);

  // Wait for an answer:
  while(answer == MBNoButton)
    XtAppProcessEvent(TopShell::app, XtIMAll);

  //  Ensure that dialog is popped down:
  XSync(XtDisplay(message_box), 0);
  XmUpdateDisplay(XtParent(message_box));

  return answer;
}
