
// This is file Thumbnail.C.  It implements an abstract base class for
// page thumbnail images, then derives classes for anti-aliased and
// one-bit versions.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/



#include "Thumbnail.h"
#include "ScanControl.h"
#include <stdlib.h>

#define	THIS	Thumbnail
#define	BASE	ReducedObj
#define	BASE_CLASSES BASE::desc()
#define	MEMBER_CLASSES
#define VIRTUAL_BASE_CLASSES

DEFINE_REDUCED_ABSTRACT_CLASS(Thumbnail,1,"$Header: /a/cellar62/home/ginsburg/browser/RCS/Thumbnail.C,v 2.1 1996/05/05 02:25:34 ginsburg Exp $",NULL)


bool Thumbnail::isEqual(const Object& p) const
// Test two objects for equality
{
  return (this == &p);
}


// Create a new thumbnail with the specified page name, number, and
// orientation

Thumbnail::Thumbnail(char * const name, const int number,
		     const unsigned int width, const unsigned int height,
		     const int new_orient)
:page_name(name),
 page_number(number),
 original_orientation(new_orient),
 orientation(new_orient),
 image_width(width),
 image_height(height),
 pixmaps_ready(0)
{
}


int Thumbnail::compare(const Object& p) const
// Compare two objects.  If *this > p return >0,
// *this == p return 0, and if *this < p return <0.
{
  const Thumbnail &test_obj = (const Thumbnail&)p;

  if (test_obj.page_number < page_number)
    return 1;
  if (test_obj.page_number > page_number)
    return -1;
  return 0;
}


// Rotate pm clockwise by pi/2 * difference, then return the new
// pixmap

Pixmap Thumbnail::transform_pixmap(Pixmap pm, int difference)
{
  // Three turns CCW == one turn CW, and vice versa:
  if (abs(difference) == 3)
    difference = difference / abs(difference);

  GC gc = XCreateGC(dpy, pm, 0, NULL);

  // Get the pixmap dimensions:
  Window root_win;
  int x,y;
  unsigned int width_return, height_return, depth_return, border_return;

  XGetGeometry(dpy, pm, &root_win, &x, &y, &width_return, &height_return,
	       &border_return, &depth_return);

  // Perform the rotations:
  while(difference)
    {
      // Copy the pixmap into an XImage:
      XImage *xi;
      xi = XGetImage(dpy, pm, 0, 0, width_return, height_return, AllPlanes, XYPixmap);

      if (abs(difference) >= 2)
	{
	  for(int i = 0; i < width_return; i++)
	    for(int j = 0; j < height_return; j++)
	      {
		const unsigned long temp = XGetPixel(xi, i, j);
		XPutPixel(xi, i, j, XGetPixel(xi, width_return - 1 - i,
					      height_return - 1 - j));
		XPutPixel(xi, width_return - 1 - i, height_return - 1 - j, temp);
	      }

	  XPutImage(dpy, pm, gc, xi, 0, 0, 0, 0, width_return, height_return);
	  XDestroyImage(xi);

	  // Move difference closer to zero by two
	  difference += -2 * (difference / abs(difference));
	}
      else
	{
	  // Create a new pixmap with the same visual/depth as pm:
	  Pixmap new_bitmap =
	    XCreatePixmap(dpy, pm, height_return, width_return, depth_return);

	  if (difference > 0)
	    {
	      // The bitmap should be rotated clockwise
	      for(int i = 0; i < height_return; i++)
		for(int j = 0; j < width_return; j++)
		  {
		    XSetForeground(dpy, gc, XGetPixel(xi, j, height_return - 1 - i));
		    XDrawPoint(dpy, new_bitmap, gc, i, j);
		  }
	      difference--;
	    }
	  else
	    {
	      // rotate counter-clockwise
	      for(int i = 0; i < height_return; i++)
		for(int j = 0; j < width_return; j++)
		  {
		    XSetForeground(dpy, gc, XGetPixel(xi, j, i));
		    XDrawPoint(dpy, new_bitmap, gc, i, width_return - 1 - j);
		  }
	      difference++;
	    }

	  // Swap height_return and width_return
	  swap(width_return, height_return);

	  // install the new bitmap
	  XFreePixmap(dpy, pm);
	  pm = new_bitmap;
	  XDestroyImage(xi);
	}
    }

  XFreeGC(dpy, gc);

  return pm;
}


// Implementation of class BitmapThumbnail:

#undef THIS
#define	THIS	BitmapThumbnail
#undef BASE
#define	BASE	Thumbnail
#undef BASE_CLASSES
#define	BASE_CLASSES BASE::desc()

DEFINE_REDUCED_CLASS(BitmapThumbnail,1,"$Header: /a/cellar62/home/ginsburg/browser/RCS/Thumbnail.C,v 2.1 1996/05/05 02:25:34 ginsburg Exp $",NULL)


// Create a new one-bit thumbnail without a bitmap image; initialize
// the bitmap component to XmUNSPECIFIED_PIXMAP instead.  When the
// thumbnail has been rendered, the bitmap will be replaced with an
// image of the page.

BitmapThumbnail::BitmapThumbnail(const unsigned int width, const unsigned int height,
				 char * const name, const int page, const int orientation)
:Thumbnail(name, page, width, height, orientation),
 thumbnail_bitmap(XmUNSPECIFIED_PIXMAP)
{
}


// Set the bitmap page image to the given pixmap, rotating it to
// match the current orientation.

void BitmapThumbnail::set_bitmap(Pixmap new_bitmap)
{
  thumbnail_bitmap = new_bitmap;
  const int difference = orientation - original_orientation;

  if (difference)
    thumbnail_bitmap = transform_pixmap(thumbnail_bitmap, difference);

  pixmaps_ready = 1;
}


void Thumbnail::do_reorient_geometry(int new_orientation)
{
  // thumbnails should be set to a physical orientation, not to OS_AUTO
  if (new_orientation == OS_AUTO)
    {
      cerr << "Warning: bitmap orientation set to auto; defaulted to portrait."
	<< endl;
      new_orientation = OS_PORTRAIT;
    }

  // Calculate the extent of the rotation required
  int difference = new_orientation - orientation;

  // Store the new orientation
  orientation = new_orientation;

  // Swap height and width if necessary:
  if (abs(difference % 2) == 1)
    swap(image_width, image_height);
}


void BitmapThumbnail::reorient_geometry(const int new_orientation)
{
  if (thumbnail_bitmap != XmUNSPECIFIED_PIXMAP)
   {
     XFreePixmap(dpy, thumbnail_bitmap);
     thumbnail_bitmap = XmUNSPECIFIED_PIXMAP;
   }

  do_reorient_geometry(new_orientation);
}


void AntiAliasedThumbnail::reorient_geometry(const int new_orientation)
{
  if (thumbnail_pixmap != XmUNSPECIFIED_PIXMAP)
   {
     XFreePixmap(dpy, thumbnail_pixmap);
     thumbnail_pixmap = XmUNSPECIFIED_PIXMAP;
   }

  if (active_pixmap != XmUNSPECIFIED_PIXMAP)
   {
     XFreePixmap(dpy, active_pixmap);
     active_pixmap = XmUNSPECIFIED_PIXMAP;
   }

  do_reorient_geometry(new_orientation);
}


// Change the thumbnail orientation by rotating the pixmap and
// potentially exchanging the width and height.  new_orientation is
// one of OS_PORTRAIT, OS_LANDSCAPE, OS_INVERTED, OS_SEASCAPE from
// DocWindow.h.

void BitmapThumbnail::reorient(int new_orientation, ScanControl *parent)
{
  // thumbnails should be set to a physical orientation, not to OS_AUTO
  if (new_orientation == OS_AUTO)
    {
      cerr << "Warning: bitmap orientation set to auto; defaulted to portrait."
	<< endl;
      new_orientation = OS_PORTRAIT;
    }

  // Calculate the extent of the rotation required
  int difference = new_orientation - orientation;

  // Store the new orientation
  orientation = new_orientation;

  // Rotate the bitmap
/*  if (thumbnail_bitmap != XmUNSPECIFIED_PIXMAP)
    thumbnail_bitmap = transform_pixmap(thumbnail_bitmap, difference);*/

  if (pixmaps_ready == 1)
    {
      RotatePixmapStruct *rps = new RotatePixmapStruct;
      rps->pm = &thumbnail_bitmap;
      rps->difference = difference;
      rps->this_obj = this;
      rps->parent = parent;

      pixmaps_ready = 0;
      XtAppAddWorkProc(TopShell::app, (XtWorkProc)rotate_pixmapWP, (XtPointer)rps);
    }

  // Swap height and width if necessary:
  if (abs(difference % 2) == 1)
    swap(image_width, image_height);
}



// Implementation of class AntiAliasedThumbnail

#undef THIS
#define	THIS	AntiAliasedThumbnail

DEFINE_REDUCED_CLASS(AntiAliasedThumbnail,1,"$Header: /a/cellar62/home/ginsburg/browser/RCS/Thumbnail.C,v 2.1 1996/05/05 02:25:34 ginsburg Exp $",NULL)


// Create a new anti-aliased thumbnail without a pixmap image;
// initialize the pixmap components to XmUNSPECIFIED_PIXMAP instead.
// When the thumbnail has been rendered, the pixmaps will be replaced
// with an image of the page.

AntiAliasedThumbnail::AntiAliasedThumbnail(const unsigned int width,
					   const unsigned int height,
					   char * const name, const int number,
					   const int orientation)
:Thumbnail(name, number, width, height, orientation),
 thumbnail_pixmap(XmUNSPECIFIED_PIXMAP),
 active_pixmap(XmUNSPECIFIED_PIXMAP)
{
}


// Set the page images to the given pixmaps, rotating them to
// match the current orientation.

void AntiAliasedThumbnail::set_pixmaps(Pixmap new_page, Pixmap new_active)
{
  thumbnail_pixmap = new_page;
  active_pixmap = new_active;
  // 4/24/96: orientation changes no longer necessary?
/*
  const int difference = orientation - original_orientation;

  // Rotate the pixmaps if necessary:
  if (difference)
    {
      thumbnail_pixmap = transform_pixmap(thumbnail_pixmap, difference);
      active_pixmap = transform_pixmap(active_pixmap, difference);
    }
*/
  pixmaps_ready = 2;
}


// Change the thumbnail orientation by rotating the pixmaps and
// potentially exchanging the width and height.  new_orientation is
// one of OS_PORTRAIT, OS_LANDSCAPE, OS_INVERTED, OS_SEASCAPE from
// DocWindow.h.

void AntiAliasedThumbnail::reorient(int new_orientation, ScanControl *parent)
{
  // new_orientation should not be OS_AUTO; silently replace it with
  // OS_PORTRAIT:
  if (new_orientation == OS_AUTO)
    new_orientation = OS_PORTRAIT;

  int difference = new_orientation - orientation;

  // Store the new orientation
  orientation = new_orientation;

  // Rotate the pixmaps if they have been installed:
/*  if (thumbnail_pixmap != XmUNSPECIFIED_PIXMAP)
    {
      thumbnail_pixmap = transform_pixmap(thumbnail_pixmap, difference);
      active_pixmap = transform_pixmap(active_pixmap, difference);
    }*/

  if (pixmaps_ready == 2)
    {
      RotatePixmapStruct *rps1 = new RotatePixmapStruct;
      rps1->pm = &thumbnail_pixmap;
      rps1->difference = difference;
      rps1->this_obj = this;
      rps1->parent = parent;

      RotatePixmapStruct *rps2 = new RotatePixmapStruct;
      rps2->pm = &active_pixmap;
      rps2->difference = difference;
      rps2->this_obj = this;
      rps2->parent = parent;

      pixmaps_ready = 0;
      XtAppAddWorkProc(TopShell::app, (XtWorkProc)rotate_pixmapWP, (XtPointer)rps1);
      XtAppAddWorkProc(TopShell::app, (XtWorkProc)rotate_pixmapWP, (XtPointer)rps2);
    }

  // Swap height and width if necessary:
  if (abs(difference % 2) == 1)
    swap(image_width, image_height);
}


// Draw an image of this thumbnail (or a blank rectangle if the bitmap
// is not yet installed) in the Window dest.  bg_gc contains black as
// the foreground color and the thumbnail background as the background
// color.  fg_gc contains the thumbnail background as the foreground
// color.

void BitmapThumbnail::draw_page(Window dest, GC fg_gc, GC bg_gc,
				int x, int y, unsigned int height, Boolean)
{
  if ((pixmaps_ready == 1) && (thumbnail_bitmap != XmUNSPECIFIED_PIXMAP))
    // Draw the thumbnail bitmap:
    XCopyPlane(dpy, thumbnail_bitmap, dest, bg_gc, 0, 0,
	       image_width, height, x, y, 1);
  else
    // Draw a blank rectangle
    XFillRectangle(dpy, dest, fg_gc, x, y, image_width, height);
}


// Draw an image of this thumbnail (or a blank rectangle if the bitmap
// is not yet installed) in the Window dest.  bg_gc contains black as
// the foreground color and the thumbnail background as the background
// color.  fg_gc contains the thumbnail background as the foreground
// color.  If active is true, the active_pixmap is drawn; otherwise
// the thumbnail_pixmap is used.  active_pixmap should have a lighter
// background that thumbnail_pixmap to indicate that it is currently
// selected.

void AntiAliasedThumbnail::draw_page(Window dest, GC fg_gc, GC bg_gc,
				     int x, int y, unsigned int height,
				     Boolean active)
{
  if ((pixmaps_ready == 2) && (active_pixmap != XmUNSPECIFIED_PIXMAP)
      && (thumbnail_pixmap != XmUNSPECIFIED_PIXMAP))
    {
      if (active)
	XCopyArea(dpy, active_pixmap, dest, bg_gc, 0, 0,
		  image_width, height, x, y);
      else
	XCopyArea(dpy, thumbnail_pixmap, dest, bg_gc, 0, 0,
		  image_width, height, x, y);
    }
  else
    XFillRectangle(dpy, dest, fg_gc, x, y, image_width, height);
}



Boolean Thumbnail::rotate_pixmapWP(RotatePixmapStruct *rps)
{
  *(rps->pm) =
    transform_pixmap(*(rps->pm), rps->difference);

  (rps->this_obj->pixmaps_ready)++;

  rps->parent->refresh_thumbnail(rps->this_obj);

  delete rps;

  return True;
}
