
// This is file TopShell.C.  It implements a class for a top-level
// application shell, defined in TopShell.h, as well as the
// application's global functions and variables.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/



#include "TopShell.h"
#include "DocWindow.h"
#include "MessageBox.h"
#include "Bitmap.h"
#include <stdio.h>
#include <signal.h>
#include <sys/wait.h>

#define String X11String
#define Object X11Object

#include <X11/StringDefs.h>
#include <xpm.h>
#include <X11/Xresource.h>
#include <Xm/Xm.h>
#include <Xm/AtomMgr.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/FileSB.h>
#include <Xm/SelectioB.h>
#include <Xm/DialogS.h>
#include <Xm/MessageB.h>
#include <Xm/DragIcon.h>
#include <Xm/Separator.h>

#undef String
#undef Object


// Bitmaps and pixmaps:

#include "icons/thumbnail"
#include "icons/icon.xpm"
#include "icons/dest.xpm"
#include "icons/empty.xpm"
#include "icons/note.xpm"
#include "icons/link_arrow.xpm"


#define	THIS	TopShell
#define	BASE	UIObject
#define	BASE_CLASSES BASE::desc()
#define	MEMBER_CLASSES
#define VIRTUAL_BASE_CLASSES

DEFINE_REDUCED_CLASS(TopShell,1,"$Header: /home/cellar/ginsburg/browser/RCS/TopShell.C,v 2.4 1996/07/05 15:42:38 ginsburg Exp ginsburg $",NULL)





static char info_msg[] =
"    DeckView, version 1.0, BETA 1

    DeckView -- a browser for Postscript(r) language documents
    Copyright (C) 1996  Adam Ginsburg

    Please send comments and bug reports to
    \"deckview@eecs.harvard.edu\".

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

";


XtAppContext TopShell::app;	// application context
AppData TopShell::appdata;	// application resource data


// Global variables:

// Contexts for storing "this" pointers in drag-and-drop routines
XContext docwin_context, scan_context;

// The atoms for different drag data
Atom existing_annotation_atom, create_link_atom, thumbnail_atom;
// Atom for the delete window window manager protocol
Atom delete_window_atom;

// Pixmaps/pixmap arrays:
// The "unlinked" icon used to create annotations
Pixmap annotation_icon_pixmaps[PIXMAPS_ARRAY_SIZE];
// A generic thumbnail page
Pixmap thumbnail_icon_pixmap;
// The "linked" icon used in annotation dialogs
Pixmap destination_icon_pixmaps[PIXMAPS_ARRAY_SIZE];
// The annotation text button icon
Pixmap note_pixmaps[PIXMAPS_ARRAY_SIZE];
// The empty marker icon
Pixmap marker_pixmaps[PIXMAPS_ARRAY_SIZE];
// The hyperlink button icon
Pixmap link_pixmaps[PIXMAPS_ARRAY_SIZE];

// The dimensions of some of the pixmaps:
unsigned int annotation_width, annotation_height;
unsigned int destination_width, destination_height;
unsigned int marker_width, marker_height;
unsigned int note_width, note_height;
unsigned int link_width, link_height;

// The display and screen on which the application is running:
Display *dpy;
Screen *scrn;


// The application resource definitions (defines static XtResource
// resources[]):

#include "resources.h"


// Command-line options:

static XrmOptionDescRec options[] =
{
  {"-antialias", "*antialias", XrmoptionNoArg, "False"},
  {"+antialias", "*antialias", XrmoptionNoArg, "True"},
  {"-gs", "*interpreter", XrmoptionSepArg, NULL},
  {"-dpi", "*thumbnailDPI", XrmoptionSepArg, NULL},
};


bool TopShell::isEqual(const Object& p) const
// Test two objects for equality
{
#ifdef FULL_NIHCL
	return p.isSpecies(classDesc) && *this==castdown(p);
#else
	return this == &p;
#endif
}


// Fallback resources to be combined with the resource database
// (defines static char fallback_resources[]).

#include "fallback.h"



// Create a new top-level shell

TopShell::TopShell(int argc, char **argv)
{
  // Initialize the application:
  toplevel = XtVaAppInitialize(&app, "DeckView", options, XtNumber(options),
			       &argc, argv, NULL,
			       XmNtitle, "DeckView",
			       XmNiconName, "DeckView Console",
			       XmNwidth, 225,
			       XmNheight, 75,
			       XmNdeleteResponse, XmDO_NOTHING,
			       NULL);

  XrmDatabase fallbackDB;

  fallbackDB = XrmGetStringDatabase(fallback_resources);
  XrmCombineDatabase(XrmGetDatabase(XtDisplay(toplevel)), &fallbackDB, True);
  XrmSetDatabase(XtDisplay(toplevel), fallbackDB);

  XtGetApplicationResources(toplevel, &appdata, resources,
			    XtNumber(resources), NULL, 0);

  // Set global variables:
  dpy = XtDisplay(toplevel);
  scrn = XtScreen(toplevel);
  delete_window_atom = XmInternAtom(dpy, "WM_DELETE_WINDOW", False);

  // Create the thumbnail and annotation icons for drag-and-drop:

  // Create the generic thumbnail page icon:
  // Load the bitmap
  Pixmap thumbnail_bitmap_pixmap =
    XCreateBitmapFromData(dpy, RootWindowOfScreen(scrn),
			  (const char *)thumbnail_bits,
			  thumbnail_width, thumbnail_height);

  // Deepen the bitmap to a pixmap:
  Bitmap thumbnail_bitmap(thumbnail_bitmap_pixmap,
			  thumbnail_width, thumbnail_height);
  thumbnail_icon_pixmap = thumbnail_bitmap.deepen();

  // Load the xpm images into the pixmap arrays:
  XpmAttributes attributes;
  attributes.valuemask = XpmSize;

  // Create the annotation text button pixmap:
  if (XpmCreatePixmapFromData(dpy, RootWindowOfScreen(scrn), note,
			      note_pixmaps + PIXMAPS_IMAGE,
			      note_pixmaps + PIXMAPS_MASK, &attributes)
      != XpmSuccess)
    cerr << "Warning: could not create icon pixmap" << endl;
  note_width = attributes.width;
  note_height = attributes.height;
  note_pixmaps[PIXMAPS_LABEL_PIXMAP] =
    note_pixmaps[PIXMAPS_ARM_PIXMAP] =
      XmUNSPECIFIED_PIXMAP;

  // Create the hyperlink button pixmap:
  if (XpmCreatePixmapFromData(dpy, RootWindowOfScreen(scrn), link_arrow,
			      link_pixmaps + PIXMAPS_IMAGE,
			      link_pixmaps + PIXMAPS_MASK, &attributes)
      != XpmSuccess)
    cerr << "Warning: could not create icon pixmap" << endl;

  link_width = attributes.width;
  link_height = attributes.height;
  link_pixmaps[PIXMAPS_LABEL_PIXMAP] =
    link_pixmaps[PIXMAPS_ARM_PIXMAP] =
      XmUNSPECIFIED_PIXMAP;

  // Create the empty marker pixmap:
  if (XpmCreatePixmapFromData(dpy, RootWindowOfScreen(scrn), empty,
			      marker_pixmaps + PIXMAPS_IMAGE,
			      marker_pixmaps + PIXMAPS_MASK, &attributes)
      != XpmSuccess)
    cerr << "Warning: could not create icon pixmap" << endl;

  marker_width = attributes.width;
  marker_height = attributes.height;
  marker_pixmaps[PIXMAPS_LABEL_PIXMAP] = XmUNSPECIFIED_PIXMAP;
  marker_pixmaps[PIXMAPS_ARM_PIXMAP] = XmUNSPECIFIED_PIXMAP;

  // Create the "unlinked" icon:
  if (XpmCreatePixmapFromData(dpy, RootWindowOfScreen(scrn), icon,
			      annotation_icon_pixmaps + PIXMAPS_IMAGE,
			      annotation_icon_pixmaps + PIXMAPS_MASK,
			      &attributes)
      != XpmSuccess)
    cerr << "Warning: could not create icon pixmap" << endl;

  annotation_width = attributes.width;
  annotation_height = attributes.height;
  annotation_icon_pixmaps[PIXMAPS_LABEL_PIXMAP] = XmUNSPECIFIED_PIXMAP;
  annotation_icon_pixmaps[PIXMAPS_ARM_PIXMAP] = XmUNSPECIFIED_PIXMAP;

  // Create the "linked" icon:
  if (XpmCreatePixmapFromData(dpy, RootWindowOfScreen(scrn), dest,
			      destination_icon_pixmaps + PIXMAPS_IMAGE,
			      destination_icon_pixmaps + PIXMAPS_MASK,
			      &attributes)
      != XpmSuccess)
    cerr << "Warning: could not create icon pixmap" << endl;

  destination_width = attributes.width;
  destination_height = attributes.height;
  destination_icon_pixmaps[PIXMAPS_LABEL_PIXMAP] = XmUNSPECIFIED_PIXMAP;
  destination_icon_pixmaps[PIXMAPS_ARM_PIXMAP] = XmUNSPECIFIED_PIXMAP;

  // Register the delete_window protocol callback:
  register_wm_cb(toplevel, delete_window_atom,
		 (CBType)&TopShell::quitCB, NULL);

  // Create the form widget and its pushbutton children:
  form = XtVaCreateManagedWidget("form",
				 xmFormWidgetClass, toplevel,
				 XmNfractionBase, 3,
				 NULL);

  XmString str = XmStringCreateSimple("Open");
  openbutton = XtVaCreateManagedWidget("openbutton",
				       xmPushButtonWidgetClass, form,
				       XmNlabelString, str,
				       XmNtopAttachment, XmATTACH_FORM,
				       XmNbottomAttachment, XmATTACH_FORM,
				       XmNleftAttachment, XmATTACH_FORM,
				       XmNrightAttachment, XmATTACH_POSITION,
				       XmNrightPosition, 1,
				       NULL);
  XmStringFree(str);
  str=XmStringCreateSimple("Info");
  infobutton = XtVaCreateManagedWidget("infobutton",
				       xmPushButtonWidgetClass, form,
				       XmNlabelString, str,
				       XmNtopAttachment, XmATTACH_FORM,
				       XmNbottomAttachment, XmATTACH_FORM,
				       XmNleftAttachment, XmATTACH_POSITION,
				       XmNleftPosition, 1,
				       XmNrightAttachment, XmATTACH_POSITION,
				       XmNrightPosition, 2,
				       NULL);
  XmStringFree(str);
  str=XmStringCreateSimple("Quit");
  quitbutton = XtVaCreateManagedWidget("quitbutton",
				       xmPushButtonWidgetClass, form,
				       XmNlabelString, str,
				       XmNtopAttachment, XmATTACH_FORM,
				       XmNbottomAttachment, XmATTACH_FORM,
				       XmNleftAttachment, XmATTACH_POSITION,
				       XmNleftPosition, 2,
				       XmNrightAttachment, XmATTACH_FORM,
				       NULL);
  XmStringFree(str);

  // Create a file selection dialog:
  filesb = XmCreateFileSelectionDialog(toplevel, "filesb", NULL, 0);

  XtVaSetValues(XtParent(filesb),
		XmNtitle, "Open Document",
		NULL);

  XtVaSetValues(filesb,
		XtVaTypedArg, XmNpattern, XtRString,
		appdata.pattern, strlen(appdata.pattern) + 1,
		NULL);

  register_cb(filesb, XmNokCallback,
	      (CBType)&TopShell::open_fileCB, NULL);
  register_cb(filesb, XmNcancelCallback, UIObject::unmanageCB, filesb);

  // Remove the help button
  XtUnmanageChild(XmSelectionBoxGetChild(filesb, XmDIALOG_HELP_BUTTON));

  // Create the information dialog:
  info_dialog = XmCreateFormDialog(toplevel, "info", NULL, 0);
  XtVaSetValues(XtParent(info_dialog),
		XmNdeleteResponse, XmUNMAP,
		XmNtitle, "DeckView Information",
		NULL);

  XtVaSetValues(info_dialog,
		XmNfractionBase, 15,
		NULL);

  info_okbutton =
    XtVaCreateManagedWidget("ok_button", xmPushButtonWidgetClass,
			    info_dialog,
			    XmNlabelString, XmStringCreateSimple("OK"),
			    XmNbottomAttachment, XmATTACH_FORM,
			    XmNbottomOffset, 10,
			    XmNleftAttachment, XmATTACH_POSITION,
			    XmNleftPosition, 7,
			    NULL);

  info_separator =
    XtVaCreateManagedWidget("separator", xmSeparatorWidgetClass, info_dialog,
			    XmNleftAttachment, XmATTACH_FORM,
			    XmNrightAttachment, XmATTACH_FORM,
			    XmNbottomAttachment, XmATTACH_WIDGET,
			    XmNbottomWidget, info_okbutton,
			    XmNbottomOffset, 10,
			    NULL);

  info_stext = XmCreateScrolledText(info_dialog, "info_text", NULL, 0);

  XtVaSetValues(info_stext,
		XmNscrollVertical, True,
		XmNscrollHorizontal, True,
		XmNeditMode, XmMULTI_LINE_EDIT,
		XmNeditable, False,
		XmNcursorPositionVisible, False,
		XmNwordWrap, False,
		XmNvalue, info_msg,
		XmNrows, 15,
		XmNcolumns, 70,
		NULL);

  XtVaSetValues(XtParent(info_stext),
		XmNleftAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		XmNtopAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, info_separator,
		XmNbottomOffset, 5,
		XmNtopOffset, 5,
		XmNleftOffset, 5,
		XmNrightOffset, 5,
		NULL);

  XtManageChild(info_stext);

  register_cb(info_okbutton, XmNactivateCallback,
	      UIObject::unmanageCB, info_dialog);


  // Register callbacks for the buttons:
  register_cb(openbutton, XmNactivateCallback, UIObject::manageCB, filesb);
  register_cb(infobutton, XmNactivateCallback,
	      UIObject::manageCB, info_dialog);
  register_cb(quitbutton, XmNactivateCallback,
	      (CBType)&TopShell::quitCB, NULL);

  // Create new contexts:
  docwin_context = XUniqueContext();
  scan_context = XUniqueContext();

  // Create atoms:
  existing_annotation_atom = XInternAtom(dpy, XbA_EXISTING_ANNOTATION, False);
  create_link_atom = XInternAtom(dpy, XbA_CREATE_LINK, False);
  thumbnail_atom = XInternAtom(dpy, XbA_THUMBNAIL, False);

  XtRealizeWidget(toplevel);

  // Open the files specified on the command line:
  XmFileSelectionBoxCallbackStruct cbs;
  for (int i = 1; i < argc; i++)
    {
      cbs.value = XmStringCreateSimple(argv[i]);
      open_fileCB(filesb, NULL, &cbs);
      XmStringFree(cbs.value);
    }
}


// Create a new DocWindow object to open a file

void TopShell::do_openfile(const char *filename)
{
  // Attempt to open the file:
  FILE *file = fopen(filename, "r");

  if (!file)
    {
      MessageBox file_alert(toplevel, XmDIALOG_ERROR,
			    "File could not be opened",
			    NULL, MBOkButton,
			    XmDIALOG_PRIMARY_APPLICATION_MODAL);

      int discard = file_alert.get_response();
    }
  else
    {
      // File was opened; create a new DocWindow object:
      int successful;
      pid_t pid;
      DocWindow *window=
	new DocWindow(toplevel, file, String(filename),
		      (CBType)&TopShell::delete_childCB, this,
		      successful, pid);
      if (successful)
	file_windows.add(*window);
      else
	{
	  delete window;
	  if (pid != -1)
	    {
	      // An instance of ghostview was started; save its pid:
	      PID *pid_class = new PID;
	      pid_class->pid = pid;

	      ghostview_pids.add(*pid_class);
	    }
	}
    }
}


// The open file callback function; extract the filename from the
// callback struct, then load it.

void TopShell::open_fileCB(Widget widget, XtPointer,
			   XmFileSelectionBoxCallbackStruct *cbs)
{
  // Process exposure events:
  XmUpdateDisplay(toplevel);

  // Get the filename:
  char *filename;

  if (!XmStringGetLtoR(cbs->value,XmSTRING_DEFAULT_CHARSET,&filename))
    return;

  if (!*filename)  // No filename selected
    {
      XtFree(filename);
      return;
    }

  // Unmanage the file selection dialog
  XtUnmanageChild(widget);

  do_openfile(filename);

  XtFree(filename);
}


// Create a DragIcon widget as the child of parent, using the pixmap
// as its XmNsourceIconPixmap.

Widget TopShell::create_drag_icon(Widget parent, Pixmap pixmap,
				  unsigned int width, unsigned int height)
{
  // Create an icon for the drag:
  static Arg args[10];
  static int n;

  n = 0;

  XtSetArg(args[n], XmNhotX, (width / 2)); n++;
  XtSetArg(args[n], XmNhotY, (height / 2)); n++;
  XtSetArg(args[n], XmNwidth, width); n++;
  XtSetArg(args[n], XmNheight, height); n++;
  XtSetArg(args[n], XmNpixmap, pixmap); n++;
  XtSetArg(args[n], XmNdepth, DefaultDepthOfScreen(scrn)); n++;

  return XmCreateDragIcon(parent, "drag_icon", args, n);
}


// Currently unused

Widget TopShell::create_rect_icon(Widget parent, const int width,
				  const int height)
{
  // Create the pixmap and mask:
  Pixmap icon_pixmap = XCreatePixmap(dpy, XtWindow(parent), width, height,
				     DefaultDepthOfScreen(scrn));
  GC gc = XCreateGC(dpy, icon_pixmap, 0, NULL);

  XSetForeground(dpy, gc, BlackPixelOfScreen(scrn));
  XSetBackground(dpy, gc, WhitePixelOfScreen(scrn));
  XSetFillStyle(dpy, gc, FillStippled);
  Bitmap stipple("/usr/include/X11/bitmaps/grainy");
  XSetStipple(dpy, gc, (Pixmap)stipple);

  XFillRectangle(dpy, icon_pixmap, gc, 0, 0, width, height);

  XSetFillStyle(dpy, gc, FillSolid);
  XDrawRectangle(dpy, icon_pixmap, gc, 0, 0, width - 1, height - 1);

  XFreeGC(dpy, gc);

/*  Pixmap icon_mask = XCreatePixmap(dpy, XtWindow(parent), width, height, 1);

  gc = XCreateGC(dpy, icon_mask, NULL, 0);

  XSetForeground(dpy, gc, WhitePixelOfScreen(scrn));
  XDrawRectangle(dpy, icon_mask, gc, 1, 1, width - 2, height - 2);
  XSetForeground(dpy, gc, BlackPixelOfScreen(scrn));
  XDrawRectangle(dpy, icon_mask, gc, 0, 0, width, height);

  XFreeGC(dpy, gc);
*/

  // Create an icon for the drag:
  static Arg args[10];
  static int n;

  n = 0;

  XtSetArg(args[n], XmNhotX, (width / 2)); n++;
  XtSetArg(args[n], XmNhotY, (height / 2)); n++;
  XtSetArg(args[n], XmNwidth, width); n++;
  XtSetArg(args[n], XmNheight, height); n++;
//  XtSetArg(args[n], XmNbackground, WhitePixelOfScreen(scrn)); n++;
//  XtSetArg(args[n], XmNforeground, BlackPixelOfScreen(scrn)); n++;
  XtSetArg(args[n], XmNpixmap, icon_pixmap); n++;
  XtSetArg(args[n], XmNdepth, DefaultDepthOfScreen(scrn)); n++;
  XtSetArg(args[n], XmNattachment, XmATTACH_HOT); n++;
//  XtSetArg(args[n], XmNmask, icon_mask); n++;

  return XmCreateDragIcon(parent, "drag_icon", args, n);
}


// Delete the DocWindow objects and exit the application

void TopShell::quitCB(Widget, XtPointer, XtPointer)
{
  for (int i = 0; i < file_windows.size(); i++)
    delete (DocWindow*)(file_windows[i]);

  for (int i = 0; i < ghostview_pids.size(); i++)
    {
      // Cause the programs to exit, then wait on them
      PID *this_pid = (PID*)(ghostview_pids[i]);

      kill(this_pid->pid, SIGKILL);
      waitpid(this_pid->pid, NULL, 0);
      
      delete (PID*)(ghostview_pids[i]);
    }

  file_windows.removeAll();
  ghostview_pids.removeAll();

  exit(0);
}


// Delete the child DocWindow object passed as the client_data
// argument

void TopShell::delete_childCB(Widget, XtPointer child, XtPointer)
{
  // Delete window only if it still exists
  if (file_windows.indexOf(*(DocWindow*)child) != -1)
    {
      file_windows.remove(*(DocWindow*)child);
      delete ((DocWindow*)child);
    }
}


void TopShell::delete_and_reload(DocWindow *obj, const String filename)
{
  // Remove the window in question
  delete_childCB(NULL, obj, NULL);

  // Reload the file
  do_openfile((const char *)filename);
}


// Use the image and mask pixmaps, along with the background color
// provided, to create and return a new pixmap with the transparent
// pixels set to the background

Pixmap create_masked_pixmap(Pixmap image, Pixmap mask,
			    unsigned long transparent_pixel)
{
  // Get the pixmap geometry:
  Window root;
  int x,y;
  unsigned int width, height, border_width, depth;

  XGetGeometry(dpy, image, &root, &x, &y, &width, &height,
	       &border_width, &depth);

  GC gc = XCreateGC(dpy, image, 0, NULL);

  // Copy image to a new pixmap:
  Pixmap pixmap = XCreatePixmap(dpy, image, width, height, depth);
  XCopyArea(dpy, image, pixmap, gc, 0, 0, width, height, 0, 0);

  // Use an XImage for efficient masking:
  XImage *xi = XGetImage(dpy, mask, 0, 0, width, height, AllPlanes, XYPixmap);
  XSetForeground(dpy, gc, transparent_pixel);

  // Draw over the transparent pixels with the specified color:
  for (int i = 0; i < width; i++)
    for (int j = 0; j < height; j++)
      if (XGetPixel(xi, i, j) == 0)
	XDrawPoint(dpy, pixmap, gc, i, j);

  XFreeGC(dpy, gc);
  XDestroyImage(xi);

  return pixmap;
}


// Assign the appropriate pixmaps from pixmaps to XmNlabelPixmap and
// XmNarmPixmap.  pixmaps is an array of the form {image, mask,
// background_pixmap, arm_pixmap} (the second pair is computed from
// the first pair, then the first pair are destroyed and replaced with
// XmUNSPECIFIED_PIXMAP).

void assign_button_pixmaps(Widget button, Pixmap pixmaps[])
{
  if (pixmaps[PIXMAPS_LABEL_PIXMAP] == XmUNSPECIFIED_PIXMAP)
    {
      // Compute the second pixmap set from the first, using the
      // background colors in the widget:
      unsigned long background_pixel, arm_pixel;

      XtVaGetValues(button,
		    XmNbackground, &background_pixel,
		    XmNarmColor, &arm_pixel,
		    NULL);

      pixmaps[PIXMAPS_LABEL_PIXMAP] =
	create_masked_pixmap(pixmaps[PIXMAPS_IMAGE], pixmaps[PIXMAPS_MASK],
			     background_pixel);
      pixmaps[PIXMAPS_ARM_PIXMAP] =
	create_masked_pixmap(pixmaps[PIXMAPS_IMAGE], pixmaps[PIXMAPS_MASK],
			     arm_pixel);
    }

  // Install the pixmaps:
  XtVaSetValues(button,
		XmNlabelPixmap, pixmaps[PIXMAPS_LABEL_PIXMAP],
		XmNarmPixmap, pixmaps[PIXMAPS_ARM_PIXMAP],
		XmNlabelType, XmPIXMAP,
		NULL);
}


// Assign the appropriate pixmap from the array to XmNlabelPixmap.
// Compute the label pixmap if necessary.

void assign_label_pixmap(Widget label, Pixmap pixmaps[])
{
  if (pixmaps[PIXMAPS_LABEL_PIXMAP] == XmUNSPECIFIED_PIXMAP)
    {
      // Compute the label pixmap from the image and mask, using the
      // background color in the widget:

      unsigned long background_pixel;

      XtVaGetValues(label,
		    XmNbackground, &background_pixel,
		    NULL);

      pixmaps[PIXMAPS_LABEL_PIXMAP] =
	create_masked_pixmap(pixmaps[PIXMAPS_IMAGE], pixmaps[PIXMAPS_MASK],
			     background_pixel);
    }

  // Install the pixmap:
  XtVaSetValues(label,
		XmNlabelPixmap, pixmaps[PIXMAPS_LABEL_PIXMAP],
		XmNlabelType, XmPIXMAP,
		NULL);
}

