//  This may look like C, but it is really -*- C++ -*-

// This is file TopShell.h.  It defines a class for a top-level
// application shell, implemented in TopShell.C, as well as the
// application's global functions and variables.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/



#ifndef TopShell_H
#define TopShell_H


#include "UIObject.h"
#include "DocWindow.h"
#include <nihcl/OrderedCltn.h>
#include <nihcl/String.h>
#include <stdlib.h>
#include <sys/wait.h>


// String defines for application resources:

#define XbNdimBackgroundColor  "dimPageBackground"
#define XbCDimBackgroundColor  "DimPageBackground"
#define XbNlevelOneGray  "levelOneGray"
#define XbCLevelOneGray  "LevelOneGray"
#define XbNlevelTwoGray  "levelTwoGray"
#define XbCLevelTwoGray  "LevelTwoGray"
#define XbNlevelThreeGray  "levelThreeGray"
#define XbCLevelThreeGray  "LevelThreeGray"
#define XbNinterpreter  "interpreter"
#define XbCInterpreter  "Interpreter"
#define XbNghostview  "ghostview"
#define XbCGhostview  "Ghostview"
#define XbNthumbnailDPI  "thumbnailDPI"
#define XbCThumbnailDPI  "ThumbnailDPI"
#define XbNantialias  "antialias"
#define XbCAntialias  "AntiAlias"
#define XbNjumpSize  "jumpSize"
#define XbCJumpSize  "JumpSize"
#define XbNlargeDragIcons  "largeDragIcons"
#define XbCLargeDragIcons  "LargeDragIcons"
#define XbNpattern  "pattern"
#define XbCPattern  "Pattern"
#define XbNprintCommand  "printCommand"
#define XbCPrintCommand  "PrintCommand"



// String defines for drag-and-drop atoms:

#define XbA_EXISTING_ANNOTATION  "existing_annotation"
#define XbA_CREATE_LINK  "create_link"
#define XbA_THUMBNAIL  "thumbnail"


// Label and button images are stored as arrays of four pixmaps.  The
// image and mask are loaded from the xpm data, and the label and arm
// pixmaps are created by replacing the transparent pixels in the
// image with the background and arm background colors.  The following
// defines serve as indices into the arrays:

#define PIXMAPS_IMAGE  0
#define PIXMAPS_MASK  1
#define PIXMAPS_LABEL_PIXMAP  2
#define PIXMAPS_ARM_PIXMAP  3

#define PIXMAPS_ARRAY_SIZE  4


// Global variables:

// Contexts for storing "this" pointers in drag-and-drop routines
extern XContext docwin_context, scan_context;

// The atoms for different drag data
extern Atom existing_annotation_atom, create_link_atom, thumbnail_atom;
// Atom for the delete window window manager protocol
extern Atom delete_window_atom;

// Pixmaps/pixmap arrays:
extern Pixmap annotation_icon_pixmaps[]; // the "unlinked" icon used
					 // to create annotations
extern Pixmap thumbnail_icon_pixmap;	 // a generic thumbnail page
extern Pixmap destination_icon_pixmaps[]; // the "linked" icon used in
					  // annotation dialogs
extern Pixmap note_pixmaps[];		  // the annotation text
					  // button icon
extern Pixmap link_pixmaps[];		  // the hyperlink button icon
extern Pixmap marker_pixmaps[];		  // the empty marker icon

// The dimensions of some of the pixmaps:
extern unsigned int annotation_width, annotation_height;
extern unsigned int marker_width, marker_height;
extern unsigned int destination_width, destination_height;
extern unsigned int note_width, note_height;
extern unsigned int link_width, link_height;

// The display and screen on which the application is running:
extern Display *dpy;
extern Screen *scrn;


// Global functions:

// Use the image and mask pixmaps, along with the background color
// provided, to create and return a new pixmap with the transparent
// pixels set to the background
extern Pixmap create_masked_pixmap(Pixmap, Pixmap, unsigned long);

// Assign the appropriate pixmaps from the array to XmNlabelPixmap and
// XmNarmPixmap
extern void assign_button_pixmaps(Widget, Pixmap[]);

// Assign the appropriate pixmap from the array to XmNlabelPixmap
extern void assign_label_pixmap(Widget, Pixmap[]);

// Swap two variables:
inline void swap(float &x, float &y) {const float temp = x; x = y; y = temp;}
inline void swap(Dimension &x, Dimension &y) {const Dimension temp = x; x = y; y = temp;}
inline void swap(int &x, int &y) {const int temp = x; x = y; y = temp;}
inline void swap(unsigned int &x, unsigned int &y)
{const unsigned int temp = x; x = y; y = temp;}
inline void swap(double &x, double &y) {const double temp = x; x = y; y = temp;}


// Application resource data:

struct AppData
{
  Pixel dim_page_background;
  Pixel level_one;
  Pixel level_two;
  Pixel level_three;
  Boolean antialias;
  int thumbnail_dpi;
  X11String gs_path;
  X11String gv_path;
  int jump_size;
  Boolean large_icons;
  X11String pattern;
  X11String print_command;
};


// The top-level shell class:

class TopShell: public UIObject {
  DECLARE_REDUCED_MEMBERS(TopShell);

public:
  static XtAppContext app;	// application context
  static AppData appdata;	// application resource data

  // Create a drag icon from the given pixmap
  static Widget create_drag_icon(Widget, Pixmap, unsigned int, unsigned int);
  // Create a blank rectangle icon with the given dimensions
  static Widget create_rect_icon(Widget, const int, const int);

  void delete_and_reload(DocWindow *, const String);

private:
  // Widgets for the top-level shell, form manager, buttons, and
  // dialogs:
  Widget toplevel, form, openbutton, infobutton, quitbutton;
  Widget filesb, info_dialog;
  Widget info_separator, info_stext, info_okbutton;

  // A collection of DocWindow instances
  OrderedCltn file_windows;

  // A collection of Ghostview pid's
  OrderedCltn ghostview_pids;

  void quitCB(Widget, XtPointer, XtPointer); // exit the application

  // delete the DocWindow instance passed as the client_data argument
  void delete_childCB(Widget, XtPointer, XtPointer);

  // create a new DocWindow to load a file
  void open_fileCB(Widget, XtPointer, XmFileSelectionBoxCallbackStruct*);
  void do_openfile(const char *);

public:

  // Create a new TopShell instance (there should be only
  // one/application)
  TopShell(int, char **);

  void main_loop()
    { XtAppMainLoop(app); }

public:
  bool operator==(const TopShell& a) const
    { return (this == &a); }
  bool operator!=(const TopShell& a) const
    { return !(*this==a); }
  virtual bool isEqual(const Object&) const;
#ifdef FULL_NIHCL
  virtual const Class* species() const
    { return &classDesc; }
#endif
};


class PID: public ReducedObj {
  DECLARE_REDUCED_MEMBERS(PID);

public:
  pid_t pid;

public:
  bool operator==(const PID& a) const
    { return (this == &a); }
  bool operator!=(const PID& a) const
    { return !(*this==a); }
  virtual bool isEqual(const Object&p) const
    {
#ifdef FULL_NIHCL
      return p.isSpecies(classDesc) && *this==castdown(p);
#else
      return this == &p;
#endif
    }

#ifdef FULL_NIHCL
  virtual const Class* species() const
    { return &classDesc; }
#endif
};


#endif
