
// This is file UIObject.C.  It implements a base class for deriving
// classes which support user-interface objects, described in
// UIObject.h.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/

#include "UIObject.h"
#include "TopShell.h"
#include <stdio.h>

#define String X11String
#define Object X11Object

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/Protocols.h>
#include <Xm/AtomMgr.h>
#include <Xm/RowColumn.h>
#include <Xm/CascadeB.h>
#include <Xm/PushB.h>

#undef String
#undef Object


#define	THIS	UIObject
#define	BASE	ReducedObj
#define	BASE_CLASSES BASE::desc()
#define	MEMBER_CLASSES
#define VIRTUAL_BASE_CLASSES

DEFINE_REDUCED_CLASS(UIObject,1,"$Header: /a/cellar62/home/ginsburg/browser/RCS/UIObject.C,v 2.0 1996/03/13 20:38:30 ginsburg Exp $",NULL)


bool UIObject::isEqual(const Object& p) const
// Test two objects for equality
{
#ifdef FULL_NIHCL
  return p.isSpecies(classDesc) && *this==castdown(p);
#else
  return this == &p;
#endif
}


// Register a callback by creating a new CBData structure,
// initializing the members, and registering do_callback as the actual
// callback.

void UIObject::register_cb(Widget widget, const char *name, CBType cbfn,
			   XtPointer call_data, UIObject *obj)
{
  CBData *datastruct = new CBData;
  datastruct->this_obj = obj;
  datastruct->data = call_data;
  datastruct->cb = cbfn;
  XtAddCallback(widget, name, (XtCallbackProc)do_callback,
		(XtPointer)datastruct);
}


// Register a window manager protocol callback by creating a new
// CBData structure, initializing the members, and registering
// do_callback as the actual protocol callback.

void UIObject::register_wm_cb(Widget widget, Atom atom, CBType callback,
			      XtPointer call_data, UIObject *obj)
{
  CBData *datastruct = new CBData;
  datastruct->this_obj = obj;
  datastruct->data =  call_data;
  datastruct->cb = callback;

  XmAddWMProtocolCallback(widget, atom, (XtCallbackProc)do_callback,
			  (XtPointer)datastruct);
}


// Invoke the callback described by the CBData structure referenced by
// client_data.

void UIObject::do_callback(Widget widget, CBData *client_data,
				   XtPointer call_data)
{
  ((client_data->this_obj)->*(client_data->cb))
    (widget, client_data->data, call_data);
}


// Create a menu as a child of parent, with the specified
// title and mnemonic.  The menu items are described by the array
// items, which is terminated by a final element with a NULL label.

// This function is derived from an example in the
// _Motif_Programming_Manual_, by Dan Heller.

Widget UIObject::BuildPulldownMenu(Widget parent, const char *menu_title,
				   const char mnemonic, MenuItem items[],
				   const int menu_type)
{
  // Create the menu and (optional) cascade button:

  Widget pulldown, cascade;

  if (menu_type == XmMENU_PULLDOWN)
    pulldown = XmCreatePulldownMenu(parent, "_pulldown", NULL, 0);
  else
    pulldown = XmCreatePopupMenu(parent, "_popup", NULL, 0);

  XmString str;

  if (menu_type == XmMENU_PULLDOWN)
    {
      str = XmStringCreateSimple((char*)menu_title);
      cascade = XtVaCreateManagedWidget(menu_title,
					xmCascadeButtonWidgetClass, parent,
					XmNsubMenuId, pulldown,
					XmNlabelString, str,
					XmNmnemonic, mnemonic,
					NULL);
      XmStringFree(str);
    }

  // Create the menu items
  Widget widget;
  for(int i = 0; items[i].label != NULL; i++)
    {
      // If the item is a submenu, create it recursively; otherwise
      // create the requested widget:
      if (items[i].subitems)
	widget = BuildPulldownMenu(pulldown, items[i].label,
				   items[i].mnemonic, items[i].subitems);
      else
	widget = XtVaCreateManagedWidget(items[i].label, *(items[i].wclass),
					 pulldown, NULL);

      // Set the mnemonic and accelerator:
      if(items[i].mnemonic)
	XtVaSetValues(widget, XmNmnemonic, items[i].mnemonic, NULL);

      if(items[i].accelerator)
	{
	  str = XmStringCreateSimple(items[i].accel_text);
	  XtVaSetValues(widget,
			XmNaccelerator, items[i].accelerator,
			XmNacceleratorText, str,
			NULL);
	  XmStringFree(str);
	}

      // Register the activation callback:
      if(items[i].cb)
	register_cb(widget, XmNactivateCallback, items[i].cb,
		    items[i].cb_data);
    }

  return (menu_type == XmMENU_PULLDOWN) ? cascade : pulldown;
}


// Register an action proc by creating a new APData structure,
// initializing the members, and registering do_action as the actual
// XtActionProc.  The address of the APData structure is printed to a
// string, and this buffer is returned.  This address must be the
// last argument to the action whenever it is used in a translation.

char *UIObject::register_ap(const char *name, APType ap, UIObject *this_obj)
{
  APData *datastruct = new APData;
  datastruct->this_obj = this_obj;
  datastruct->action = ap;

  XtActionsRec actions;
  actions.string = (char *)name;
  actions.proc = (XtActionProc) do_action;

  XtAppAddActions(TopShell::app, &actions, 1);

  char *buf = new char[20];

  sprintf(buf, "%p", datastruct);

  return buf;
}


// Retrieve the APData structure address from the last argument, then
// call the appropriate action proc

void UIObject::do_action(Widget widget, XEvent *event,
			 char *args[], int *num_args)
{
  if (*num_args < 1)
    {
      cerr << "ActionProc: invalid arguments" << endl;
      return;
    }

  int actual_num_args = *num_args - 1;

  // Retrieve the APData address:
  APData *data;
  sscanf(args[actual_num_args], "%p", &data);

  // call the action proc:
  ((data->this_obj)->*(data->action))
    (widget, event, args, actual_num_args);
}


// Get the coordinates of the specified widget with respect to the
// root window

void UIObject::get_root_coords(Widget widget, int &x_return, int &y_return)
{
/*  int x,y;
  Window root_return;
  unsigned int width_return, height_return;
  unsigned int border_width_return, depth_return;
  int x_accum = 0, y_accum = 0;
*/
  // Use XtTranslateCoordinates to avoid a roundtrip to the server:
  Position x_pos, y_pos;

  XtTranslateCoords(widget, 0, 0, &x_pos, &y_pos);

  x_return = x_pos;
  y_return = y_pos;

/*  Window win = XtWindow(widget);

  XGetGeometry(dpy, win, &root_return, &x, &y, &width_return,
	       &height_return, &border_width_return, &depth_return);

  XTranslateCoordinates(dpy, win, root_return, 
	       

  while(win != RootWindowOfScreen(XtScreen(widget)))
    {
      XGetGeometry(XtDisplay(widget), win, &root_return,
		   &x, &y, &width_return, &height_return,
		   &border_width_return, &depth_return);

      x_accum += x;
      y_accum += y;
      Window parent_return, *children;
      unsigned int n_return;

      XQueryTree(XtDisplay(widget), win, &root_return, &parent_return,
		 &children, &n_return);
      XFree(children);

      win = parent_return;
    }

  x_return = x_accum;
  y_return = y_accum;*/
}


// Create a new EHData structure, set its members, and add do_handler
// as the actual event handler with the structure as its client data.

void UIObject::register_eh(Widget widget, EventMask mask, EHType eh,
			   XtPointer client_data, Boolean nm_events)
{
  EHData *datastruct = new EHData;
  datastruct->this_obj = this;
  datastruct->data = client_data;
  datastruct->handler = eh;

  XtAddEventHandler(widget, mask, nm_events, (XtEventHandler) do_handler,
		    (XtPointer)datastruct);
}


// Invoke the event handler described by the EHData structure
// referenced by client_data

void UIObject::do_handler(Widget widget, EHData *client_data,
			  XEvent *event, Boolean *discard)
{
  ((client_data->this_obj)->*(client_data->handler))
    (widget,client_data->data,event,discard);
}


// Traverse the widget tree until the first shell parent is found

Widget UIObject::get_shell(Widget widget)
{
  while(!XtIsShell(widget))
    widget = XtParent(widget);

  return widget;
}


void UIObject::set_labelString(Widget widget, const char *label)
{
  XmString xm_label = XmStringCreateLtoR((char*)label, XmFONTLIST_DEFAULT_TAG);

  XtVaSetValues(widget,
		XmNlabelString, xm_label,
		NULL);

  XmStringFree(xm_label);
}


// client_data argument specifies the Widget ID of the popup menu
void UIObject::display_popupEH(Widget, XtPointer client_data,
			       XButtonPressedEvent *event, Boolean*)
{
  if (event->button != 3) return;

  XmMenuPosition((Widget)client_data, event);
  XtManageChild((Widget)client_data);
}
