//  This may look like C, but it is really -*- C++ -*-

// This is file UIObject.h.  It describes a base class for deriving
// classes which support user-interface objects, implemented in
// UIObject.C.

/*
DeckView - A browser for Postscript documents.
Version 1.0, BETA 1

Copyright (C) 1996  Adam Ginsburg
Design by Adam Ginsburg, Joe Marks, and Stuart M. Shieber

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

Author: Adam Ginsburg
	Aiken G-20
	33 Oxford Street
	Cambridge, MA 02138 USA

Comments and bug reports to:
	deckview@eecs.harvard.edu
*/


#ifndef UIObject_H
#define UIObject_H

// 	$Id: UIObject.h,v 2.0 1996/03/13 20:46:41 ginsburg Exp $


#include "ReducedObj.h"

#define String X11String
#define Object X11Object
#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#undef String
#undef Object



class UIObject;

// Function pointers to callback procs, action procs, and event
// handlers:
typedef void (UIObject::*CBType)(Widget,XtPointer,XtPointer);
typedef void (UIObject::*APType)(Widget,XEvent*,char*[],int);
typedef void (UIObject::*EHType)(Widget,XtPointer,XEvent*,Boolean*);



class UIObject: public ReducedObj {
  DECLARE_REDUCED_MEMBERS(UIObject);

protected:
  // Create and return a pulldown menu from a specification given as
  // an array of MenuItem structures:
  struct MenuItem
    {
      char *label;
      WidgetClass *wclass;
      char mnemonic;
      char *accelerator;
      char *accel_text;
      void (UIObject::*cb)(Widget,XtPointer,XtPointer);
      XtPointer cb_data;
      MenuItem *subitems;
    };

  Widget BuildPulldownMenu(Widget,const char *, const char, MenuItem[],
			   const int = XmMENU_PULLDOWN);

  // Add the specified function as a callback:
  void register_cb(Widget,const char *,
		   void (UIObject::*)(Widget,XtPointer,XtPointer),
		   XtPointer, UIObject*);

  void register_cb(Widget widget, const char *callback,
		   void (UIObject::*cb)(Widget,XtPointer,XtPointer),
		   XtPointer client_data)
    { register_cb(widget, callback, cb, client_data, this); }

  // Add a function as a window manager protocol callback:
  void register_wm_cb(Widget, Atom, CBType, XtPointer, UIObject*);
  void register_wm_cb(Widget w, Atom atom, CBType cb, XtPointer client_data)
    { register_wm_cb(w, atom, cb, client_data, this); }

  // Define the function as an actionProc and return a string that
  // must be the last argument when the action is used in a
  // translation:
  char *register_ap(const char *, APType, UIObject*);
  char *register_ap(const char *name, APType callback)
    { return register_ap(name, callback, this); }

  // Register an event handler:
  void register_eh(Widget widget, EventMask mask, EHType eh, XtPointer ptr)
    { register_eh(widget, mask, eh, ptr, False); }
  void register_eh(Widget, EventMask, EHType, XtPointer, Boolean);

  // The following functions manage/unmanage/popup/popdown the widget
  // specified as their second argument:
  void unmanageCB(Widget, XtPointer client_data, XtPointer)
    { XtUnmanageChild((Widget)client_data); }
  void manageCB(Widget, XtPointer client_data, XtPointer)
    { XtManageChild((Widget)client_data); }
  void popupCB(Widget, XtPointer client_data, XtPointer)
    { XtPopup((Widget)client_data, XtGrabNonexclusive); }
  void popdownCB(Widget, XtPointer client_data, XtPointer)
    { XtPopdown((Widget)client_data); }

  // Destroy the sourcePixmapIcon from a dragContext widget:
  static void destroyWidgetCB(Widget widget, XtPointer, XtPointer)
    { Widget icon; Arg args[1]; XtSetArg(args[0], "sourcePixmapIcon", &icon);
      XtGetValues(widget, args, 1); if (icon) XtDestroyWidget(icon); }

  // Display a popup menu at the button press location
  static void display_popupEH(Widget, XtPointer, XButtonPressedEvent*, Boolean*);

  // Get the location of a widget with respect to the root window
  void get_root_coords(Widget, int &, int &);
  // Get the shell containing a widget
  static Widget get_shell(Widget);
  // Set the XmNlabelString resource
  static void set_labelString(Widget, const char *);

private:
  // Data structures for callback, actionProc, and event handler
  // management:
  struct CBData
    {
      UIObject *this_obj; XtPointer data;
      void (UIObject::*cb)(Widget,XtPointer,XtPointer);
    };

  struct APData
    {
      UIObject *this_obj;
      APType action;
    };

  struct EHData
    {
      UIObject *this_obj;
      XtPointer data;
      EHType handler;
    };

  // The actual callback/action/event handler routines.  When a
  // function is registered, a new data structure is created with its
  // address, object address, and client_data pointer (except
  // actionProcs).  One of the following functions is registered with
  // the widget, and the data structure is passed as its client data
  // or last action argument:
  static void do_callback(Widget,CBData*,XtPointer);
  static void do_action(Widget, XEvent*, char*[], int*);
  static void do_handler(Widget, EHData *, XEvent*, Boolean*);

public:
  UIObject():ReducedObj() {};

  bool operator==(const UIObject& a) const
    { return (this == &a); }
  bool operator!=(const UIObject& a) const
    { return !(*this==a); }
  virtual bool isEqual(const Object&) const;
#ifdef FULL_NIHCL
  virtual const Class* species() const
    { return &classDesc; }
#endif
};


#endif
